module Copar.Algorithm
  ( refine
  , AlgoStatistics(..)
  , refineWithStats
  ) where

import           Control.Monad.ST
import           Data.STRef
import           Data.Proxy

import           Control.Monad.Extra (whileM)
import           Control.Monad.Reader

import           Data.BlockQueue (BlockQueue)
import qualified Data.BlockQueue as Queue
import           Data.MorphismEncoding
import           Data.Partition (Partition)
import qualified Data.RefinablePartition as Partition
import           Copar.RefinementInterface (RefinementInterface, Label, F1)
import           Copar.Algorithm.Types
import           Copar.Algorithm.Initialize
import           Copar.Algorithm.Split
import           Copar.Timing
import           Data.List.Utils (deleteLargestM)

-- | Metrics about the algorithm's execution.
data AlgoStatistics = AlgoStatistics
  { initialBlocks :: Int -- ^ Size of the initial partition
  , splitCount :: Int -- ^ Overall count of 'split' operations
  , initTime :: Integer -- ^ Time in picoseconds spend in the initialize
                        -- procedure
  , refineTime :: Integer -- ^ Time in picoseconds spend in the rest of the
                          -- algorithm (calling split until the queue is empty)
  , size1Skipped :: Int -- ^ How often did we skip a block instead of calling
                        -- update/splitBlock on it due to our
                        -- size-1-optimization.
  }

-- | Returns the count of 'split' calls
processQueue  :: RefinementInterface f => BlockQueue s -> AlgoState s f -> ST s Int
processQueue queue as = do
  count <- newSTRef 0
  whileM $ Queue.dequeue queue >>= \case
    Nothing -> return False
    Just block -> do
      modifySTRef' count (+1)
      runReaderT (split block) (as, queue)
      return True
  readSTRef count
{-# INLINE processQueue #-}

refine :: forall f s.
     RefinementInterface f
  => Proxy f
  -> Encoding (Label f) (F1 f)
  -> Bool
  -> ST s Partition
refine Proxy encoding enableOptimization = do
  queue <- Queue.empty (size encoding)
  (blocks, state) <- initialize @f encoding enableOptimization
  deleteLargestM (Partition.blockSize (partition state)) blocks >>= mapM_ (Queue.enqueue queue)

  _ <- processQueue queue state

  Partition.freeze (partition state)


-- | Same as 'refine', but also reports some statistics about the algorithm's
-- execution.
refineWithStats
  :: forall f
   . RefinementInterface f
  => Proxy f
  -> Encoding (Label f) (F1 f)
  -> Bool
  -> IO (Partition, AlgoStatistics)
refineWithStats Proxy encoding enableOptimization = do
  queue <- stToIO $ Queue.empty (size encoding)
  (initTime, (blocks, state)) <- withTime (stToIO $ initialize @f encoding enableOptimization)
  deleteLargestM (stToIO . Partition.blockSize (partition state)) blocks >>= mapM_ (stToIO . Queue.enqueue queue)
  let initialBlocks = length blocks

  (refineTime, splitCount) <- withTime (stToIO $ processQueue queue state)

  part <- stToIO $ Partition.freeze (partition state)
  size1Skipped <- stToIO $ readSTRef (size1Count state)

  let stats = AlgoStatistics initialBlocks splitCount initTime refineTime size1Skipped

  return (part, stats)
{-# INLINE refineWithStats #-}
