{-# LANGUAGE BangPatterns #-}

module Copar.Algorithm.Split
  (
  -- * Main interface
    SplitM
  , split

  -- * Internal functions

  -- | These are mainly exported for testing and benchmarking. Use 'split' or
  -- even better: 'Copar.Algorithm.refine'.

  , collectTouchedBlocks
  , updateBlock
  , splitBlock
  , addBlocksToQueue
  ) where

import           Prelude hiding (pred)

import           Control.Monad
import           Control.Monad.ST
import           Control.Monad.ST.Unsafe (unsafeSTToIO)
import           Data.Maybe (maybeToList)
import           Data.STRef
import           System.IO.Unsafe (unsafeDupablePerformIO)
import           Data.Ord (comparing)

import           Control.Monad.Extra (unlessM, whenM)
import           Control.Monad.Reader
import           Data.Tuple.Extra (snd3)
import qualified Data.Vector as V
import qualified Data.Vector.Mutable as VM
import qualified Data.Vector.Unboxed as VU

import           Data.List.Utils (deleteLargestM)
import           Data.Algorithm.PossibleMajorityCandidate
import           Data.BlockQueue (BlockQueue)
import qualified Data.BlockQueue as Queue
import           Data.MorphismEncoding
import           Data.RefinablePartition (Block)
import qualified Data.RefinablePartition as Partition
import           Copar.RefinementInterface (RefinementInterface, F3)
import qualified Copar.RefinementInterface as RI
import           Copar.Algorithm.Types

type SplitM s f = ReaderT (AlgoState s f, BlockQueue s) (ST s)

split :: RefinementInterface f => Block -> SplitM s f ()
split blockS = do
  (as, _) <- ask
  touchedBlocks <- collectTouchedBlocks blockS

  forM_ touchedBlocks $ \(b, v0) -> do
    updateBlock b v0
    whenM (lift $ Partition.hasMarked (partition as) b) $
      splitBlock b >>= addBlocksToQueue b
{-# INLINE split #-}

-- | Update weights for all marked states in the given block @b@.
--
-- This calls 'update' for all marked states in the block with @[labels to S]@
-- and @weightToC@ as parameters and saves the resulting weight to S in @lastW@.
-- @toSub@ in reset to an empty list for all marked states. Also, the @F3@
-- values returned from 'update' are saved in @f3Cache@.
--
-- As a precondition, toSub must not be empty for marked states.
updateBlock :: forall s f. RefinementInterface f => Block -> F3 f -> SplitM s f ()
updateBlock b v0 = ask >>= \(as, _) -> lift $ do
  markB <- Partition.markedStates (partition as) b
  VU.forM_ markB $ \x -> do
    -- We can use `head` here, since states are only marked if they have at
    -- least one edge into S => toSub[x] can't be empty.
    !pc <- (fromEdgeRef . head <$> VM.read (toSub as) x)
            >>= VM.read (lastW as)

    labelsToS <- VM.read (toSub as) x >>= mapM (\e -> do
      let Edge _ !lab _ = graph (encoding as) e
      return $! lab)

    (!wxS, !vx, !wxCwithoutS) <- RI.update @f labelsToS <$> readSTRef pc
    writeSTRef pc wxCwithoutS
    !ps <- newSTRef wxS

    VM.read (toSub as) x >>= \(!edges) -> forM_ edges $ \(EdgeRef !e) ->
      (VM.write (lastW as) $! e) $! ps

    VM.write (toSub as) x []

    if vx == v0
      then Partition.unmark (partition as) x
      else VM.write (f3Cache as) x $! vx
{-# INLINE updateBlock #-}

-- | Split block according to marked/unmarked status and saved F3s.
--
-- @b@ must have at least one marked state
--
-- Returns a list of new sub-blocks of @b@
splitBlock :: RefinementInterface f => Block -> SplitM s f [Block]
splitBlock b = ask >>= \(as, _) -> lift $ do
  -- b has marked states, so b1 is guaranteed to be non-empty
  (Just b1, bunmarked) <- Partition.splitMarked (partition as) b

  -- NOTE: We need to use unsafePerformIO here, because all vector sortBy
  -- functions expect a pure predicate. Since our predicate is only monadic
  -- because we need to _read_ from a mutable vector and doesn't have side
  -- effects, this should be safe.
  let unsafeF3 = unsafeDupablePerformIO . unsafeSTToIO . VM.read (f3Cache as)

  -- We can use unsafeStatesOfBlock here, since the resulting vector is
  -- immediatly consumed and not referenced afterwards.
  !pmc <- possibleMajorityCandidateBy' unsafeF3 <$>
          Partition.unsafeStatesOfBlock (partition as) b1

  -- the pmc occurs in b1, so b1' has to be non-empty
  (Just b1', b2) <- Partition.splitBy (partition as) b1
                    ((==pmc) . unsafeF3)

  ((b1':maybeToList bunmarked) ++) <$> case b2 of
    Nothing -> return []
    Just b2' -> Partition.groupBy (partition as) b2' (comparing unsafeF3)
{-# INLINE splitBlock #-}

-- | Add sub-blocks that were split off of a super-block to the queue.
--
-- If the original @b@ is already in the queue, we add all sub-blocks except the
-- one that shares its identity with @b@. If @b@ is not in the queue, we add all
-- blocks except a largest one.
--
-- Runtime: @O(|sub-blocks|)@
addBlocksToQueue
  :: RefinementInterface f
  => Block -- ^ Original super-block
  -> [Block] -- ^ List of split-off sub-blocks
  -> SplitM s f ()
addBlocksToQueue b blocks = ask >>= \(as, queue) -> lift $ do
  bInQueue <- Queue.elem b queue

  blocks'  <- if bInQueue
    then return blocks
    else deleteLargestM (Partition.blockSize (partition as)) blocks

  mapM_ (Queue.enqueue queue) blocks'
{-# INLINE addBlocksToQueue #-}

-- | Returns a list of blocks that have at least one predecessor state of the
-- given block @S@.
--
-- Such predecessor states are marked for subsequent splitting and their edges
-- into @S@ are added to @toSub@.
collectTouchedBlocks :: forall s f. RefinementInterface f => Block -> SplitM s f [(Block, F3 f)]
collectTouchedBlocks blockS = do
  (as, _) <- ask

  statesOfS <- lift $ Partition.statesOfBlock (partition as) blockS

  markedBlocks <- lift $ newSTRef []

  lift $ VU.forM_ statesOfS $ \y -> forM_ (pred as V.! y) $ \e -> do
    let Edge !x _ _ = graph (encoding as) e
    b <- Partition.blockOfState (partition as) x

    sizeOfB <- Partition.blockSize (partition as) b

    when (sizeOfB == 1 && size1Enabled as) $
      -- We found an edge with an outgoing block that has only one state.
      modifySTRef' (size1Count as) (+1)

    unless (sizeOfB == 1 && size1Enabled as) $ do
      unlessM (Partition.hasMarked (partition as) b) $ do
        wCx <- readSTRef =<< VM.read (lastW as) (fromEdgeRef e)
        let v0 = snd3 $ RI.update @f [] wCx
        modifySTRef markedBlocks ((b, v0):)

      whenM (null <$> VM.read (toSub as) x) $
        Partition.mark (partition as) x

      VM.modify (toSub as) (e:) x

  lift $ readSTRef markedBlocks
{-# INLINE collectTouchedBlocks #-}
