{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE AllowAmbiguousTypes #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ScopedTypeVariables #-}
module Copar.Coalgebra.Parser
  ( parseMorphisms
  , morphismsParser
  , SymbolTable(..)
  , module Copar.Coalgebra.Parser.Class
  , HasSanityChecks(noSanityChecks)
  , SanityChecks(..)
  ) where

import           Control.Monad (void, forM_)
import           Data.Bifunctor
import           Data.Tuple
import           Data.Void (Void)

import           Control.Monad.State.Strict (execStateT, MonadState)
import qualified Data.HashMap.Strict as M
import           Data.Text (Text)
import qualified Data.Text as T
import           Lens.Micro.Platform
import           Text.Megaparsec hiding (State)
import qualified Data.Vector as V
import qualified Data.Vector.Mutable as VM
import           Control.DeepSeq (NFData)

import           Data.MorphismEncoding (Encoding, State)
import qualified Data.MorphismEncoding as Encoding
import           Copar.Coalgebra.Parser.Internal
import           Copar.RefinementInterface
import           Copar.FunctorExpression.Sorts (Sort, Sorted(..))
import           Copar.FunctorExpression.Type
import           Copar.FunctorExpression.Desorting (Desorted, mkDesortedLabel)
import qualified Copar.Parser.Lexer as L
import           Copar.Parser.Types
import           Copar.Coalgebra.Parser.Class


class HasSanityChecks m where
  -- | When this returns 'True', morphism parsers are allowed (encouraged) to skip
  -- sanity checks on the input.
  --
  -- In this case, the user is responsible for providing correct inputs and nasal
  -- demons will be set free if the input is incorrect.
  noSanityChecks :: m Bool

instance MonadState (ParserState l f1) m => HasSanityChecks m where
  noSanityChecks = use disableSanity


newState :: MorphParser l f1 State
newState = nextState <<%= succ
{-# INLINE newState #-}

defineSymbol :: Text -> MorphParser l f1 State
defineSymbol sym = use (symbolTable . at sym) >>= \case
  Just (_, Defined) -> fail $ T.unpack sym ++ " defined twice" -- TODO: Better error
  Just (x, Undefined) -> define x
  Nothing -> newState >>= define

  where
    define :: State -> MorphParser l f1 State
    define node = (symbolTable . at sym .= Just (node, Defined))
      *> return node
{-# INLINE defineSymbol #-}

lookupSymbol :: Text -> MorphParser l f1 State
lookupSymbol sym = use (symbolTable . at sym) >>= \case
  Nothing -> do { x <- newState; markUndefined x; return x }
  Just (x, _) -> return x

  where
    markUndefined :: State -> MorphParser l f1 ()
    markUndefined node = symbolTable . at sym .= Just (node, Undefined)
{-# INLINE lookupSymbol #-}

checkUndefinedRefs :: MorphParser l f1 ()
checkUndefinedRefs = use (symbolTable . to M.toList . to (filter isUndefined)) >>= \case
  [] -> return ()
  ((sym,_):_) -> fail $ T.unpack sym ++ " is undefined"

  where
    isUndefined = (==Undefined) . snd . snd

newtype SymbolTable = SymbolTable { fromSymbolTable :: M.HashMap State Text }
  deriving (Show,Eq,Ord,NFData)

finalizeState :: forall f.
     ParserState (Label f) (F1 f) -> (SymbolTable, Encoding (Label (Desorted f)) (F1 (Desorted f)))
finalizeState state =
  let
    f1s = state ^. f1Map
    f1Vec = V.create $ do
      v <- VM.unsafeNew (M.size f1s)
      forM_ (M.toList f1s) $ \(i, f1) -> VM.unsafeWrite v i f1
      return v
    !edges = V.concat (state ^. graph)
    !symTab = M.fromList (map swap (M.toList (fmap fst (state ^. symbolTable))))

  in
    ( SymbolTable symTab
    , bimap (mkDesortedLabel @f) id (Encoding.new f1Vec edges)
    )

-- | Whether or not to do sanity checks on the input.
data SanityChecks = EnableSanityChecks | DisableSanityChecks

morphismsParser :: forall f.
     (Functor f, ParseMorphism f)
  => FunctorExpression f Sort
  -> SanityChecks
  -> Parser (SymbolTable, Encoding (Label (Desorted f)) (F1 (Desorted f)))
morphismsParser Variable _ =
  error "should not happen: variable" -- FIXME: Useful error message
morphismsParser (Functor sort f) sanity = finalizeState @f <$> execStateT p initial
  where
    initial = case sanity of
      EnableSanityChecks -> initState
      DisableSanityChecks -> initState & disableSanity .~ True
    p = do
      void (some parsePoint)
      checkUndefinedRefs

    parsePoint = do
      from <- L.name >>= defineSymbol
      void $ L.symbol ":"
      (f1, succs) <- parseMorphismPoint (fmap wrapper f)
      L.newlinesOrEof

      !succs' <- V.forM succs $ \(to, l) ->
        return $! Encoding.Edge from (Sorted sort l) to

      graph %= (succs':)
      f1Map %= M.insert from (Sorted sort f1)

parseMorphisms ::
     (Functor f, ParseMorphism f)
  => FunctorExpression f Sort
  -> SanityChecks
  -> String
  -> Text
  -> Either (ParseErrorBundle Text Void) ( SymbolTable
                                   , Encoding (Label (Desorted f)) (F1 (Desorted f)))
parseMorphisms expr = parse . morphismsParser expr

wrapper ::
     (Functor f, ParseMorphism f)
  => FunctorExpression f Sort
  -> MorphParser (Label f) (F1 f) State
wrapper Variable = L.name >>= lookupSymbol
wrapper (Functor nextSort f) = do
  from <- newState
  (f1, succs) <- parseMorphismPoint (fmap wrapper f)

  !succs' <- V.forM succs $ \(to, l) ->
    return $! Encoding.Edge from (Sorted nextSort l) to

  graph %= (succs':)
  f1Map %= M.insert from (Sorted nextSort f1)
  return from
