module Copar.Dot
  ( printDot
  , writeDot
  , DotConfig(..)
  , defaultDotConfig
  ) where

import           Data.Text.Lazy.Builder (Builder)
import qualified Data.Text.Lazy.Builder as Build
import qualified Data.Text.Lazy.Builder.Int as Build
import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Text.Lazy.IO as LazyIO
import           Data.Text (Text)
import qualified Data.Text as T
import qualified Data.HashMap.Strict as HM

import           Data.MorphismEncoding (Encoding)
import qualified Data.MorphismEncoding as Encoding
import           Copar.Coalgebra.Parser (SymbolTable(..))
import           Copar.FunctorExpression.Sorts (Sort, Sorted(..))
import           Copar.PrettyShow
import           Data.Partition (Partition)
import qualified Data.Partition as Part

data DotConfig = DotConfig
  { edgeLabels :: Bool
  , nodeLabels :: Bool
  }

defaultDotConfig :: DotConfig
defaultDotConfig = DotConfig
  { edgeLabels = True
  , nodeLabels = False
  }

printDot ::
     (PrettyShow label, PrettyShow f1)
  => DotConfig
  -> SymbolTable
  -> Encoding label (Sorted f1)
  -> Maybe Partition
  -> IO ()
printDot config symbolTable encoding maybePartition =
  LazyIO.putStr
    (Build.toLazyText (dot config symbolTable encoding maybePartition))

writeDot ::
     (PrettyShow label, PrettyShow f1)
  => FilePath
  -> DotConfig
  -> SymbolTable
  -> Encoding label (Sorted f1)
  -> Maybe Partition
  -> IO ()
writeDot file config symbolTable encoding maybePartition =
  (LazyIO.writeFile
     file
     (Build.toLazyText (dot config symbolTable encoding maybePartition)))

dot ::
     (PrettyShow label, PrettyShow f1)
  => DotConfig
  -> SymbolTable
  -> Encoding label (Sorted f1)
  -> Maybe Partition
  -> Builder
dot config symbolTable encoding maybePartition =
  let clusters =
        case maybePartition of
          Nothing ->
            nodes config
              symbolTable
              (Encoding.structure encoding)
              (Encoding.states encoding)
          Just partition ->
            foldMap
              (uncurry (cluster config symbolTable (Encoding.structure encoding)))
              (zip [0..] (Part.toBlocks partition))
   in "digraph {\n" <> "  compound=true;\n" <> clusters <>
      transitions config (Encoding.edges encoding) <>
      "}\n"

cluster ::
     PrettyShow f1
  => DotConfig
  -> SymbolTable
  -> Vector (Sorted f1)
  -> Int
  -> [Int]
  -> Builder
cluster config symbolTable structure block states =
  "subgraph cluster" <> Build.decimal block <> " {\n" <>
  nodes config symbolTable structure states <>
  "}\n"

nodes ::
     PrettyShow f1
  => DotConfig
  -> SymbolTable
  -> Vector (Sorted f1)
  -> [Int]
  -> Builder
nodes config symbolTable structure =
  foldMap (\i -> node config i (getSymbol i) (structure V.! i) <> "\n")
  where
    getSymbol i = HM.lookup i (fromSymbolTable symbolTable)

node ::
     PrettyShow f1 => DotConfig -> Int -> Maybe Text -> (Sorted f1) -> Builder
node config i name f1 =
  let label =
        case name of
          Nothing -> f1text
          Just n -> Build.fromText n <> "\\n" <> f1text
      color = sortColor (sortedSort f1)
   in "  s" <> Build.decimal i <> " [label=\"" <> label <> "\", " <>
      "style=\"filled\", fillcolor=\"" <>
      color <>
      "\"]"
  where
    f1text =
      if nodeLabels config
        then Build.fromText (escape (prettyShow (sortedElem f1)))
        else ""

transitions ::
     PrettyShow label => DotConfig -> Vector (Encoding.Edge label) -> Builder
transitions config edges = foldMap (\e -> transition config e <> "\n") edges

transition :: PrettyShow label => DotConfig -> Encoding.Edge label -> Builder
transition config (Encoding.Edge from label to) =
  "  s" <> Build.decimal from <> " -> s" <> Build.decimal to <>
  if edgeLabels config
    then " [label=\"" <> Build.fromText (escape (prettyShow label)) <> "\"]"
    else mempty

escape :: Text -> Text
escape = T.replace "\"" "'" -- TODO Better escaping


sortColor :: Sort -> Builder
sortColor s = "/pastel28/" <> Build.decimal ((s `mod` 8) + 1)
