{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE DerivingStrategies #-}

-- | Assignment and handling of "sorts" for nested functor expressions
module Copar.FunctorExpression.Sorts
  ( Sort
  , Sorted(..)
  -- * Annotation of functor expressions
  , annotateSorts
  -- * Extracting a sort table
  , sortTable
  , formatSortAsScope
  ) where

import           Data.Foldable
import           GHC.Generics

import           Control.Monad.State
import           Data.Text.Lazy.Builder as Build
import           Data.Text.Lazy.Builder.Int as Build
import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Vector.Mutable as VM
import           Control.DeepSeq (NFData)

import           Copar.FunctorExpression.Type
import           Copar.PrettyShow

-- | Sorts are basically unique integers for every sub-expression in a functor
-- expression.
newtype Sort = Sort Int
  deriving newtype (Eq, Show, Ord, Num, Integral, Real, Enum, NFData)

data Sorted a = Sorted
  { sortedSort :: {-# UNPACK #-} !Sort
  , sortedElem :: !a
  }
  deriving stock (Ord, Eq, Show, Generic, Functor)
  deriving anyclass (NFData)

instance PrettyShow a => PrettyShow (Sorted a) where
  prettyShow s = "(" <> prettyShow (sortedSort s) <> ", " <> prettyShow (sortedElem s) <> ")"

formatSortAsScope :: Sort -> Build.Builder
formatSortAsScope sort = "'" <> Build.decimal sort

-- | Assigns each sub-expression a different sort, starting with 1.
annotateSorts :: Traversable f => FunctorExpression f a -> FunctorExpression f Sort
annotateSorts expr = evalState (traverse anon expr) initState
  where
    -- Begin with state 0, because we increment it with (++i) instead of (i++)
    initState = 0

    -- increment state by one
    inc :: State Sort Sort
    inc = modify (+ 1) >> get

    -- write next sort from state into annotation of current sub-expression
    anon :: a -> State Sort Sort
    anon = const inc

-- | Return a vector mapping sorts to functors
--
-- ==== Example
--
-- >>> sortTable (Functor 1 (Just (Functor 2 (Just Variable))))
-- [Just (),Just ()]
sortTable :: (Foldable f, Functor f) => FunctorExpression f Sort -> Vector (f ())
sortTable expr = V.create $ do
  v <- VM.new (fromIntegral (foldr max 0 expr))
  populateVector v expr
  return v

  where
    populateVector _ Variable = return ()
    populateVector v (Functor sort f) = do
      VM.write v (fromIntegral sort - 1) (fmap (const ()) f)
      traverse_ (populateVector v) f
