{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE StrictData #-}
{-# LANGUAGE BangPatterns #-}
{-# OPTIONS_GHC -Wno-orphans #-}

module Copar.Functors.AbsorbingPolynomial
  ( AbsorbingPolynomial(..)
  , absorbInnerFunctors
  -- * Exported for easier testing
  , PolyF1(..)
  , Inner(..)
  , PolyLabel(..)
  )
where

import           Prelude                 hiding ( init )
import           Control.Monad
import           Data.Bifunctor
import           Data.List.NonEmpty             ( NonEmpty(..) )
import           GHC.Generics                   ( Generic )
import           Data.Maybe                     ( catMaybes )
import           Data.Either                    ( lefts
                                                , rights
                                                )
import qualified Data.List.NonEmpty            as NonEmpty
import           Control.Monad.ST
import           Unsafe.Coerce
import           Data.Foldable

import           Data.Eq.Deriving               ( deriveEq1 )
import           Text.Show.Deriving             ( deriveShow1 )
import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import qualified Data.Vector.Mutable           as VM
import qualified Data.Vector.Unboxed           as VU
import           Control.DeepSeq                ( NFData
                                                , rnf
                                                )
import           Lens.Micro.Platform
import           Data.Text                      ( Text )
import           Text.Megaparsec

import           Copar.Coalgebra.Parser
import qualified Copar.Parser.Lexer               as L
import           Copar.RefinementInterface
import           Copar.Functors.Polynomial  hiding ( PolyF1(..) )
import           Copar.FunctorExpression.Type
import           Copar.FunctorExpression.Printable
import           Copar.Parser.Types
import qualified Data.Vector.Utils             as V
import           Data.Primitive.SmallArray


data Inner f a = Direct a | Absorbed (f a)
  deriving (Functor, Foldable, Traversable)

$(deriveShow1 ''Inner)
$(deriveEq1 ''Inner)


newtype AbsorbingPolynomial f a = AbsorbingPolynomial (Polynomial (Inner f a))
  deriving (Functor, Foldable, Traversable)

instance (Printable f) => Printable (AbsorbingPolynomial f) where
  precedence _ = 10

  printMe cfg (AbsorbingPolynomial poly) =
    let name = withName cfg "AbsorbingPolynomial" 
    in name <> printPolynomial cfg ((\case
        Direct a -> a
        -- We want to always print the surrounding braces in case functor names
        -- are printed.
        Absorbed inner -> if precedence inner > 10 && name == mempty
          then printMe cfg inner
          else "(" <> printMe cfg inner <> ")"
      ) <$> poly)

$(deriveShow1 ''AbsorbingPolynomial)
$(deriveEq1 ''AbsorbingPolynomial)

absorbInnerFunctors
  :: Polynomial (FunctorExpression f a)
  -> Maybe (AbsorbingPolynomial f (FunctorExpression f a))
absorbInnerFunctors poly
  -- In case there aren't any nested functors beneath the polynomial to absorb,
  -- we just leave the original Polynomial and don't convert it to an
  -- AbsorbingPolynomial. This is because the refinement interface
  -- implementation of Polynomial is faster than that of AbsorbingPolynomial.
  | isLeaf poly = Nothing
  -- Otherwise we can absorb, since the reduced number of intermediate states
  -- outweighs the slower refinement interface.
  | otherwise   = Just $ AbsorbingPolynomial (fmap wrapInner poly)


-- | Check if this Polynomial has nested functors beneath it. Return True if it
-- doesn't.
isLeaf :: Polynomial (FunctorExpression f a) -> Bool
isLeaf = all isVariable
  where
    isVariable Variable = True
    isVariable _ = False

wrapInner
  :: FunctorExpression f a -> Inner f (FunctorExpression f a)
wrapInner Variable = Direct Variable
wrapInner (Functor _ inner) = Absorbed inner


-- | F1 for Polynomial
data PolyF1 f = PolyF1
  { polyF1Summand :: {-# UNPACK #-} Int
  -- ^ Index into sum
  , polyF1Variables :: {-# UNPACK #-} (SmallArray (Maybe (F1 f)))
  -- ^ F1 of inner functors (or Nothing for variables)
  , polyF1Constants :: {-# UNPACK #-} (VU.Vector Int)
  -- ^ Values of constant factors
  }
  deriving (Generic)

deriving instance Eq (F1 f) => Eq (PolyF1 f)
deriving instance Ord (F1 f) => Ord (PolyF1 f)
deriving instance Show (F1 f) => Show (PolyF1 f)
deriving instance NFData (F1 f) => NFData (PolyF1 f)

-- TODO To get rid of this orphan instance, we should probably create a newtype
-- wrapper around smallvec, probably even with a bit of useful API around it.
instance NFData a => NFData (SmallArray a) where
  rnf arr = rnfAll 0
    where
      rnfAll ix | ix < s    = rnf (indexSmallArray arr ix) `seq` rnfAll (ix + 1)
                | otherwise = ()
      !s = sizeofSmallArray arr


type instance F1 (AbsorbingPolynomial f) = PolyF1 f


-- | Index of this edge in the product and the inner label.
--
-- Note that this ignores constant factors, so an edge to the X in (2, 5, X)
-- would have 0 as Label, not 2. Also, the top-level co-product doesn't appear
-- in the label at all. It already appears in "F1".
data PolyLabel f = PolyLabel {-# UNPACK #-} Int (Maybe (Label f))
  deriving (Generic)

deriving instance Show (Label f) => Show (PolyLabel f)
deriving instance NFData (Label f) => NFData (PolyLabel f)

type instance Label (AbsorbingPolynomial f) = PolyLabel f


-- NB: We use SmallArray for Weight and F3, because they are slightly more memory
-- efficient for small sizes than Vector. These arrays have as many entries as
-- there are factors in the product which is usually rather few compared the the
-- number of states and edges in the encoding.


data PolyWeightEntry f =
    PolyWeightToSub
  | PolyWeightToRest
  | PolyWeightWrapped (Weight f)

-- | Defined as F2
--
-- For each variable in the product, this is True if the corresponding successor
-- state belongs to the subblock and False otherwise.
--
-- For functors this just wraps the Weight of the inner functor.
type instance Weight (AbsorbingPolynomial f) = SmallArray (PolyWeightEntry f)


data PolyF3Entry f =
    PolyF3ToRest
  | PolyF3ToCompound
  | PolyF3ToSub
  | PolyF3Wrapped (F3 f)

deriving instance Eq (F3 f) => Eq (PolyF3Entry f)
deriving instance Ord (F3 f) => Ord (PolyF3Entry f)

-- | Same as for Weight, but now with the three states in the left component.
type instance F3 (AbsorbingPolynomial f) = SmallArray (PolyF3Entry f)


instance RefinementInterface f => RefinementInterface (AbsorbingPolynomial f) where
  init
    :: F1 (AbsorbingPolynomial f)
    -> [Label (AbsorbingPolynomial f)]
    -> Weight (AbsorbingPolynomial f)
  init f1 labels =
    smallArrayFromListN (sizeofSmallArray (polyF1Variables f1)) $ map
      (\(i, x) -> (maybe PolyWeightToSub (PolyWeightWrapped . initInner i)) x)
      (zip [0 ..] (toList (polyF1Variables f1)))
    where
      initInner i innerF1 =
        init @f innerF1 (catMaybes $ labelsByIndex V.! i)
      labelsByIndex = V.create $ do
        v <- VM.replicate (sizeofSmallArray (polyF1Variables f1)) []
        forM_ labels $ \(PolyLabel i l) -> VM.modify v (l :) i
        return v

  update
    :: [Label (AbsorbingPolynomial f)]
    -> Weight (AbsorbingPolynomial f)
    -> ( Weight (AbsorbingPolynomial f)
       , F3 (AbsorbingPolynomial f)
       , Weight (AbsorbingPolynomial f)
       )
  update labels w = runST $ do
    let size = sizeofSmallArray w

    v <- newSmallArray size []
    forM_ labels $ \(PolyLabel i l) -> do
      readSmallArray v i >>= \ls -> writeSmallArray v i (l : ls)

    w1 <- newSmallArray size (unsafeCoerce (0 :: Int))
    f3 <- newSmallArray size (unsafeCoerce (0 :: Int))
    w2 <- newSmallArray size (unsafeCoerce (0 :: Int))

    forM_ [0 .. size - 1] $ \i -> do
      ls <- readSmallArray v i
      (!a, !b, !c) <- doUpdate ls <$> indexSmallArrayM w i
      writeSmallArray w1 i a
      writeSmallArray f3 i b
      writeSmallArray w2 i c

    (,,)
      <$> unsafeFreezeSmallArray w1
      <*> unsafeFreezeSmallArray f3
      <*> unsafeFreezeSmallArray w2
    where
      doUpdate
        :: [Maybe (Label f)]
        -> PolyWeightEntry f
        -> (PolyWeightEntry f, PolyF3Entry f, PolyWeightEntry f)
      doUpdate ls (PolyWeightWrapped inner) =
        let (!w1, !f3, !w2) = update @f (catMaybes ls) inner
        in  (PolyWeightWrapped w1, PolyF3Wrapped f3, PolyWeightWrapped w2)
      doUpdate _ PolyWeightToRest =
        (PolyWeightToRest, PolyF3ToRest, PolyWeightToRest)
      doUpdate [] PolyWeightToSub =
        (PolyWeightToRest, PolyF3ToCompound, PolyWeightToSub)
      doUpdate _ PolyWeightToSub =
        (PolyWeightToSub, PolyF3ToSub, PolyWeightToRest)


-- | Parse morphism for inner functor
--
-- If the inner value is just a variable, return (Nothing, [(x, Nothing)]). If
-- the inner value is a wrapped functor, return (f1, labels) that the inner
-- parser returns, wrapped in 'Just' constructors.
parseInner
  :: (Ord x, ParseMorphism f)
  => Inner f (MorphParser l f1 x)
  -> MorphParser
       l
       f1
       (Maybe (F1 f), Vector (x, Maybe (Label f)))
parseInner (Absorbed inner) = do
  (f1, labels) <- parseMorphismPoint inner
  return (Just f1, V.map (second Just) labels)
parseInner (Direct inner) = do
  x <- inner
  return (Nothing, V.singleton (x, Nothing))


instance ParseMorphism f => ParseMorphism (AbsorbingPolynomial f) where
  parseMorphismPoint (AbsorbingPolynomial (Polynomial sum)) = parseSum1 sum


----------- Coproducts parser

-- | Parse either a single product or an injection into the coproduct, depending
-- on the number of co-factors.
parseSum1
  :: (Ord a, ParseMorphism f)
  => Sum (Inner f (MorphParser l f1 a))
  -> MorphParser
       l
       f1
       (F1 (AbsorbingPolynomial f), Vector (a, Label (AbsorbingPolynomial f)))
parseSum1 sum@(Sum (_ :| [])) = do
  -- only a single summand => parse product directly

  -- We first try to parse 'inj i' and delegate to the product parser only if
  -- that fails.
  -- This avoids strange situations where a constant called 'inj' exists and the
  -- input starts with inj.
  (try parseSumPrefix >>= parseSum sum) <|> (parseSum sum 0)
parseSum1 other =
  parseSumPrefix
    >>= parseSum other -- otherwise, require 'inj'
    <?> "coproduct injection"


-- | parses @inj i@ where @i@ is a decimal integer
parseSumPrefix :: MonadParser m => m Int
parseSumPrefix = L.symbol "inj" *> L.decimal


-- | Parse an injection into the coproduct with the syntax 'inj i _'
parseSum
  :: (Ord a, ParseMorphism f)
  => Sum (Inner f (MorphParser l f1 a))
  -> Int
  -> MorphParser
       l
       f1
       (F1 (AbsorbingPolynomial f), Vector (a, Label (AbsorbingPolynomial f)))
parseSum (Sum summands) i = do
  sanity <- not <$> noSanityChecks
  when (sanity && (i < 0 || i >= length summands))
    $ fail ("polynomial: injection " ++ show i ++ " is out of bounds")

  (constants, successors) <- parseProduct1 (summands NonEmpty.!! i)
  let
    f1 = PolyF1
      { polyF1Summand   = i
      , polyF1Constants = constants
      , polyF1Variables =
        smallArrayFromListN (V.length successors) $ map fst $ V.toList
          successors
      }

  return (f1, foldMap snd successors)


----------- Products parser

-- | Parse either a single factor without parens or a tuple.
parseProduct1
  :: (Ord a, ParseMorphism f)
  => Product (Inner f (MorphParser l f1 a))
  -> MorphParser
       l
       f1
       ( VU.Vector Int
       , Vector (Maybe (F1 f), Vector (a, (PolyLabel f)))
       )
parseProduct1 product@(Product (factor :| [])) =
  let mkProduct = either
        (\i -> (VU.singleton i, V.empty))
        (\v ->
          (VU.empty, V.imap (\i x -> x & _2 . each . _2 %~ (PolyLabel i)) v)
        )
  in  (mkProduct <$> parseFactor factor) <|> parseProduct product
parseProduct1 other = parseProduct other


parseProduct
  :: (Ord a, ParseMorphism f)
  => Product (Inner f (MorphParser l f1 a))
  -> MorphParser
       l
       f1
       ( VU.Vector Int
       , Vector (Maybe (F1 f), Vector (a, (PolyLabel f)))
       )
parseProduct (Product l@(f :| fs)) =
  label ("a product of " ++ show (length l) ++ " element(s)") $ L.parens $ do
    factors <-
      (:) <$> parseFactor f <*> (traverse (\x -> L.comma *> parseFactor x) fs)

    let constants  = VU.fromList (lefts factors)
        vars       = V.concat (rights factors)
  --      labels    = factors ^.. each . _Right
        successors = V.imap (\i x -> x & _2 . each . _2 %~ (PolyLabel i)) vars

    return (constants, successors)


----------- Factor parser

-- | Returns either the value of a constants, or (f1, successors) for every
-- inner functor (There can be more than one because of exponentials).
parseFactor
  :: (Ord a, ParseMorphism f)
  => Factor (Inner f (MorphParser l f1 a))
  -> MorphParser
       l
       f1
       ( Either
           Int
           ( Vector
               ( Maybe (F1 f)
               , Vector (a, Maybe (Label f))
               )
           )
       )
parseFactor (Const (ExplicitSet names)) = do
  !f1 <- Left <$> someName names
  return f1 -- const has no successors
parseFactor (Const IntSet) = do
  x <- L.signed L.decimal <?> "integer"
  return (Left x)
parseFactor (Const NatSet) = do
  x <- L.decimal <?> "natural number"
  return (Left x)
parseFactor (Const (FiniteNatSet n)) = do
  x      <- L.decimal <?> ("natural number small than " ++ show n)
  sanity <- not <$> noSanityChecks
  when (sanity && x >= n) $ fail
    (  "out of range constant: "
    ++ show x
    ++ "(must be between 0 and "
    ++ show n
    ++ ")"
    )
  return (Left x)
parseFactor (Identity inner) = do
  successors <- parseInner inner
  return (Right (V.singleton successors))
parseFactor (Exponential inner exp) = L.braces $ do
  sanity     <- not <$> noSanityChecks

  successors <-
    (if sanity then V.sortOn fst else id)
    .       V.fromList
    <$>     ((,) <$> parseExpValue exp <*> (L.colon *> parseInner inner))
    `sepBy` L.comma

  when (sanity && allExpValues exp /= (V.map fst successors))
    $ fail ("exponential: map must be well-defined on " ++ showExp exp)

  return (Right (V.map snd successors))

parseExpValue :: (MonadParser m, HasSanityChecks m) => Exponent -> m Int
parseExpValue (ExplicitExp  names) = someName names
parseExpValue (FiniteNatExp n    ) = do
  x      <- L.decimal
  sanity <- not <$> noSanityChecks
  -- L.decimal returns only positive ints
  when (sanity && x >= n) $ fail
    (  "Value "
    ++ show x
    ++ "is out of bounds. (must be between 0 and "
    ++ show n
    ++ ")"
    )
  return x

allExpValues :: Exponent -> Vector Int
allExpValues (ExplicitExp  names) = V.enumFromN 0 (length names)
allExpValues (FiniteNatExp n    ) = V.enumFromN 0 n

showExp :: Exponent -> String
showExp (ExplicitExp  names) = show names
showExp (FiniteNatExp n    ) = "{0.." ++ show n ++ "}"

someName :: MonadParser m => Vector Text -> m Int
someName v = do
  name <- try L.name
  -- TODO Better error
  maybe empty return (V.elemIndex name v)
