{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- | The bag functor.
--
-- This is essentially a list without respect for order (but law). Formally,
-- it's the monoid-valued functor for the monoid @(ℕ, +, 0)@.
--
-- Coalgebras for this functor correspond to nondeterministic (unlabeled)
-- transition systems.
module Copar.Functors.Bag (Bag(..)) where

import           Prelude hiding (init)

import qualified Data.Vector as V
import           Text.Megaparsec
import           Data.Eq.Deriving (deriveEq1)
import           Text.Show.Deriving (deriveShow1)
import qualified Data.Text.Prettyprint     as Doc
import           Data.Text.Prettyprint     ((<+>))

import           Copar.RefinementInterface
import qualified Copar.Parser.Lexer as L
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Printable
import           Copar.Coalgebra.Parser
import           Copar.Functors.GroupValued
import           Copar.FunctorDescription

newtype Bag a = Bag a
  deriving (Functor,Foldable,Traversable)

instance Printable Bag where
  precedence _ = 40

  printMe cfg (Bag x) = withName cfg "Bag" <> "B" <> x

instance FunctorDescription Bag where
  name = "Bag"
  syntaxExample = "BX | ƁX"
  description = Just bagHelp
  precedence = 40
  functorExprParser = prefix ((L.symbol "B" <|> L.symbol "Ɓ") >> pure Bag)

bagHelp :: Doc.Doc Doc.AnsiStyle
bagHelp =
  Doc.reflow "This is like the Powerset functor, except that states can have \
             \multiple edges to the same successor. I.e the successors for a \
             \state form a multiset instead of a set."
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "B(F) | Ɓ(F)"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' x ',' ... '}'"

type instance Label Bag = Label (GroupValued Int)
type instance Weight Bag = Weight (GroupValued Int)
type instance F1 Bag = F1 (GroupValued Int)
type instance F3 Bag = F3 (GroupValued Int)

$(deriveEq1 ''Bag)
$(deriveShow1 ''Bag)

deriving instance Show (Bag ())

instance ParseMorphism Bag where
  parseMorphismPoint (Bag inner) = L.braces $ do
    successors <- V.fromList <$> inner `sepBy` L.comma

    let f1 = length successors

    return (f1, V.map (,1) successors)

instance RefinementInterface Bag where
  init = init @(GroupValued Int)
  update = update @(GroupValued Int)
