{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE BangPatterns #-}

module Copar.Functors.Distribution (Distribution(..)) where

import           Prelude hiding (init)
import           Control.Monad (when)

import           Text.Megaparsec
import           Data.Eq.Deriving (deriveEq1)
import           Text.Show.Deriving (deriveShow1)
import qualified Data.Text.Prettyprint     as Doc
import           Data.Text.Prettyprint     ((<+>))
import           Control.Monad.Extra (unlessM)

import           Data.Float.Utils (EqDouble)
import           Copar.Coalgebra.Parser
import           Copar.FunctorDescription
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Printable
import           Copar.Functors.GroupValued
import qualified Copar.Parser.Lexer as L
import           Copar.RefinementInterface


newtype Distribution x = Distribution x
  deriving (Functor, Foldable, Traversable)

instance Printable Distribution where
  precedence _ = 30

  printMe cfg (Distribution x) = withName cfg "Distribution" <> "D" <> x

instance FunctorDescription Distribution where
  name = "Distribution"
  syntaxExample = "DX | ƊX"
  description = Just distHelp
  precedence = 30
  functorExprParser = prefix ((L.symbol "D" <|> L.symbol "Ɗ")
                              >> pure Distribution)

distHelp :: Doc.Doc Doc.AnsiStyle
distHelp =
  Doc.reflow "Coalgebras for the distribution functor correspond to markov \
             \chains. Each state has a probability distribution over the \
             \successor states"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "DX | ƊX"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' X ':' real, ... '}'"
  <> Doc.line
  <> Doc.reflow "where the 'real's sum to 1 and zeros can be omitted."

type instance Label Distribution = Label (GroupValued EqDouble)
type instance Weight Distribution = Weight (GroupValued EqDouble)
type instance F1 Distribution = F1 (GroupValued EqDouble)
type instance F3 Distribution = F3 (GroupValued EqDouble)

$(deriveEq1 ''Distribution)
$(deriveShow1 ''Distribution)

deriving instance Show (Distribution ())

instance ParseMorphism Distribution where
  parseMorphismPoint (Distribution inner) = do
    (f1, succs) <- parseMorphismPoint (GroupValued @EqDouble inner)

    unlessM noSanityChecks $ do
      when (f1 /= 1) $
        fail "distribution: Sum of outgoing labels is not 1"

      when (not (all (\x -> x >= 0 && x <= 1) (fmap snd succs))) $ do
        fail "distribution: Weights must be between 0 and 1"

    return (f1, succs)

instance RefinementInterface Distribution where
  init _ _ = 1
  update weightsToS w =
    let
      !toS = sum weightsToS
      !toCwithoutS = w - toS
      !f3 = mkGroupF3 toCwithoutS toS
    in (toS, f3, toCwithoutS)
