{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE TemplateHaskell #-}

-- | The monoid valued functor @M^X@.
--
-- The 'RefinementInterface' for this functor is implemented in an abstract
-- fashion over any type that satisfies the 'Num' constraint, but concrete
-- parsers only exist for integers and reals.
module Copar.Functors.GroupValued
  ( GroupValued(..)
  , IsGroupF3(..)
  , OrderedComplex(..)
  , IntValued
  , RealValued
  , RationalValued
  , ComplexValued
  ) where

import           Control.Monad (when)
import           Data.Complex
import           Control.DeepSeq (NFData)
import           Data.Ratio
import           Data.Functor.Classes

import           Data.Vector (Vector)
import qualified Data.Vector as V
import           Text.Megaparsec
import           Text.Show.Deriving (deriveShow1)
import qualified Data.Text.Prettyprint     as Doc
import           Data.Text.Prettyprint     ((<+>))

import           Data.Float.Utils (EqDouble)
import qualified Data.Vector.Utils as V
import           Copar.RefinementInterface
import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Printable
import qualified Copar.Parser.Lexer as L
import           Copar.Parser.Types
import           Copar.FunctorDescription

data GroupValued m a = GroupValued a

instance Eq1 (GroupValued m) where
  liftEq f (GroupValued a1) (GroupValued a2) = f a1 a2

deriving instance Show (GroupValued m ())
deriving instance Functor (GroupValued m)
deriving instance Foldable (GroupValued m)
deriving instance Traversable (GroupValued m)

$(deriveShow1 ''GroupValued)

instance Printable (GroupValued Int) where
  precedence _ = 150

  printMe cfg (GroupValued x) = withName cfg "Integer-valued" <> "Z^" <> x

instance Printable (GroupValued EqDouble) where
  precedence _ = 140

  printMe cfg (GroupValued x) = withName cfg "Real-valued" <> "R^" <> x


instance Printable (GroupValued OrderedComplex) where
  precedence _ = 130

  printMe cfg (GroupValued x) = withName cfg "Complex-valued" <> "C^" <> x

instance Printable (GroupValued (Ratio Int)) where
  precedence _ = 120

  printMe cfg (GroupValued x) = withName cfg "Rational-valued" <> "Q^" <> x


type IntValued = GroupValued Int
instance FunctorDescription (GroupValued Int) where
  name = "Integer-valued"
  syntaxExample = "Z^X | ℤ^X"
  description = Just intHelp
  precedence = 150
  functorExprParser =
      prefix ((L.symbol "Z" <|> L.symbol "ℤ") >> L.symbol "^" >> pure GroupValued)

intHelp :: Doc.Doc Doc.AnsiStyle
intHelp =
  Doc.reflow "Weighted systems with integer weights"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "Z^X | ℤ^X"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' X ':' int, ... '}'"

type RealValued = GroupValued EqDouble
instance FunctorDescription (GroupValued EqDouble) where
  name = "Real-valued"
  syntaxExample = "R^X | ℝ^X"
  description = Just realHelp
  precedence = 140
  functorExprParser = prefix
    ((L.symbol "R" <|> L.symbol "ℝ") >> L.symbol "^" >> pure GroupValued)

realHelp :: Doc.Doc Doc.AnsiStyle
realHelp =
  Doc.reflow "Weighted systems with real weights"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "R^X | ℝ^X"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' X ':' real, ... '}'"

type RationalValued = GroupValued (Ratio Int)
instance FunctorDescription (GroupValued (Ratio Int)) where
  name = "Rational-valued"
  syntaxExample = "Q^X | ℚ^X"
  description = Just rationalHelp
  precedence = 120
  functorExprParser = prefix
    ((L.symbol "Q" <|> L.symbol "ℚ") >> L.symbol "^" >> pure GroupValued)

rationalHelp :: Doc.Doc Doc.AnsiStyle
rationalHelp =
  Doc.reflow ("Weighted systems with rational weights")
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "Q^X | ℚ^X"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' X ':' rational, ... '}'"
  <> Doc.line
  <> Doc.reflow "where 'rational' is either an integer or of the form a/b, \
                \where a and b are integers."

newtype OrderedComplex = OrderedComplex (Complex EqDouble)
  deriving (Eq, Num, NFData, Show)

instance Ord OrderedComplex where
  compare (OrderedComplex a) (OrderedComplex b) = case (a, b) of
    (r1 :+ i1, r2 :+ i2) -> compare r1 r2 <> compare i1 i2

type ComplexValued = GroupValued OrderedComplex
instance FunctorDescription (GroupValued OrderedComplex) where
  name = "Complex-valued"
  syntaxExample = "C^X | ℂ^X"
  description = Just complexHelp
  precedence = 130
  functorExprParser = prefix
    ((L.symbol "C" <|> L.symbol "ℂ") >> L.symbol "^" >> pure GroupValued)

complexHelp :: Doc.Doc Doc.AnsiStyle
complexHelp =
  Doc.reflow ("Weighted systems with complex weights")
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "C^X | ℂ^X"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' X ':' complex, ... '}'"
  <> Doc.line
  <> Doc.reflow "where 'complex' is of the form 'a + ib"


class IsGroupF3 m where
  data GroupF3 m
  f3ToCompound :: GroupF3 m -> m
  f3ToSub :: GroupF3 m -> m
  mkGroupF3 :: m -> m -> GroupF3 m


instance IsGroupF3 Int where
  data GroupF3 Int = IntGroupF3 {-# UNPACK #-} !Int {-# UNPACK #-} !Int
  f3ToCompound (IntGroupF3 x _) = x
  f3ToSub (IntGroupF3 _ x) = x
  mkGroupF3 = IntGroupF3

instance IsGroupF3 EqDouble where
  data GroupF3 EqDouble = EqDoubleGroupF3 {-# UNPACK #-} !EqDouble {-# UNPACK #-} !EqDouble
  f3ToCompound (EqDoubleGroupF3 x _) = x
  f3ToSub (EqDoubleGroupF3 _ x) = x
  mkGroupF3 = EqDoubleGroupF3

instance IsGroupF3 OrderedComplex where
  data GroupF3 OrderedComplex = OrderedComplexGroupF3 {-# UNPACK #-} !OrderedComplex {-# UNPACK #-} !OrderedComplex
  f3ToCompound (OrderedComplexGroupF3 x _) = x
  f3ToSub (OrderedComplexGroupF3 _ x) = x
  mkGroupF3 = OrderedComplexGroupF3

instance IsGroupF3 (Ratio Int) where
  data GroupF3 (Ratio Int) = RationalGroupF3 {-# UNPACK #-} !(Ratio Int) {-# UNPACK #-} !(Ratio Int)
  f3ToCompound (RationalGroupF3 x _) = x
  f3ToSub (RationalGroupF3 _ x) = x
  mkGroupF3 = RationalGroupF3

instance (Eq a, IsGroupF3 a) => Eq (GroupF3 a) where
  x == y =
         (f3ToCompound x == f3ToCompound y)
      && (f3ToSub x == f3ToSub y)

instance (Ord a, IsGroupF3 a) => Ord (GroupF3 a) where
  compare x y =
         compare (f3ToCompound x) (f3ToCompound y)
      <> compare (f3ToSub x)      (f3ToSub y)

instance (Show a, IsGroupF3 a) => Show (GroupF3 a) where
  showsPrec p x =
    showParen (p > 10)
      $ showString "GroupF3 "
      . showsPrec 11 (f3ToCompound x)
      . showChar ' '
      . showsPrec 11 (f3ToSub x)


type instance Label (GroupValued m) = m
type instance Weight (GroupValued m) = m
type instance F1 (GroupValued m) = m
type instance F3 (GroupValued m) = GroupF3 m

parseMorphismPointHelper ::
     (Num w, Ord x, MonadParser m)
  => m x
  -> m w
  -> Bool
  -> m (w, Vector (x, w))
parseMorphismPointHelper inner weightParser sanity = do
  !successors <- case sanity of
    True -> do
      succs <- V.sortOn fst . V.fromList <$> L.braces (edge `sepBy` L.comma)

      when (V.hasDuplicates (fmap fst succs)) $
        fail "group valued: Duplicate edges"

      return succs

    False -> V.fromList <$> L.braces (edge `sepBy` L.comma)

  let !f1 = V.sum (V.map snd successors)
  return (f1, successors)

  where
    edge = (,) <$> inner <*> (L.colon *> weightParser)
{-# INLINE parseMorphismPointHelper #-}
{-# SPECIALIZE parseMorphismPointHelper :: MorphParser l f1 Int -> MorphParser l f1 Int -> Bool -> MorphParser l f1 (Int, Vector (Int, Int)) #-}
{-# SPECIALIZE parseMorphismPointHelper :: MorphParser l f1 Int -> MorphParser l f1 EqDouble -> Bool -> MorphParser l f1 (EqDouble, Vector (Int, EqDouble)) #-}

instance ParseMorphism (GroupValued Int) where
  parseMorphismPoint (GroupValued inner) =
    parseMorphismPointHelper inner (L.signed L.decimal)
      =<< (not <$> noSanityChecks)

instance ParseMorphism (GroupValued EqDouble) where
  parseMorphismPoint (GroupValued inner) =
    parseMorphismPointHelper inner (L.signed L.adouble)
      =<< (not <$> noSanityChecks)

instance ParseMorphism (GroupValued OrderedComplex) where
  parseMorphismPoint (GroupValued inner) =
    parseMorphismPointHelper inner (OrderedComplex <$> L.complex L.adouble)
      =<< (not <$> noSanityChecks)

instance ParseMorphism (GroupValued (Ratio Int)) where
  parseMorphismPoint (GroupValued inner) =
    parseMorphismPointHelper inner (L.signed (L.fraction L.decimal))
      =<< (not <$> noSanityChecks)

instance (IsGroupF3 m, Ord m, Num m) => RefinementInterface (GroupValued m) where
  {-# SPECIALIZE instance RefinementInterface (GroupValued Int) #-}
  {-# SPECIALIZE instance RefinementInterface (GroupValued EqDouble) #-}
  {-# SPECIALIZE instance RefinementInterface (GroupValued OrderedComplex) #-}
  init :: F1 (GroupValued m) -> [Label (GroupValued m)] -> Weight (GroupValued m)
  init _ weights = let !x = (sum weights) in x

  update :: [Label (GroupValued m)] -> Weight (GroupValued m)
         -> (Weight (GroupValued m), F3 (GroupValued m), Weight (GroupValued m))
  update weightsToS !w =
    let
      !toS = sum weightsToS
      !toCwithoutS = w - toS
    in
      ( toS
      , mkGroupF3 toCwithoutS toS
      , toCwithoutS
      )
