{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE AllowAmbiguousTypes #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE FlexibleInstances #-}

-- | This module implements the @M^X@ functor, where @M@ is a monoid that
-- doesn't have to be a group like in "Copar.Functors.GroupValued".
--
-- The 'RefinementInterface' implementation for such functors doesn't fulfil the
-- same runtime complexity criteria as the other functors and it also uses tons
-- of space, but it works and satisfies the axioms.
module Copar.Functors.MonoidValued 
  ( SlowMonoidValued(..)
  , MaxIntValued
  , MinIntValued
  , MaxDoubleValued
  , MinDoubleValued
  , BitAndValued
  , BitOrValued
  )
where

import           Data.List                      ( foldl', intersperse )
import           Data.Semigroup                 ( Max(..), Min(..) )
import           Data.Functor.Classes
import           Control.Monad
import           Data.Foldable
import           Data.Word

import qualified Data.Vector                   as V
import           Text.Megaparsec
import qualified Data.Text.Prettyprint         as Doc
import           Data.Text.Prettyprint          ( (<+>) )
import           Data.Text                      ( Text )

import qualified Data.Vector.Utils             as V
import           Copar.RefinementInterface
import           Copar.FunctorDescription
import qualified Copar.Parser.Lexer               as L
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Printable
import           Copar.Functors.SomeFunctor
import           Copar.Coalgebra.Parser
import           Data.Float.Utils               ( MaxDouble(..)
                                                , MinDouble(..)
                                                )
import           Copar.Parser.Types
import           Data.SumBag (SumBag)
import qualified Data.SumBag as SumBag
import           Data.Bits.Monoid

import           Data.Proxy


newtype SlowMonoidValued m a = SlowMonoidValued a

-- TODO Remove Printable in favor of FunctorDescription
instance Printable (SlowMonoidValued (Max Int)) where
  precedence _ = 250

  printMe cfg (SlowMonoidValued x) = withName cfg "Max-valued" <> "(Z, max)^" <> x

instance Printable (SlowMonoidValued (Min Int)) where
  precedence _ = 240

  printMe cfg (SlowMonoidValued x) = withName cfg "Min-valued" <> "(Z, min)^" <> x

instance Printable (SlowMonoidValued MaxDouble) where
  precedence _ = 230

  printMe cfg (SlowMonoidValued x) =  withName cfg "Max-valued" <> "(R, max)^" <> x

instance Printable (SlowMonoidValued MinDouble) where
  precedence _ = 220

  printMe cfg (SlowMonoidValued x) =  withName cfg "Min-valued" <> "(R, min)^" <> x

instance Printable (SlowMonoidValued (BitAnd Word64)) where
  precedence _ = 210

  printMe cfg (SlowMonoidValued x) =  withName cfg "BitAnd-valued" <> "(Word, and)^" <> x

instance Printable (SlowMonoidValued (BitOr Word64)) where
  precedence _ = 200

  printMe cfg (SlowMonoidValued x) =  withName cfg "BitOr-valued" <> "(Word, or)^" <> x

instance Eq1 (SlowMonoidValued m) where
  liftEq f (SlowMonoidValued a1) (SlowMonoidValued a2) = f a1 a2

instance Show1 (SlowMonoidValued m) where
  liftShowsPrec f _ i (SlowMonoidValued a) = f i a

deriving instance Show (SlowMonoidValued m ())
deriving instance Functor (SlowMonoidValued m)
deriving instance Foldable (SlowMonoidValued m)
deriving instance Traversable (SlowMonoidValued m)

class (Suitable (SlowMonoidValued m)) => MonoidValuedDescription m where
  mvName :: Text
  mvDescription :: Text
  mvSet :: [Text] -- often [ascii, unicode]
  mvOperation :: Text
  mvTerm :: Text
  mvPrecedence :: Int

instance MonoidValuedDescription m => FunctorDescription (SlowMonoidValued m) where
  name = mvName @m <> "-valued"
  syntaxExample = fold (intersperse " | " (map syntax (mvSet @m)))
    where syntax x = functorSyntax x (mvOperation @m)
  description = Just (makeMVHelp (Proxy :: Proxy m))
  precedence = 200 + mvPrecedence @m
  functorExprParser =
    prefix
      -- We need this try here, so that parenthesis can still be parsed as
      -- normal if they don't contain exactly (Z, max)
      (  try
          (L.parens
            ((choice (map L.symbol (mvSet @m)))
            >> L.comma
            >> L.symbol (mvOperation @m)
            )
          )
        >> L.symbol "^"
        >> pure SlowMonoidValued
      )

functorSyntax :: Text -> Text -> Text
functorSyntax s o = "(" <> s <> ", " <> o <> ")^X"

makeMVHelp :: MonoidValuedDescription m => Proxy m -> Doc.Doc Doc.AnsiStyle
makeMVHelp (Proxy :: Proxy m) =
  Doc.reflow ("Weighted systems with " <> mvDescription @m)
  <> Doc.line <> Doc.line
  <> Doc.reflow "This is similar to all the group valued functors (int, real, \
                \etc) but isn't actually a group as it lacks an inverse. Thus \
                \the refinement interface implementation for this functor is\
                \asymptotically slower than for the others."
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow (fold (intersperse " | " (map syntax (mvSet @m))))
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow ("'{' X ':' " <> mvTerm @m <> ", ... '}'")
  where syntax x = functorSyntax x (mvOperation @m)

-- | The @(ℤ, max)^X@ functor
type MaxIntValued = SlowMonoidValued (Max Int)
instance MonoidValuedDescription (Max Int) where
  mvName = "Max"
  mvSet = ["Z", "ℤ"]
  mvOperation = "max"
  mvDescription = "the monoid (Z, max)"
  mvPrecedence = 50
  mvTerm = "int"

-- | The @(ℤ, min)^X@ functor
type MinIntValued = SlowMonoidValued (Min Int)
instance MonoidValuedDescription (Min Int) where
  mvName = "Min"
  mvSet = ["Z", "ℤ"]
  mvOperation = "min"
  mvDescription = "the monoid (Z, min)"
  mvPrecedence = 40
  mvTerm = "int"

-- | The @(ℝ, max)^X@ functor
type MaxDoubleValued = SlowMonoidValued (MaxDouble)
instance MonoidValuedDescription MaxDouble where
  mvName = "Max"
  mvSet = ["R", "ℝ"]
  mvOperation = "max"
  mvDescription = "the monoid (R, max)"
  mvPrecedence = 30
  mvTerm = "real"

-- | The @(ℝ, min)^X@ functor
type MinDoubleValued = SlowMonoidValued (MinDouble)
instance MonoidValuedDescription MinDouble where
  mvName = "Min"
  mvSet = ["R", "ℝ"]
  mvOperation = "min"
  mvDescription = "the monoid (R, min)"
  mvPrecedence = 20
  mvTerm = "real"

-- | The @(Word, and)^X@ functor
type BitAndValued = SlowMonoidValued (BitAnd Word64)
instance MonoidValuedDescription (BitAnd Word64) where
  mvName = "BitAnd"
  mvSet = ["Word"]
  mvOperation = "and"
  mvDescription = "bitvectors and bitwise 'and' as monoid weight"
  mvPrecedence = 10
  mvTerm = "0xCAFE"

-- | The @(Word, or)^X@ functor
type BitOrValued = SlowMonoidValued (BitOr Word64)
instance MonoidValuedDescription (BitOr Word64) where
  mvName = "BitOr"
  mvSet = ["Word"]
  mvOperation = "or"
  mvDescription = "bitvectors and bitwise 'or' as monoid weight"
  mvPrecedence = 0
  mvTerm = "0xCAFE"


type instance Label (SlowMonoidValued m) = m
type instance Weight (SlowMonoidValued m) = (m, SumBag m)
type instance F1 (SlowMonoidValued m) = m
type instance F3 (SlowMonoidValued m) = (m, m, m)

instance (Monoid m, Ord m) => RefinementInterface (SlowMonoidValued m) where
  init
    :: F1 (SlowMonoidValued m)
    -> [Label (SlowMonoidValued m)]
    -> Weight (SlowMonoidValued m)
  init _ labels =
    (mempty, foldl' (flip SumBag.insert) SumBag.empty labels)

  update
    :: [Label (SlowMonoidValued m)]
    -> Weight (SlowMonoidValued m)
    -> ( Weight (SlowMonoidValued m)
       , F3 (SlowMonoidValued m)
       , Weight (SlowMonoidValued m)
       )
  update labels (sumRest, counts) =
    let toS  = foldl' (flip SumBag.insert) SumBag.empty labels
        toCWithoutS = foldl' (flip SumBag.delete) counts labels
        sumS = fold toS
        sumCWithoutS = fold toCWithoutS
        f3   = (sumRest, sumCWithoutS, sumS)
        w1   = (sumRest <> sumCWithoutS, toS)
        w2   = (sumRest <> sumS, toCWithoutS)
    in  (w1, f3, w2)


instance ParseMorphism (SlowMonoidValued (Max Int)) where
  parseMorphismPoint = parseMorphismPointHelper (Max <$> (L.signed L.decimal))

instance ParseMorphism (SlowMonoidValued (Min Int)) where
  parseMorphismPoint = parseMorphismPointHelper (Min <$> (L.signed L.decimal))

instance ParseMorphism (SlowMonoidValued MaxDouble) where
  parseMorphismPoint = parseMorphismPointHelper (MaxDouble <$> L.signed L.float)

instance ParseMorphism (SlowMonoidValued MinDouble) where
  parseMorphismPoint = parseMorphismPointHelper (MinDouble <$> L.signed L.float)

instance ParseMorphism (SlowMonoidValued (BitAnd Word64)) where
  parseMorphismPoint = parseMorphismPointHelper L.hex

instance ParseMorphism (SlowMonoidValued (BitOr Word64)) where
  parseMorphismPoint = parseMorphismPointHelper L.hex


parseMorphismPointHelper :: (HasSanityChecks m, MonadParser m, Ord x, Monoid w) => m w -> SlowMonoidValued w (m x) -> m (w, V.Vector (x, w))
parseMorphismPointHelper weightParser (SlowMonoidValued inner) = do
  sanity <- not <$> noSanityChecks

  !successors <- case sanity of
    True -> do
      succs <- V.sortOn fst . V.fromList <$> L.braces (edge `sepBy` L.comma)

      when (V.hasDuplicates (fmap fst succs)) $
        fail "monoid valued: Duplicate edges"

      return succs

    False -> V.fromList <$> L.braces (edge `sepBy` L.comma)

  let !f1 = fold (V.map snd successors)
  return (f1, successors)
  where edge = (,) <$> inner <*> (L.colon *> weightParser)
