{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE PartialTypeSignatures #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE StrictData #-}
{-# LANGUAGE BangPatterns #-}

-- | Polynomial functor with co-products, products, exponentials and constants
module Copar.Functors.Polynomial
  ( printPolynomial
    -- * Types exported for easier testing
  , Polynomial(..)
  , Sum(..)
  , Product(..)
  , Factor(..)
  , ConstSet(..)
  , Exponent(..)
  , PolyF1(..)
  ) where

import           Control.Monad
import           Data.Bifunctor
import           Data.Foldable (fold)
import           Data.List.NonEmpty (NonEmpty(..))
import qualified Data.List.NonEmpty as NonEmpty
import           Data.Traversable
import           Data.Word (Word8)
import           GHC.Generics (Generic)
import           Data.Functor.Classes
import           Control.Monad.ST
import           Data.Either                    ( lefts
                                                , rights
                                                )

import qualified Data.Vector.Utils as V
import           Data.Text (Text)
import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Vector.Unboxed as VU
import qualified Data.Vector.Unboxed.Mutable as VUM
import           Text.Megaparsec
import           Data.Eq.Deriving (deriveEq1)
import           Text.Show.Deriving (deriveShow1)
import           Lens.Micro
import           Control.DeepSeq
import           Data.Text.Lazy.Builder as Build
import           Data.Text.Lazy.Builder.Int as Build
import qualified Data.Text.Prettyprint     as Doc
import           Data.Text.Prettyprint     ((<+>))

import           Copar.Coalgebra.Parser
import qualified Copar.Parser.Lexer as L
import           Copar.Parser.Types
import           Copar.RefinementInterface
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Printable
import           Copar.FunctorDescription


newtype Polynomial a = Polynomial (Sum a)
  deriving (Functor, Foldable, Traversable)

newtype Sum a = Sum (NonEmpty (Product a))
  deriving (Functor, Foldable, Traversable)

newtype Product a = Product (NonEmpty (Factor a))
  deriving (Functor, Foldable, Traversable)

data ConstSet
  = IntSet
  | NatSet
  | FiniteNatSet Int
  | ExplicitSet (Vector Text)
  deriving (Show, Eq)

data Exponent
  = FiniteNatExp Int
  | ExplicitExp (Vector Text)
  deriving (Show, Eq)

data Factor a
  = Const ConstSet
  | Identity a
  | Exponential a Exponent
  deriving (Functor, Foldable, Traversable)

$(deriveEq1 ''Factor)
$(deriveEq1 ''Product)
$(deriveEq1 ''Sum)
$(deriveEq1 ''Polynomial)

$(deriveShow1 ''Factor)
$(deriveShow1 ''Polynomial)

instance Printable Polynomial where
  precedence _ = 10

  printMe cfg poly = withName cfg "Polynomial" <> printPolynomial cfg poly
    

printPolynomial :: PrintConfig -> Polynomial Build.Builder -> Build.Builder
printPolynomial cfg (Polynomial (Sum (prods))) = fold (NonEmpty.intersperse " + " (fmap (printProduct cfg) prods))

printProduct :: PrintConfig -> Product Build.Builder -> Build.Builder
printProduct cfg (Product (factors)) = fold (NonEmpty.intersperse "x" (fmap (printFactor cfg) factors))

printFactor :: PrintConfig -> Factor Build.Builder -> Build.Builder
printFactor _ (Const IntSet) = "Z"
printFactor _ (Const NatSet) = "N"
printFactor _ (Const (FiniteNatSet v)) = Build.decimal v
printFactor _ (Const (ExplicitSet elems)) =
  (V.foldl
    (\s e -> s <> ", " <> Build.fromText e)
    ("{" <> Build.fromText (V.head elems)) (V.tail elems)
  ) <> "}"
printFactor _ (Identity a) = a
printFactor _ (Exponential a exp) = a <> "^" <> printExp exp

printExp :: Exponent -> Build.Builder
printExp (FiniteNatExp v) = Build.decimal v
printExp (ExplicitExp elems) =
  (V.foldl (\s e -> s <> ", " <> Build.fromText e) ("{" <> Build.fromText (V.head elems)) (V.tail elems)) <> "}"



-- | Custom Show1 instance to show NonEmpty list like a normal list for easier
-- reading.
instance Show1 Product where
  liftShowsPrec shows' showsList' prec (Product factors) = showsUnaryWith
    (const (liftShowList shows' showsList'))
    "Product"
    prec
    (NonEmpty.toList factors)


-- | Custom Show1 instance to show NonEmpty list like a normal list for easier
-- reading.
instance Show1 Sum where
  liftShowsPrec shows' showsList' prec (Sum summands) = showsUnaryWith
    (const (liftShowList shows' showsList'))
    "Sum"
    prec
    (NonEmpty.toList summands)

instance FunctorDescription Polynomial where
  name = "Polynomial"
  syntaxExample = "2xX + X^3 | {a, b}xX + X^{c, d, e}"
  description = Just polynomialHelp
  precedence = 10
  functorExprParser = polynomialp

polynomialHelp :: Doc.Doc Doc.AnsiStyle
polynomialHelp =
  Doc.reflow ("This is a catch-all functor for polynomial expressions, \
              \consisting of sums, products and constants. The informal \
              \semantics for each of those is:")
  <>  Doc.line <>  Doc.line
  <> Doc.reflow "Sums represent choice. A state for the functor A+B can be \
                \either A or B."
  <>  Doc.line <>  Doc.line
  <> Doc.reflow "Products combine. A state for AxB has successors in A and \
                \in B."
  <>  Doc.line <>  Doc.line
  <> Doc.reflow "Constants don't have any successors and exist solely for \
                \their F1 values. Constants can be used in a product to \
                \represent state labels or initial partitions like for AxX."
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.align ("P ::= S_1 '+' S_2 '+' ... | S" <> Doc.line
                <> "S ::= F_1 'x' F_2 'x' ... | F" <> Doc.line
                <> "F ::= X | X '^' A | C")
  <> Doc.line
  <> Doc.reflow "where A and C are constants and can be written as set \
                \'{' a, b, ...'}' or a fixed set of integers like 4, meaning \
                \{0,1,2,3}. Additionally, C can also be N for all integers."
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.align ("P ::= 'inj' i S" <> Doc.line
                <> "S ::= '(' F_1 ',' F_2 ',' ... ')'" <> Doc.line
                <> "F ::= C | X | '{' X ':' C ',' ... '}'")
  <> Doc.line
  <> Doc.reflow "where i is the index of the summand and C a member of the \
                \specified set."
  <> Doc.line
  <> Doc.reflow "The syntax for one-element sums and products can be omitted, \
                \i.e. 'inj 0 p' is the same as just 'p' and '(f)' is the same \
                \as just 'f'."


polynomialp :: FunctorParser Polynomial
polynomialp = FunctorParser $ \inner -> do
  parseSumExpr inner >>= \case
    Sum (Product (Identity a :| []) :| []) -> return (Left a)
    other -> return (Right (Polynomial other))

parseSumExpr :: MonadParser m => m a -> m (Sum a)
parseSumExpr inner = do
  left <- parseProductExpr inner
  rest <- many (L.symbol "+" *> parseProductExpr inner)
  return $ Sum (left :| rest)

parseProductExpr :: MonadParser m => m a -> m (Product a)
parseProductExpr inner = do
  left <- parseFactorExpr inner
  let multSign = L.symbol "x" <|> L.symbol "×"
  rest <- many (multSign *> parseFactorExpr inner)
  return $ Product (left :| rest)

parseFactorExpr :: MonadParser m => m a -> m (Factor a)
parseFactorExpr inner = (Const <$> parseConstExpr) <|> (parseIdOrExp inner)

parseConstExpr :: MonadParser m => m ConstSet
parseConstExpr =
  -- We have use `try` here to prevent us from erroneously parsing the `Z` of `Z^X` as an `IntSet`.
  try (
    (((L.symbol "Z" <|> L.symbol "ℤ") >> return IntSet) <|>
    ((L.symbol "N" <|> L.symbol "ℕ") >> return NatSet) <|>
    (FiniteNatSet <$> L.decimal) <|>
    ((ExplicitSet . V.fromList) <$> (L.braces (L.name `sepBy` L.comma)))) <* notFollowedBy  (L.symbol "^")
  ) 

parseIdOrExp :: MonadParser m => m a -> m (Factor a)
parseIdOrExp inner = do
  x <- inner
  (Exponential x <$> (try (L.symbol "^") *>
                      (parseExplicitExp <|> parseFiniteNatExp)))
    <|> (return (Identity x))

parseExplicitExp :: MonadParser m => m Exponent
parseExplicitExp = L.braces $ do
  names <- (V.sort . V.fromList) <$> L.name `sepBy` L.comma

  when (V.hasDuplicates names) $
    fail "exponential: domain must be uniquely defined"

  return (ExplicitExp names)

parseFiniteNatExp :: MonadParser m => m Exponent
parseFiniteNatExp = FiniteNatExp <$> L.decimal


-- | Morally the type '3'.
--
-- We use Word8 for efficiency, so this should only ever have the values 0, 1
-- and 2 with the following meanings:
--
-- 0: Edge goes to rest
-- 1: Edge goes to compound block
-- 2: Edge goes to subblock
type Three = Word8

toSub :: Three
toSub = 2

boolToThree :: Bool -> Three
boolToThree False = 0
boolToThree True = 1

-- | F1 for Polynomial
--
-- TODO: Use unboxed vector for constants
data PolyF1 = PolyF1
  { polyF1Summand :: {-# UNPACK #-} Int
  -- ^ Index into sum
  , polyF1Variables :: {-# UNPACK #-} Int
  -- ^ Number of variable factors
  , polyF1Constants :: {-# UNPACK #-} (VU.Vector Int)
  -- ^ Values of constant factors
  }
  deriving (Eq,Show,Ord,NFData,Generic)

type instance F1 Polynomial = PolyF1


-- | Index of this edge into the product
--
-- Note that this ignores constant factors, so an edge to the X in (2, 5, X)
-- would have 0 as Label, not 2. Also, the top-level co-product doesn't appear
-- in the label at all. It already appears in "F1".
type instance Label Polynomial = Int


-- | Defined as F2
--
-- For each variable in the product, this is True if the corresponding successor
-- state belongs to the subblock and False otherwise.
type instance Weight Polynomial = VU.Vector Bool


-- | Defined as F3
--
-- Same as for Weight, but now with the three states described in the
-- documentation for Three.
type instance F3 Polynomial = VU.Vector Three


instance ParseMorphism Polynomial where
  parseMorphismPoint (Polynomial expr) = parseSum1 expr

----------- Coproducts parser

-- | Parse either a single product or an injection into the coproduct, depending
-- on the number of co-factors.
parseSum1
  -- :: (MonadParser m, HasSanityChecks m)
  -- => Sum (m a)
  -- -> m (F1 Polynomial, Vector (a, Label Polynomial))
  :: Ord a
  => Sum (MorphParser l f1 a)
  -> MorphParser
       l
       f1
       (F1 Polynomial, Vector (a, Label Polynomial))
parseSum1 sum@(Sum (product :| [])) = do
  -- only a single summand => parse product directly

  -- We first try to parse 'inj i' and delegate to the product parser only if
  -- that fails.
  -- This avoids strange situations where a constant calle 'inj' exists and the
  -- input starts with inj.
  (try parseSumPrefix >>= parseSum sum)
    <|> (first (uncurry (PolyF1 0)) <$> parseProduct1 product)
parseSum1 other =
  parseSumPrefix
    >>= parseSum other -- otherwise, require 'inj'
    <?> "coproduct injection"

-- | parses @inj i@ where @i@ is a decimal integer
parseSumPrefix :: MonadParser m => m Int
parseSumPrefix = L.symbol "inj" *> L.decimal

-- | Parse an injection into the coproduct with the syntax 'inj i _'
parseSum
  :: Ord a
  => Sum (MorphParser l f1 a)
  -> Int
  -> MorphParser l f1 (F1 Polynomial, Vector (a, Label Polynomial))
parseSum (Sum summands) i = do
  sanity <- not <$> noSanityChecks
  when (sanity && (i < 0 || i >= length summands))
    $ fail ("polynomial: injection " ++ show i ++ " is out of bounds")

  (f1', successors) <- parseProduct1 (summands NonEmpty.!! i)

  let !f1 = PolyF1 i (fst f1') (snd f1')

  return (f1, successors)

----------- Products parser

-- | Parse either a single factor without parens or a tuple.
parseProduct1
  :: Ord a
  => Product (MorphParser l f1 a)
  -> MorphParser l f1 ((Int, VU.Vector Int), Vector (a, Label Polynomial))
parseProduct1 product@(Product (factor :| [])) =
  let mkProduct = either (\i -> ((0, VU.singleton i), V.empty))
                         (\v -> ((length v, VU.empty), v))
  in  (mkProduct <$> parseFactor factor) <|> parseProduct product
parseProduct1 other = parseProduct other


parseProduct
  :: Ord a
  => Product (MorphParser l f1 a)
  -> MorphParser l f1 ((Int, VU.Vector Int), Vector (a, Label Polynomial))
parseProduct (Product l@(f :| fs)) =
  label ("a product of " ++ show (length l) ++ " element(s)") $ L.parens $ do
    factors <-
      (:)
      <$> parseFactor f
      <*> (traverse (\x -> L.comma *> parseFactor x) fs)

    let
      !constants                 = VU.fromList (lefts factors)
      !labels                    = rights factors
      (!numFactors, !successors) = mapAccumL
        (\cur f -> (cur + length f, f & each . _2 %~ (+ cur)))
        0
        labels

    return ((numFactors, constants), V.concat successors)

----------- Factor parser

parseFactor
  :: Ord a
  => Factor (MorphParser l f1 a)
  -> MorphParser l f1 (Either Int (Vector (a, Int)))
parseFactor (Const (ExplicitSet names)) = do
  !f1 <- Left <$> someName names
  return f1 -- const has no successors
parseFactor (Const IntSet) = do
  x <- L.signed L.decimal <?> "integer"
  return (Left x)
parseFactor (Const NatSet) = do
  x <- L.decimal <?> "natural number"
  return (Left x)
parseFactor (Const (FiniteNatSet n)) = do
  x      <- L.decimal <?> ("natural number small than " ++ show n)
  sanity <- not <$> noSanityChecks
  when (sanity && x >= n) $ fail
    (  "out of range constant: "
    ++ show x
    ++ "(must be between 0 and "
    ++ show n
    ++ ")"
    )
  return (Left x)
parseFactor (Identity inner) = do
  successor <- inner
  return (Right (V.singleton (successor, 0)))
parseFactor (Exponential inner exp) = L.braces $ do
  sanity <- not <$> noSanityChecks

  successors <-
    (if sanity then V.sortOn snd else id)
    .       V.fromList
    <$>     (flip (,) <$> parseExpValue exp <*> (L.colon *> inner))
    `sepBy` L.comma

  when (sanity && allExpValues exp /= (V.map snd successors))
    $ fail ("exponential: map must be well-defined on " ++ showExp exp)

  return (Right successors)

parseExpValue :: (MonadParser m, HasSanityChecks m) => Exponent -> m Int
parseExpValue (ExplicitExp names) = someName names
parseExpValue (FiniteNatExp n) = do
  x <- L.decimal
  sanity <- not <$> noSanityChecks
  when (sanity && x >= n) $ -- L.decimal returns only positive ints
    fail ("Value " ++ show x ++
          "is out of bounds. (must be between 0 and " ++ show n ++ ")")
  return x

allExpValues :: Exponent -> Vector Int
allExpValues (ExplicitExp names) = V.enumFromN 0 (length names)
allExpValues (FiniteNatExp n) = V.enumFromN 0 n

showExp :: Exponent -> String
showExp (ExplicitExp names) = show names
showExp (FiniteNatExp n) = "{0.." ++ show n ++ "}"

someName :: MonadParser m => Vector Text -> m Int
someName v = do
  name <- try L.name
  -- TODO Better error
  maybe empty return (V.elemIndex name v)

instance RefinementInterface Polynomial where
  init :: F1 Polynomial -> [Label Polynomial] -> Weight Polynomial
  init f1 _ = VU.replicate (polyF1Variables f1) True

  update ::
       [Label Polynomial]
    -> Weight Polynomial
    -> (Weight Polynomial, F3 Polynomial, Weight Polynomial)
  update labels toC = runST $ do
    toS <- VUM.new (VU.length toC) -- toS initialized with False
    toCNoS <- VU.thaw toC -- toCNoS initialized with the content of toC
    -- f3 also initialized with toC, but converted to Three. We can use
    -- unsafeThaw here since VU.map yields a fresh copy.
    f3 <- VU.unsafeThaw (VU.map boolToThree toC)

    forM_ labels $ \i -> do
      VUM.write toS i True
      VUM.write toCNoS i False
      VUM.write f3 i toSub

    (,,)
      <$> VU.unsafeFreeze toS
      <*> VU.unsafeFreeze f3
      <*> VU.unsafeFreeze toCNoS
