{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE BangPatterns #-}

module Copar.Functors.Powerset
  ( Powerset(..)
  -- * For testing
  , packWeight
  , unpackWeight
  , PowerF3(..)
  , mkPowerF3
  ) where

import           Control.Monad (when)
import           Data.Word (Word8)
import           Data.Bits

import           Text.Megaparsec
import qualified Data.Vector as V
import qualified Data.Vector.Utils as V
import           Data.Eq.Deriving (deriveEq1)
import           Text.Show.Deriving (deriveShow1)
import qualified Data.Text.Prettyprint as Doc
import           Data.Text.Prettyprint ((<+>))

import           Copar.RefinementInterface
import qualified Copar.Parser.Lexer as L
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Printable
import           Copar.Coalgebra.Parser
import           Copar.FunctorDescription

newtype Powerset a = Powerset a
  deriving (Show,Functor,Foldable,Traversable)

$(deriveShow1 ''Powerset)
$(deriveEq1 ''Powerset)

instance Printable Powerset where
  precedence _ = 50

  printMe cfg (Powerset x) = withName cfg "Powerset" <> "P" <> x

instance FunctorDescription Powerset where
  name = "Powerset"
  syntaxExample = "PX | ƤX"
  description = Just powersetHelp
  precedence = 50
  functorExprParser =
    prefix ((L.symbol "P" <|> L.symbol "Ƥ") >> pure Powerset)

powersetHelp :: Doc.Doc Doc.AnsiStyle
powersetHelp =
  Doc.reflow "Coalgebras for this functor correspond to transition systems \
             \where each state has a possibly empty set of successor states. \
             \Combined with the polynomial functor to P(AxX), we get labelled \
             \transition systems."
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Functor syntax:"
  <+> Doc.reflow "P(F) | Ƥ(F)"
  <> Doc.line <> Doc.line
  <> Doc.annotate Doc.bold "Coalgebra syntax:"
  <+> Doc.reflow "'{' x ',' ... '}'"

data PowerWeight = PowerWeight {-# UNPACK #-} !Int
  deriving (Show,Eq)

newtype PowerF3 = PowerF3 Word8
  deriving (Show, Ord, Eq)

mkPowerF3 :: Bool -> Bool -> Bool -> PowerF3
mkPowerF3 !a !b !c = PowerF3 $ set 0 a .|. set 1 b .|. set 2 c
  where set !i True = bit i
        set !_ False = 0
{-# INLINE mkPowerF3 #-}

-- | No edge labels
type instance Label Powerset = ()
-- | Tuple of (Edges to X\C, |edgesToC|), see packWeight / unpackWeight
type instance Weight Powerset = Int
-- | Does this state have at least one successor?
type instance F1 Powerset = Bool
-- | Tuple of:
--    - do we have edges to the rest?
--    - do we have edges to C\S?
--    - do we have edges to S?
type instance F3 Powerset = PowerF3

instance ParseMorphism Powerset where
  parseMorphismPoint (Powerset inner) = do
    successors' <- L.braces (inner `sepBy` L.comma)
    sanity <- not <$> noSanityChecks

    !successors <- case sanity of
      True -> do
        let succs = V.sort (V.fromList successors')
        when (V.hasDuplicates succs) $
          fail "powerset: Duplicate edges"
        return succs

      False -> return (V.fromList successors')

    let f1 = not (null successors)

    return (f1, fmap (,()) successors)

packWeight :: Bool -> Int -> Int
packWeight True edgesToC = (edgesToC `shift` 1) `setBit` 0
packWeight False edgesToC = edgesToC `shift` 1

unpackWeight :: Int -> (Bool,Int)
unpackWeight weight = (testBit weight 0, weight `shift` (-1))

instance RefinementInterface Powerset where
  init :: F1 Powerset -> [Label Powerset] -> Weight Powerset
  init _ = packWeight False . length

  update :: [Label Powerset] -> Weight Powerset
         -> (Weight Powerset, F3 Powerset, Weight Powerset)
  update labels weight =
    let
      (!toRest, !toC) = unpackWeight weight
      !toS = length labels
      !toCwithoutS = toC - toS
      !weightToS = packWeight (toRest || toCwithoutS > 0) toS
      !f3 = mkPowerF3 toRest (toCwithoutS > 0) (toS > 0)
      !weightToCwithoutS = packWeight (toRest || toS > 0) toCwithoutS
    in
      (weightToS, f3, weightToCwithoutS)
