{-# LANGUAGE CPP #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ConstraintKinds #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE Strict #-}

module Copar.Functors.SomeFunctor
  ( SomeFunctor(..)
  , Suitable
  , transformInner
  , SomeWeight(..)
  , SomeLabel
  , SomeF1
  ) where

import           Prelude hiding (init)

import           Type.Reflection
#ifdef RELEASE
import           GHC.Exts (Any)
import           Unsafe.Coerce
#else
import           Data.Maybe (mapMaybe)
#endif
import           Data.Functor.Classes

import           Control.DeepSeq (NFData(..))
import qualified Data.Vector as V
import qualified Data.Text as T

import           Copar.Coalgebra.Parser.Class
import           Copar.FunctorExpression.Printable
import           Copar.PrettyShow
import           Copar.RefinementInterface

type Suitable f
   = ( RefinementInterface f
     , ParseMorphism f
     , Printable f
     , Eq1 f
     , Functor f
     , Foldable f
     , Typeable f
     , Traversable f
     , Show1 f
     , NFData (F1 f)
     , NFData (Label f)
     , PrettyShow (Label f)
     , PrettyShow (F1 f))


data SomeFunctor a where
  SomeFunctor
    :: Suitable f
    => f a
    -> SomeFunctor a

deriving instance Functor SomeFunctor
deriving instance Foldable SomeFunctor
deriving instance Traversable SomeFunctor

instance Show a => Show (SomeFunctor a) where
  showsPrec = showsPrec1

instance Show1 SomeFunctor where
  liftShowsPrec shows' showsList' pred (SomeFunctor f) =
    liftShowsPrec shows' showsList' pred f

instance Eq a => Eq (SomeFunctor a) where
  (==) = eq1

instance Eq1 SomeFunctor where
  liftEq eq' (SomeFunctor (lhs :: f' a)) (SomeFunctor (rhs :: f'' b)) =
    case eqTypeRep (typeRep @f') (typeRep @f'') of
      Nothing    -> False
      Just HRefl -> liftEq eq' lhs rhs

instance Printable SomeFunctor where
  precedence (SomeFunctor x) = precedence x

  printMe cfg (SomeFunctor x) = printMe cfg x

-- | Apply a natural transformation under the 'SomeFunctor'.
--
-- This can change the functor stored in the 'SomeFunctor'. Transformations with
-- the wrong domain are silently ignored.
transformInner
  :: forall f g a
   . (Typeable f, Typeable g, Suitable g, ParseMorphism g)
  => (f a -> Maybe (g a))
  -> SomeFunctor a
  -> SomeFunctor a
transformInner t (SomeFunctor (x :: f1 a)) =
  case eqTypeRep (typeRep @f) (typeRep @f1) of
    Nothing    -> SomeFunctor x
    Just HRefl -> case t x of
      Nothing  -> SomeFunctor x
      Just new -> SomeFunctor new

-- A note on all this Any stuff:
--
-- It's safe to unsafeCoerce a type to Any and than coerce it back, but what we
-- do here is actually "unsafeCoerce :: [Any] -> [Label]" instead of the more
-- obviously safe "map (unsafeCoerce :: Any -> Label)". According to #haskell,
-- it's "about as safe".


#ifdef RELEASE
newtype SomeLabel = SomeLabel Any

instance PrettyShow SomeLabel where
  prettyShow _ = "[elided]"

instance NFData SomeLabel where
  rnf (SomeLabel inner) = inner `seq` ()
#else
data SomeLabel where
  SomeLabel :: (Suitable f) => TypeRep f -> Label f -> SomeLabel

instance PrettyShow SomeLabel where
  prettyShow (SomeLabel _ inner) = prettyShow inner

instance NFData SomeLabel where
  rnf (SomeLabel !_ !inner) = rnf inner
#endif

instance Show SomeLabel where
  show = T.unpack . prettyShow


data SomeWeight where
  SomeWeight :: Suitable f => TypeRep f -> Weight f -> SomeWeight

data SomeF1 where
  SomeF1 :: (Suitable f) => TypeRep f -> F1 f -> SomeF1

instance Eq SomeF1 where
  (SomeF1 f1 a) == (SomeF1 f2 b) = case eqTypeRep f1 f2 of
    Nothing -> False
    Just HRefl -> a == b

instance Ord SomeF1 where
  compare (SomeF1 f1 a) (SomeF1 f2 b) = case eqTypeRep f1 f2 of
    Nothing -> compare @Int 1 2 -- XXX: should not happen
    Just HRefl -> compare a b

instance PrettyShow SomeF1 where
  prettyShow (SomeF1 _ a) = prettyShow a

instance Show SomeF1 where
  show = T.unpack . prettyShow

instance NFData SomeF1 where
  rnf (SomeF1 !_ !inner) = rnf inner

data SomeF3 where
  SomeF3 :: Suitable f => TypeRep f -> F3 f -> SomeF3

instance Eq SomeF3 where
  (SomeF3 f1 a) == (SomeF3 f2 b) = case eqTypeRep f1 f2 of
    Nothing -> False
    Just HRefl -> a == b

instance Ord SomeF3 where
  compare (SomeF3 f1 a) (SomeF3 f2 b) = case eqTypeRep f1 f2 of
    Nothing -> compare @Int 1 2 -- XXX: should not happen
    Just HRefl -> compare a b

type instance Label SomeFunctor = SomeLabel
type instance Weight SomeFunctor = SomeWeight
type instance F1 SomeFunctor = SomeF1
type instance F3 SomeFunctor = SomeF3

instance RefinementInterface SomeFunctor where
  init (SomeF1 (f :: TypeRep tf) f1) labels =
#ifdef RELEASE
    SomeWeight f (init @tf f1 (unsafeCoerce labels))
#else
    let myLabels = mapMaybe isSameType labels
    in SomeWeight f (init @tf f1 myLabels)

    where
      isSameType :: SomeLabel -> Maybe (Label tf)
      isSameType (SomeLabel f2 l) = case eqTypeRep f f2 of
        Nothing -> Nothing
        Just HRefl -> Just l
#endif

  update labels (SomeWeight (f :: TypeRep tf) w) =
#ifdef RELEASE
    let (a, b, c) = update @tf (unsafeCoerce labels) w
    in (SomeWeight f a, SomeF3 f b, SomeWeight f c)
#else
    let myLabels = mapMaybe isSameType labels
        (a, b, c) = update @tf myLabels w
    in (SomeWeight f a, SomeF3 f b, SomeWeight f c)

    where
      isSameType :: SomeLabel -> Maybe (Label tf)
      isSameType (SomeLabel f2 l) = case eqTypeRep f f2 of
        Nothing -> Nothing
        Just HRefl -> Just l
#endif

instance ParseMorphism SomeFunctor where
  parseMorphismPoint (SomeFunctor (f :: tf (MorphParser l f1 x))) = do
    (f1, succs) <- parseMorphismPoint f
    let fRep = typeRep @tf
        newF1 = SomeF1 fRep f1

#ifdef RELEASE
    let newSuccs = V.map (\(x, y) -> (x, unsafeCoerce y)) succs
#else
    let newSuccs = V.map (\(x, y) -> (x, SomeLabel fRep y)) succs
#endif

    return (newF1, newSuccs)
