{-# LANGUAGE PartialTypeSignatures #-}
{-# LANGUAGE PolyKinds #-}

-- | High-level API for the various parsers.
--
-- This module combines the different parsers (functor, coalgebra) and exposes a
-- high level API to parse the complete input for CoPaR. This is the parser API
-- that should be used in the main function and high-level tests.
module Copar.Parser
  ( -- * IO Parsers
    readFile
  , readStdin
    -- * Pure Parsers
  , parseCoalgebra
  , parseFunctor
    -- * Parser Configuration
  , Config(..)
  , defaultConfig
  , ApplyFunctorTransformations(..)
  , SanityChecks(..)
  ) where

import           Prelude                 hiding ( readFile )
import           Data.Bifunctor
import           Data.Function                 ( on )
import           Data.Proxy
import qualified Data.List.NonEmpty            as E
import qualified Data.List                     as L

import           Data.Text                      ( Text )
import qualified Data.Text.IO                  as T
import           Text.Megaparsec

import           Data.MorphismEncoding
import           Copar.RefinementInterface         (F1, Label)
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Sorts
import           Copar.Coalgebra.Parser
import           Copar.Parser.Types
import qualified Copar.Parser.Lexer               as L
import           Copar.Functors
import           Copar.FunctorExpression.Desorting
import           Copar.FunctorExpression.Type
import           Copar.FunctorDescription
import           Copar.Functors.Polynomial
import           Copar.Functors.SomeFunctor        ( SomeFunctor(SomeFunctor) )
import           Copar.RewriteFunctors

functorExpressionParser
  :: (Traversable f, ParseMorphism f)
  => FunctorExpression f Precedence
  -> [FunctorParser f]
  -> Parser (FunctorExpression f Sort)
functorExpressionParser varReplacement functors = do
  ((annotateSorts . replaceVar) <$> functorsParser functors)
  where
    replaceVar Variable = varReplacement
    replaceVar other    = other


-- | Whether to apply transformations on the parsed functor expression. These
-- transformations can e.g replace certain functors with more efficient ones or
-- optimize the expression in other ways.
data ApplyFunctorTransformations = ApplyTransformations | DontApplyTransformations


coalgebraParser
  :: Either
       (FunctorExpression SomeFunctor Precedence, [FunctorParser SomeFunctor])
       (FunctorExpression SomeFunctor Sort)
  -> ApplyFunctorTransformations
  -> SanityChecks
  -> Parser
       ( SymbolTable
       , Encoding
           (Label (Desorted SomeFunctor))
           (F1 (Desorted SomeFunctor))
       )
coalgebraParser functor transPolicy sanity = do
  L.space *> L.newlines
  f <- case functor of
    Left (varReplacement, functors) ->
      functorExpressionParser varReplacement functors <* L.newlines1
    Right f' -> return f'
  morphismsParser (transformFunctor f) sanity
    <?> "morphism definition"
  where
    transformFunctor = case transPolicy of
      ApplyTransformations     -> applyFunctorRewrites
      DontApplyTransformations -> id


-- TODO: Needs better name
type TheFunctor = Desorted SomeFunctor

-- | Parse functor expression from text.
--
-- The functor expression contains 'SomeFunctor's, which in turn can wrap any of
-- the registered functors.
--
-- It returns either an error message or the resulting (sorted) functor
-- expression.
parseFunctor
  :: String -- ^ Name of the input (filename, etc)
  -> Text -- ^ Input text
  -> Either String (FunctorExpression SomeFunctor Sort)
parseFunctor name input =
  let identity =
        Functor
          (Precedence 0)
          (SomeFunctor
             (Polynomial
                (Sum (E.fromList [Product (E.fromList [Identity Variable])]))))
   in bimap
        errorBundlePretty
        id
        (parse (functorExpressionParser identity functorParsers <* eof) name input)
  where
    functorParsers = map dynFunctorExprParser (reverse (L.sortBy (compare `on` dynPrecedence) registeredFunctors))


-- | Options for coalgebra parsing.
data Config = Config
  { -- | Controls whether to apply functor transformations. See
    -- 'ApplyTransformations' for details.
    functorTransforms :: ApplyFunctorTransformations
    -- | Controls sanity checks during parsing. Disabling them can make the
    -- parser faster but fail horribly if the input is not correct.
  , sanityChecks :: SanityChecks
  -- | The functor for the coalgebra. If this is @Nothing@ (the default), the
  -- functor expression will be read from the first non-comment line of the
  -- input.
  , functor :: Maybe (FunctorExpression SomeFunctor Sort)
  }


-- | Default config: Apply functor transforms, enable sanity checks and read
-- functor from input file.
defaultConfig :: Config
defaultConfig = Config
  { functorTransforms = ApplyTransformations
  , sanityChecks = EnableSanityChecks
  , functor = Nothing
  }


-- | Parse a coalgebra from text.
--
-- If the parser configuration specifies a functor expression, this functor is
-- used to parse the coalgebra. Otherwise, a functor expression is read from the
-- first non-comment line in the input.
--
-- The returned value is either an error message, or as follows:
--
--   - A proxy for the type of the functor. This is always @Desorted
--     SomeFunctor@ but can be used to conveniently pass the proxy to functions
--     that require the type as input.
--   - A symbol table, mapping the names of all states in sort 1 to their
--     numerical id.
--   - The coalgebra encoding.
parseCoalgebra
  :: Config -- ^ Parser configuration
  -> String -- ^ Name of the input (filename, etc)
  -> Text -- ^ Input text
  -> Either
       String
       ( Proxy TheFunctor
       , (SymbolTable, Encoding (Label TheFunctor) (F1 TheFunctor))
       )
parseCoalgebra config name input =
  let identity = Functor
        (Precedence 0)
        (SomeFunctor
          (Polynomial
            (Sum (E.fromList [Product (E.fromList [Identity Variable])]))
          )
        )
      eitherFunctor =
          maybe (Left (identity, functorParsers)) Right (functor config)
  in  bimap
        errorBundlePretty
        (Proxy, )
        (parse
          (coalgebraParser eitherFunctor
                           (functorTransforms config)
                           (sanityChecks config)
          )
          name
          input
        )
  where functorParsers = map dynFunctorExprParser (reverse (L.sortBy (compare `on` dynPrecedence) registeredFunctors))


-- | Read coalgebra from file. See 'parseCoalgebra' for details.
readFile
  :: Config
  -> FilePath
  -> IO
       ( Either
           String
           ( Proxy TheFunctor
           , (SymbolTable, Encoding (Label TheFunctor) (F1 TheFunctor))
           )
       )
readFile config filename = do
  content <- T.readFile filename
  return $ parseCoalgebra config filename content


-- | Read coalgebra from stdin. See 'parseCoalgebra' for details.
readStdin
  :: Config
  -> IO
       ( Either
           String
           ( Proxy TheFunctor
           , (SymbolTable, Encoding (Label TheFunctor) (F1 TheFunctor))
           )
       )
readStdin config = do
  content <- T.getContents
  return $ parseCoalgebra config "(stdin)" content
