{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Copar.Parser.Lexer
  ( newlines1
  , newlines
  , newlinesOrEof
  , space
  , space1
  , symbol
  , lexeme
  , braces
  , parens
  , angles
  , brackets
  , semicolon
  , comma
  , colon
  , dot
  , name
  , decimal
  , float
  , signed
  , adouble
  , complex
  , hex
  , fraction
  ) where

import           Data.Char
import           Control.Monad
import           Data.Complex
import           Control.Applicative
import           Data.Maybe (fromMaybe)
import           Data.Tuple (swap)
import           Data.Ratio

import           Data.Text (Text)
import qualified Data.Text as T
import           Text.Megaparsec
import           Text.Megaparsec.Char hiding (space, space1, newline)
import qualified Text.Megaparsec.Char.Lexer as L

import           Copar.Parser.Types
import           Data.Float.Utils (EqDouble, fromDouble)

newline :: MonadParser m => m ()
newline = (void (takeWhile1P (Just "newline") (=='\n'))
          <|> (skipLineComment <* char '\n')) *> space
{-# INLINE newline #-}

-- | Parses newlines or the end of input
newlinesOrEof :: MonadParser m => m ()
newlinesOrEof = newlines1 <|> eof
{-# INLINE newlinesOrEof #-}

-- | Parses one or more newlines or comments
newlines1 :: MonadParser m => m ()
newlines1 = skipSome newline
{-# INLINE newlines1 #-}

-- | Parses zero or more newlines or comments
newlines :: MonadParser m => m ()
newlines = skipMany newline
{-# INLINE newlines #-}

skipLineComment :: MonadParser m => m ()
skipLineComment = char '#' *> void (takeWhileP Nothing (/= '\n'))
{-# INLINE skipLineComment #-}

-- | Parsers zero or more whitespace characters (except newlines)
space :: MonadParser m => m ()
space = void (takeWhileP (Just "space") simpleIsSpace)
{-# INLINE space #-}

-- | Parsers one or more whitespace characters (except newlines)
space1 :: MonadParser m => m ()
space1 = void (takeWhile1P (Just "space") simpleIsSpace)
{-# INLINE space1 #-}

-- | Variant of 'isSpace' that doesn't handle unicode space characters and also
-- isn't 'True' for @\n@.
simpleIsSpace :: Char -> Bool
simpleIsSpace ' '  = True
simpleIsSpace '\t' = True
simpleIsSpace _    = False
{-# INLINE simpleIsSpace #-}

lexeme :: MonadParser m => m a -> m a
lexeme = L.lexeme space
{-# INLINE lexeme #-}

symbol :: MonadParser m => Text -> m Text
symbol = L.symbol space
{-# INLINE symbol #-}

character :: MonadParser m => Char -> m Char
character = lexeme . char
{-# INLINE character #-}

braces :: MonadParser m => m a -> m a
braces = between (character '{') (character '}')
{-# INLINE braces #-}

parens :: MonadParser m => m a -> m a
parens = between (character '(') (character ')')
{-# INLINE parens #-}

angles :: MonadParser m => m a -> m a
angles = between (character '<') (character '>')
{-# INLINE angles #-}

brackets :: MonadParser m => m a -> m a
brackets = between (character '[') (character ']')
{-# INLINE brackets #-}

semicolon :: MonadParser m => m Char
semicolon = character ';'
{-# INLINE semicolon #-}

comma :: MonadParser m => m Char
comma = character ','
{-# INLINE comma #-}

colon :: MonadParser m => m Char
colon = character ':'
{-# INLINE colon #-}

dot :: MonadParser m => m Char
dot = character '.'
{-# INLINE dot #-}

name :: MonadParser m => m Text
name = lexeme (takeWhile1P Nothing isNameChar <?> "name")
  where
    isNameChar c = c == '_' || isLetter c || isDigit c
{-# INLINE name #-}

decimal :: (MonadParser m) => m Int
decimal = lexeme (mkNum <$> takeWhile1P (Just "digit") isDigit)
{-# INLINE decimal #-}

float :: (MonadParser m) => m Double
float = lexeme (mkFloat <$> intPart <*> optional fracPart)
  where
    intPart = takeWhile1P (Just "digit") isDigit
    fracPart = char '.' *> intPart
    mkFloat :: Text -> Maybe Text -> Double
    mkFloat i f = (fromIntegral (mkNum i)) +
      (maybe 0.0 (\e -> fromIntegral (mkNum e) * (10.0 ^^ (negate (T.length e)))) f)
{-# INLINE float #-}

adouble :: (MonadParser m) => m EqDouble
adouble = fromDouble <$> float
{-# INLINE adouble #-}

mkNum :: Text -> Int
mkNum = T.foldl' step 0
  where
    step a c = a * 10 + digitToInt c
{-# INLINE mkNum #-}

signed :: (MonadParser m, Num a) => m a -> m a
signed p = option id (char '-' *> return negate) <*> p
{-# INLINE signed #-}


complex :: (MonadParser m, Num a) => m a -> m (Complex a)
complex inner =
  try (uncurry (:+) . swap <$> sumOf imag real)
    <|> (uncurry (:+) <$> sumOf real imag)
  where
    real = inner
    imag = inner <* symbol "i"
    mandatorySigned p =
      (symbol "-" *> pure negate <|> symbol "+" *> pure id) <*> p
    optionalNum = fmap (fromMaybe 0) . optional
    sumOf p1 p2 = (,) <$> signed p1 <*> optionalNum (mandatorySigned p2)
{-# INLINE complex #-}

-- | Parse a fraction consisting of either a single integral number or a
-- numerator and divisor separated by @/@
fraction :: (MonadParser m, Integral a) => m a -> m (Ratio a)
fraction inner = (%) <$> inner <*> denom
  where
    denom = ((symbol "/" *> notZero) <|> pure 1)
    notZero = inner >>= \x -> if x == 0 then fail "division by zero" else pure x
{-# INLINE fraction #-}

-- | Parse a hexadecimal number consisting of the prefix "0x" followed by
-- hexadecimal digits (both upper and lower case).
hex :: (MonadParser m, Integral a) => m a
hex = lexeme (string "0x" *> L.hexadecimal)
{-# INLINE hex #-}
