{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE StrictData #-}

-- | A queue for blocks that can be used in the lumping algorithm.
--
-- Each block can only be present once in the queue.
--
-- Has O(1) operations for everything including membership test, which is
-- required by the algorithm mentioned above.
module Data.BlockQueue
  ( BlockQueue
  , empty
  , null
  , enqueue
  , dequeue
  , elem
  , delete
  , clear
  , toList
  ) where

import           Prelude hiding (null, elem)

import           Control.Monad.ST
import qualified Data.Foldable as Foldable

import           Data.Primitive.MutVar
import           Data.Sequence (Seq)
import qualified Data.Sequence as Seq
import qualified Data.Vector.Unboxed.Mutable as UnboxedV
import           Lens.Micro
import           Lens.Micro.TH
import           Lens.Micro.Platform ()
import           Control.Monad.Extra (unlessM)

import           Data.RefinablePartition (Block(..))

-- | A mutable fifo queue specialized for blocks
data BlockQueue s = Q
  { _queue :: MutVar s (Seq Block)
  -- This vector has one entry for each sort
  , _presence :: UnboxedV.MVector s Bool
  }

makeLenses ''BlockQueue

-- | Create an empty queue
--
-- `@empty max@` creates an empty queue where the maximum block id that will be inserted (probably the
-- number of states) is given in `@max@`.
--
-- Runtime: O(total number of possible blocks)
empty :: Int -- ^ The maximum block id that will be added
      -> ST s (BlockQueue s)
empty maxSize = do
  q <- newMutVar Seq.empty
  p <- UnboxedV.replicate maxSize False
  return $ Q q p

-- | Check if this queue is empty
--
-- Runtime: O(1)
null :: BlockQueue s -> ST s Bool
null q = Seq.null <$> readMutVar (q^.queue)

-- | Add an element to the end of the queue if it isn't already present.
--
-- Runtime: O(1)
enqueue :: BlockQueue s -> Block -> ST s ()
enqueue q block = unlessM (elem block q) $ do
  modifyMutVar (q^.queue) $ \s -> s Seq.|> block
  UnboxedV.write (q^.presence) (fromBlock block) True

-- | Read and delete the first element in the queue.
--
-- Returns 'Nothing' and doesn't modify the queue if it was empty.
--
-- Runtime: O(1)
dequeue :: BlockQueue s -> ST s (Maybe Block)
dequeue q = Seq.viewl <$> readMutVar (q^.queue) >>= \case
  Seq.EmptyL -> return Nothing
  (x Seq.:< rest) -> do
    UnboxedV.write (q^.presence) (fromBlock x) False
    writeMutVar (q^.queue) rest
    return (Just x)

-- | Tests whether an element is in the queue.
--
-- Runtime: O(1)
elem :: Block -> BlockQueue s -> ST s Bool
elem (Block block) !q = UnboxedV.read (q^.presence) block

-- | Removes all items from the queue.
--
-- Runtime: O(max size of queue)
clear :: BlockQueue s -> ST s ()
clear q = do
  writeMutVar (q^.queue) Seq.empty
  UnboxedV.set (q^.presence) False

-- | Return a list of currently queued items in the order in which they were
-- added to the queue.
--
-- Runtime: O(n)
toList :: BlockQueue s -> ST s [Block]
toList q = Foldable.toList <$> readMutVar (q^.queue)


-- | Delete the given block from the queue if it is queued or do nothing
-- otherwise.
--
-- The block id must be smaller than the maximum size of the queue.
--
-- Runtime: O(current size of queue)
delete :: Block -> BlockQueue s -> ST s ()
delete (Block b) q = do
  UnboxedV.write (q ^. presence) b False
  modifyMutVar (q ^. queue) deleteFromSeq

  where
    deleteFromSeq s = case Seq.elemIndexL (Block b) s of
      Nothing  -> s
      Just idx -> Seq.deleteAt idx s
