{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- | Various utilities for floating point numbers

module Data.Float.Utils
  ( EqDouble
  , fromDouble
  , MaxDouble(..)
  , MinDouble(..)
  )
where

import           Control.DeepSeq                ( NFData )
import qualified Numeric.IEEE                  as IEEE
import           GHC.Float                      ( double2Float )


-- | A wrapper around double that provides an approximate 'Eq' instance.
newtype EqDouble = EqDouble Double
  deriving newtype ( Read
                   , Num
                   , Ord
                   , NFData
                   , Show
                   , Fractional
                   , RealFrac
                   , Real
                   , Floating
                   , RealFloat
                   )


-- | Wrap a 'Double' up as an 'EqDouble'
fromDouble :: Double -> EqDouble
fromDouble = EqDouble
{-# INLINE fromDouble #-}


-- | We cast the double to float and compare the result exactly. This ensures a
-- transitive relation.
instance Eq EqDouble where
  (EqDouble a) == (EqDouble b) = (double2Float a) == (double2Float b)


-- | A wrapper around double that has a monoid instance with the following
-- semantics:
--
-- * 'mempty' is negative infinity
-- * '<>' is implemented in terms of 'IEEE.maxNum'
newtype MaxDouble = MaxDouble Double
  deriving newtype ( Show
                   , Eq
                   , Ord
                   , Num
                   , Real
                   , Fractional
                   , RealFrac
                   , Floating
                   , RealFloat
                   , NFData
                   )

instance Semigroup MaxDouble where
  (MaxDouble d1) <> (MaxDouble d2) = MaxDouble (IEEE.maxNum d1 d2)

instance Monoid MaxDouble where
  mempty = MaxDouble (negate IEEE.infinity)


-- | A wrapper around double that has a monoid instance with the following
-- semantics:
--
-- * 'mempty' is positive infinity
-- * '<>' is implemented in terms of 'IEEE.minNum'
newtype MinDouble = MinDouble Double
  deriving newtype ( Show
                   , Eq
                   , Ord
                   , Num
                   , Real
                   , Fractional
                   , RealFrac
                   , Floating
                   , RealFloat
                   , NFData
                   )

instance Semigroup MinDouble where
  (MinDouble d1) <> (MinDouble d2) = MinDouble (IEEE.minNum d1 d2)

instance Monoid MinDouble where
  mempty = MinDouble IEEE.infinity
