{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE StrictData #-}

-- | Graph encoding of a morphism.
--
-- This is a graph data structure optimized for:
--
-- - Fast enumeration of all edges
-- - Fast retrieval of the label for a given state
-- - Fast indexing into the list of edges
module Data.MorphismEncoding
  ( -- * Types
    Encoding
  , State
  , Edge(..)
  -- * Construction
  , new
  -- * Query
  , size
  , states
  , structure
  , edges
  , edgeRefs
  , numEdges
  , typeOf
  , EdgeRef(..)
  , graph
  ) where

import           Data.Bifunctor
import           GHC.Generics (Generic)

import           Control.DeepSeq

import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Vector.Unboxed as VU

-- | A state is represented by an integer.
--
-- States in an automaton should always start at zero.
type State = Int

-- | A reference to an edge in the graph.
--
-- This is just a newtype wrapper around the index into the edge vector.
newtype EdgeRef = EdgeRef { fromEdgeRef :: Int }
  deriving (Eq, Ord, Show)

data Edge a = Edge
  { from :: ~State
  , label :: ~a
  , to :: ~State
  }
  deriving (Show, Eq, Functor, Generic, NFData)

-- | A generic graph (automaton) structure with labeled nodes and edges.
--
-- Nodes (states) are always the set @[0..size-1]@.
data Encoding a f1 = Encoding
  { eStructure :: Vector f1
  , eEdgesFrom :: VU.Vector State
  , eEdgesLabel :: Vector a
  , eEdgesTo :: VU.Vector State
  }
  deriving (Show, Eq, Generic, NFData)

instance Bifunctor Encoding where
  first f e = e{ eEdgesLabel = fmap f (eEdgesLabel e) }
  second f e = e{ eStructure = fmap f (eStructure e) }

-- | Construct a new encoding from a vector of state labels and a vector of
-- edges.
--
-- Runtime: O(1)
new
  :: Vector f1 -- ^ State labels.
  -> Vector (Edge a) -- ^ Edges. An edge must only reference states that are
                       -- actually defined in the first vector.
  -> Encoding a f1
new !eStructure !eEdges =
  let eEdgesFrom  = V.convert $ V.map from eEdges
      eEdgesTo    = V.convert $ V.map to eEdges
      eEdgesLabel = V.map label eEdges
  in  Encoding { .. }


-- | The number of states in the encoding
--
-- Runtime: O(1)
size :: Encoding a f1 -> Int
size = length . eStructure
{-# INLINE size #-}

-- | A list of all states in the encoding. This is just @[0..size-1]@.
--
-- Runtime: O(size) (lazy)
states :: Encoding a f1 -> [State]
states !self = [0..length (eStructure self)-1]
{-# INLINE states #-}

-- | A vector of all edges in this encoding.
--
-- Runtime: O(1)
edges :: Encoding a f1 -> Vector (Edge a)
edges !self = V.imap mkEdge (eEdgesLabel self)
  where mkEdge i l = Edge ((eEdgesFrom self) VU.! i) l ((eEdgesTo self) VU.! i)
{-# INLINE edges #-}

edgeRefs :: Encoding a f1 -> [EdgeRef]
edgeRefs !self = map EdgeRef [0..numEdges self-1]
{-# INLINE edgeRefs #-}

-- | The number of edges in the encoding
--
-- Runtime: O(1)
numEdges :: Encoding a f1 -> Int
numEdges = V.length . eEdgesLabel
{-# INLINE numEdges #-}

-- | All state labels.
--
-- Runtime: O(1)
structure :: Encoding a f1 -> Vector f1
structure = eStructure
{-# INLINE structure #-}

-- | Return the label of a given state. Throws an exception if the state is not
-- in @[0..size-1]@.
--
-- Runtime: O(1)
typeOf :: Encoding a f1 -> State -> f1
typeOf !self !x = eStructure self V.! x
{-# INLINE typeOf #-}

-- | Given an edge reference, returns the corresponding edge. Throws an
-- exception if the edge reference is invalid.
--
-- Runtime: O(1)
graph :: Encoding a f1 -> EdgeRef -> Edge a
graph !self (EdgeRef !e) = Edge ((eEdgesFrom self) VU.! e)
                                ((eEdgesLabel self) V.! e)
                                ((eEdgesTo self) VU.! e)
{-# INLINE graph #-}
