{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE AllowAmbiguousTypes #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE DataKinds #-}
{-# OPTIONS_HADDOCK prune #-}

-- | A framework for type indexed open coproducts.
--
-- === Example
--
-- Given a set of type constructors of kind @* -> *@ (e.g @Maybe@ or @[]@),
-- their coproduct is:
--
-- >>> :set -XDataKinds
-- >>> type ListOrMaybe = Union '[Maybe, []]
--
-- The function @inj@ injects a value into the coproduct:
--
-- >>> let u = inj (Just 1) :: ListOrMaybe Int
--
-- And conversely, @prj@ projects a value out of the coproduct.
--
-- >>> prj u :: Maybe (Maybe Int)
-- Just (Just 1)
-- >>> prj u :: Maybe [Int]
-- Nothing
--
-- === References
--
-- The general approach for open data types is taken from "Data Types a la
-- Carte" by Wouter Swierstra, with an efficient implementation that's heavily
-- inspired by Data.OpenUnion in the freer-simple package.
module Data.OpenUnion
  ( Union
  , inj
  , prj
  , Member
  ) where

import Unsafe.Coerce

-- | An open sum type.
--
-- @Union types a@ represents the type indexed coproduct of the elements of the
-- type level list @types@.
--
-- ==== __Example__
--
-- >>> :set -XDataKinds
-- >>> type ListOrMaybe = Union '[Maybe, []]
data Union :: [* -> *] -> * -> * where
  In :: Word -> f a -> Union ls a

-- | Inject a value into the coproduct.
--
-- Returns a "Union" of types @list@ that has @elem a@ as current value. The
-- types in @list@ can be freely chosen by the caller, as long as they contain
-- the type of @elem@.
inj :: forall elem list a. Member elem list => elem a -> Union list a
inj = unsafeInj (indexOf @elem @list)

-- | Project a value out of the coproduct
--
-- Returns @Just x@ if the requested type of @x@ is contained in the coproduct,
-- otherwise returns @Nothing@.
prj :: forall elem list a. Member elem list => Union list a -> Maybe (elem a)
prj = unsafePrj (indexOf @elem @list)

-- | This class encodes the assertion that a given type @elem@ is a member of
-- the type level list @list@.
--
-- It is not necessary (or allowed) to add your own instances to this class, all
-- instances have been defined.
--
-- === Example
--
-- This can be used to write generic functions over open unions:
--
-- >>> :set -XFlexibleContexts
-- >>> :{
--  let extractList :: Member [] list => Union list a -> Maybe [a]
--      extractList = prj
-- :}
class Member (elem :: * -> *) (list :: [* -> *]) where
  indexOf :: Word

instance Member x (x ': xs) where
  indexOf = 0

instance {-# OVERLAPPABLE #-} Member x ys => Member x (y ': ys) where
  indexOf = indexOf @x @ys + 1

unsafeInj :: Word -> f a -> Union ls a
unsafeInj = In

unsafePrj :: Word -> Union ls a -> Maybe (f a)
unsafePrj i1 (In i2 x)
  | i1 == i2 = Just (unsafeCoerce x)
  | otherwise = Nothing
