{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE StrictData #-}
{-# LANGUAGE DeriveAnyClass #-}

-- | Represents a partition of a set of states into blocks.
--
-- States and blocks are both represented by all integers starting from zero.
module Data.Partition
  ( Partition
  -- * Auxiliary types
  , Block(..)
  , State
  -- * Creation
  , fromFunction
  , fromFunctionM
  , fromBlocks
  -- * Attributes
  , blocks
  , numBlocks
  , numStates
  , blockOfState
  , statesOfBlock
  -- * Conversion
  , toBlocks
  -- * Algorithms
  , isRefinementOf
  ) where

import           Data.Foldable (toList)
import           Control.Monad (forM_)

import           Control.DeepSeq
import qualified Data.HashMap.Strict as M
import qualified Data.Set as S
import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Vector.Mutable as VM
import           GHC.Generics

import           Data.Partition.Common (Block(..),State)

-- | A partition of a set of states into blocks.
data Partition = Partition
  { numBlocks :: Int -- ^ Return the number of blocks in this partition
  , stateAssignment :: Vector Block
  } deriving (Eq,Generic,NFData)

instance Show Partition where
  show p = "Partition " ++ show (toBlocks p)

-- | Return all blocks.
--
-- Since blocks are numbered from 0, this is just @[0..numBlocks-1]@.
blocks :: Partition -> [Block]
blocks p = map Block [0..numBlocks p-1]

-- | Return the number of states.
numStates :: Partition -> Int
numStates = length . stateAssignment

-- | Construct a partition from a function that maps states to blocks.
--
-- This function must be well-defined and surjective.
--
-- ==== __Examples__
--
-- A trivial partition where all states are in a single block:
--
-- >>> fromFunction 1 3 (const (Block 0))
-- Partition [[0,1,2]]
--
-- A trivial partition where every states is in its own block:
--
-- >>> fromFunction 3 3 (\s -> Block s)
-- Partition [[0],[1],[2]]
fromFunction :: Int -- ^ Number of blocks
             -> Int -- ^ Number of states
             -> (State -> Block) -- ^ State assignment
             -> Partition
fromFunction nrBlocks states mapping = Partition
  { numBlocks = nrBlocks
  , stateAssignment = V.generate states mapping
  }

-- | Construct a partition from a monadic action that maps states to blocks.
--
-- This action must be well-defined and surjective. See 'fromFunction` for
-- details.
fromFunctionM :: Monad m
              => Int -- ^ Number of blocks
              -> Int -- ^ Number of states
              -> (State -> m Block) -- ^ Monadic state assignment
              -> m Partition
fromFunctionM nrBlocks states mapping = do
  assignment <- V.generateM states mapping
  return Partition{ numBlocks = nrBlocks, stateAssignment = assignment }

-- | Construct a partition from a list of blocks, where each block is
-- represented by its states.
fromBlocks :: [[State]] -> Partition
fromBlocks states =             -- TODO Check validity of input
  let vec = V.create $ do
        mutvec <- VM.new (sum (map length states))
        forM_ (zip states [0..]) $ \(block, blockNum) ->
          forM_ block $ \i ->
            VM.write mutvec i (Block blockNum)
        return mutvec
  in
    Partition { numBlocks = length states, stateAssignment = vec }

-- | Return the partition as list of blocks, where each block is represented by
-- its elements.
toBlocks :: Partition -> [[State]]
toBlocks p =
  let
    blockMap = V.ifoldr' (\s (Block b) m -> M.insertWith (++) b [s] m) M.empty (stateAssignment p)
  in
    M.elems blockMap

-- | The block that the given state is in.
blockOfState :: Partition -> State -> Block
blockOfState part state = stateAssignment part V.! state

-- | All states in a given block.
statesOfBlock :: Partition -> Block -> [State]
statesOfBlock part block =
  toList $ V.elemIndices block (stateAssignment part)

-- | Decides whether the first partition is a refinement of the second.
--
-- That is, iff each block in the first is a subset of a block in the second.
isRefinementOf :: Partition -> Partition -> Bool
isRefinementOf refined original =
  let
    origBlocks = map (S.fromList . statesOfBlock original) (blocks original)
    refinedBlocks = map (S.fromList . statesOfBlock refined) (blocks refined)
  in
    all (\b -> any (S.isSubsetOf b) origBlocks) refinedBlocks
