{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE MagicHash #-}

module Data.RefinablePartition
  ( -- * Types
    RefinablePartition
  , State
  , Block(..)
  -- * Construction
  , make
  , make1
  -- * Accessors
  , numBlocks
  , blockSize
  , blockOfState
  , statesOfBlock
  -- * Marking
  , mark
  , unmark
  , isMarked
  , hasMarked
  , markedStates
  -- * Splitting
  , splitMarked
  , splitBy
  , splitByM
  , groupBy
  -- * Conversion
  , freeze
  , unsafeStatesOfBlock
  ) where

import           Control.Monad (forM_, when, foldM)
import           Control.Monad.ST

import           Data.Primitive.MutVar
import qualified Data.Vector as V
import qualified Data.Vector.Algorithms.Heap as VM
import qualified Data.Vector.Mutable as VM
import qualified Data.Vector.Unboxed.Mutable as VU
import qualified Data.Vector.Unboxed as VUU
import qualified Data.Vector.Unboxed as VU (freeze, unsafeFreeze, Vector)
import           Lens.Micro
import           Lens.Micro.TH

import           Data.Vector.Utils (iforM_)
import qualified Data.Vector.Unboxed.Mutable.Utils as VU
import           Data.Partition.Common (State,Block(..))
import           Data.Partition (Partition)
import qualified Data.Partition as Partition


data States s = States
  { _block :: !(VU.MVector s Block)
  , _location :: !(VU.MVector s Int)
  }
makeLenses ''States


data Blocks s = Blocks
  { _startOffset :: !(VU.MVector s Int)
  , _endOffset :: !(VU.MVector s Int) -- exclusive
  , _unmarkedOffset :: !(VU.MVector s Int)
  }
makeLenses ''Blocks

-- | Refinable partition type.
--
-- Should be thought of as a partition {A₁,..} of a set of 'State' values from
-- zero to n. The `Ai`s are called Blocks.
--
-- This type is by nature mutable and can be mutated with the operations in this
-- module. The `s` type variable is there to support the ST monad.
data RefinablePartition s = Partition
  { _blockCount :: {-# UNPACK #-} !(MutVar s Int)
  , _statesByBlock :: {-# UNPACK #-} !(VU.MVector s State)
  , _states :: {-# UNPACK #-} !(States s)
  , _blocks :: {-# UNPACK #-} !(Blocks s)
  }

makeLenses ''RefinablePartition


-- | Create a mutable refinable partition.
make :: Int -- ^ Number of states n
     -> Int -- ^ Number of initial blocks m
     -> (State -> Block)
     -- ^ Initial partition assigning a block to each state. Must be defined on
     -- {0..n} and only have values in {0..m}.
     -> ST s (RefinablePartition s)
make numStates numBlocks initPart
  | numStates < numBlocks = error "RefinablePartition.make: More blocks than states"
  | numBlocks == 0 = error "RefinablePartition.make: zero blocks"
  | otherwise = do

  statesByBlock <- VU.new numStates
  statesBlock <- VU.new numStates
  statesLocation <- VU.new numStates

  -- we need to reserve space for more blocks, to allow splitting to create them.
  -- There can be at most as many blocks as there are states
  blocksStart <- VU.new numStates
  blocksEnd <- VU.new numStates
  blocksUnmarked <- VU.new numStates

  blockCount <- newMutVar numBlocks

  -- contains a list of states for each block
  blockStates <- VM.replicate numBlocks []

  forM_ [0..numStates-1] $ \i -> do
    let Block block = initPart i
    VM.modify blockStates (i:) block

  blockStates' <- V.unsafeFreeze blockStates

  currentLocation <- newMutVar 0
  iforM_ blockStates' $ \i stateList -> do
    beginningOfBlock <- readMutVar currentLocation

    forM_ stateList $ \s -> do
      stateLocation <- readMutVar currentLocation
      writeMutVar currentLocation (stateLocation+1)
      VU.write statesByBlock stateLocation s
      VU.write statesBlock s (Block i)
      VU.write statesLocation s stateLocation

    endOfBlock <- readMutVar currentLocation

    VU.write blocksStart i beginningOfBlock
    VU.write blocksEnd i endOfBlock
    VU.write blocksUnmarked i beginningOfBlock


  return Partition { _blockCount = blockCount
                   , _statesByBlock = statesByBlock
                   , _states = States statesBlock statesLocation
                   , _blocks = Blocks blocksStart blocksEnd blocksUnmarked
                   }


-- | Create a new mutable refinable partition with one initial block.
--
-- This is a special case of 'make', that fixes the initial block count to 1 is
-- slightly faster because of that.
make1
  :: Int -- ^ Number of states n
  -> ST s (RefinablePartition s)
make1 numStates
  | numStates < 1 = error "RefinablePartition.make1: More blocks than states"
  | otherwise = do

    statesByBlock <- VUU.thaw (VUU.generate numStates id)
    blockCount    <- newMutVar 1

    statesBlock   <- VU.replicate numStates 0
    statesLocation <- VUU.thaw (VUU.generate numStates id)
    -- states        <- VUU.thaw (VUU.generate numStates (StateRepr 0))

    blocksStart <- VU.new numStates
    blocksEnd <- VU.new numStates
    blocksUnmarked <- VU.new numStates

    VU.write blocksStart 0 0
    VU.write blocksEnd 0 numStates
    VU.write blocksUnmarked 0 0

    return Partition { _blockCount    = blockCount
                     , _statesByBlock = statesByBlock
                     , _states        = States statesBlock statesLocation
                     , _blocks        = Blocks blocksStart blocksEnd blocksUnmarked
                     }


-- | Return number of blocks in this partition.
--
-- Runtime: O(1)
numBlocks :: RefinablePartition s -> ST s Int
numBlocks p = readMutVar (p ^. blockCount)
{-# INLINE numBlocks #-}

-- | Return number of states in a given block.
--
-- Runtime: O(1)
blockSize :: RefinablePartition s -> Block -> ST s Int
blockSize !p !b = do
  start <- getBlock p b startOffset
  end <- getBlock p b endOffset
  return (end - start)
{-# INLINE blockSize #-}

-- | Return the block that a given state belongs to.
--
-- Runtime: O(1)
blockOfState :: RefinablePartition s -> State -> ST s Block
blockOfState p s = getState p s block
{-# INLINE blockOfState #-}

-- | Record this state as "marked".
--
-- This is implemented in a way to efficiently split the block in marked and
-- unmarked states.
--
-- Calling 'mark' on an already marked state results in data corruption.
--
-- Runtime: O(1)
mark :: RefinablePartition s -> State -> ST s ()
mark !partition !s = do
  loc <- getState partition s location
  b <- getState partition s block

  unmarked <- getBlock partition b unmarkedOffset

  when (loc > unmarked) $
    swap partition loc unmarked

  setBlock partition b unmarkedOffset (unmarked+1)

  where
    -- swap two indices in statesByBlock array
    swap :: RefinablePartition s -> Int -> Int -> ST s ()
    swap partition a b = do
      setStateAt partition a location b
      setStateAt partition b location a

      VU.unsafeSwap (partition^.statesByBlock) a b
{-# INLINE mark #-}

-- | Revert the marking of a state.
--
-- See 'mark'
--
-- Runtime: O(1)
unmark :: RefinablePartition s -> State -> ST s ()
unmark partition s = do
  loc <- VU.read (partition^.states.location) s
  b <- VU.read (partition^.states.block) s

  unmarked <- getBlock partition b unmarkedOffset

  when (loc <= unmarked) $
    swap partition loc (unmarked-1)

  setBlock partition b unmarkedOffset (unmarked - 1)

  where
    -- swap two indices in statesByBlock array
    swap :: RefinablePartition s -> Int -> Int -> ST s ()
    swap partition a b = do
      setStateAt partition a location b
      setStateAt partition b location a

      VU.unsafeSwap (partition^.statesByBlock) a b

-- | Decide if a block has marked states.
--
-- Runtime: O(1)
hasMarked :: RefinablePartition s -> Block -> ST s Bool
hasMarked !p !b =
  (/=) <$> (getBlock p b startOffset) <*> (getBlock p b unmarkedOffset)
{-# INLINE hasMarked #-}

-- | Decide wether a state is marked.
--
-- Runtime: O(1)
isMarked :: RefinablePartition s -> State -> ST s Bool
isMarked partition s = do
  loc <- getState partition s location
  b <- getState partition s block
  unmarked <- getBlock partition b unmarkedOffset

  return $ loc < unmarked
{-# INLINE isMarked #-}

-- | Return the marked states of a block.
--
-- Runtime O(n) for n == number of states in this block
markedStates :: RefinablePartition s -> Block -> ST s (VU.Vector State)
markedStates partition b = do
  start <- getBlock partition b startOffset
  unmarked <- getBlock partition b unmarkedOffset

  let len = unmarked - start

  VU.freeze (VU.slice start len (partition^.statesByBlock))
{-# INLINE markedStates #-}

-- | Return a vector of all states in a given block.
--
-- Runtime: O(n) for n == number of states in this block
statesOfBlock :: RefinablePartition s -> Block -> ST s (VU.Vector State)
statesOfBlock partition b = do
  start <- getBlock partition b startOffset

  len <- blockSize partition b
  let slice = VU.slice start len (partition^.statesByBlock)
  VU.freeze slice
{-# INLINE statesOfBlock #-}

-- | Return a vector of all states in a given block.
--
-- Runtime: O(n) for n == number of states in this block
unsafeStatesOfBlock :: RefinablePartition s -> Block -> ST s (VU.Vector State)
unsafeStatesOfBlock partition b = do
  start <- getBlock partition b startOffset

  len <- blockSize partition b
  let slice = VU.slice start len (partition^.statesByBlock)
  VU.unsafeFreeze slice
{-# INLINE unsafeStatesOfBlock #-}


-- | Split a block into two new blocks for its marked and unmarked states.
--
-- Returns a tuple of (marked block, unmarked block). If there are no marked or
-- unmarked states, the respective block returned will be Nothing.
--
-- The largest new block will inherit the identity (i.e. block number) of the
-- old block (with bias towards the one with unmarked states).
--
-- Runtime: O(number of marked states in the old block)
splitMarked :: RefinablePartition s -> Block -> ST s (Maybe Block, Maybe Block)
splitMarked partition b = do
  start <- getBlock partition b startOffset
  end <- getBlock partition b endOffset
  unmarked <- getBlock partition b unmarkedOffset

  if start == unmarked then -- nothing marked
    return (Nothing, Just b)
  else if unmarked == end then do -- nothing unmarked
    setBlock partition b unmarkedOffset start
    return (Just b, Nothing)
  else do
    let numMarked = unmarked - start
        numUnmarked = end - unmarked
    if numMarked <= numUnmarked then do
      -- new block for marked states
      new <- newBlock partition start unmarked
      -- let old block begin at at the first unmarked state
      setBlock partition b startOffset unmarked
      return (Just new, Just b)
    else do
      -- new block for unmarked states
      new <- newBlock partition unmarked end
      -- let old block end at at the last marked state
      setBlock partition b endOffset unmarked
      -- and reset marked status for all states
      setBlock partition b unmarkedOffset start
      return (Just b, Just new)


-- | Split a block into two new blocks according to a given predicate.
--
-- Returns a tuple @(a, b)@ where @a@ is the block with all states for which the
-- predicate returns True and 'b' contains the rest of the states.
--
-- The largest block in @[a,b]@ shares the identity (i.e. the block number) with
-- the original block.
--
-- Runtime: O(number of states in the old block)
splitBy :: RefinablePartition s -> Block -> (State -> Bool) -> ST s (Maybe Block, Maybe Block)
splitBy !partition !b !predicate = do
  start <- getBlock partition b startOffset
  end <- getBlock partition b endOffset

  splitPoint <- VU.partition (partition^.statesByBlock) predicate start end

  if splitPoint == start then -- no matching states
    return (Nothing, Just b)
  else if splitPoint == end then -- all states match
    return (Just b, Nothing)
  else do
    -- update location for all states, because split moves them
    updateLocations partition start end

    let beforeSplitNum = splitPoint - start
        afterSplitNum = end - splitPoint

    if beforeSplitNum <= afterSplitNum then do
      setBlock partition b startOffset splitPoint
      setBlock partition b unmarkedOffset splitPoint
      new <- newBlock partition start splitPoint
      return (Just new, Just b)
    else do
      setBlock partition b endOffset splitPoint
      new <- newBlock partition splitPoint end
      return (Just b, Just new)

-- | Monadic version of 'splitBy'
--
-- This is the same as 'splitBy', but accepts a monadic predicate. This
-- predicate can be called any number of times and should not have side effects,
-- but can be used to read from mutable vectors and the like.
--
-- Runtime: O(number of states in the old block)
splitByM :: RefinablePartition s -> Block -> (State -> ST s Bool)
         -> ST s (Maybe Block, Maybe Block)
splitByM !partition !b !predicate = do
  -- FIXME Remove duplication with splitBy
  start <- getBlock partition b startOffset
  end <- getBlock partition b endOffset

  splitPoint <- VU.partitionM (partition^.statesByBlock) predicate start end

  if splitPoint == start then -- no matching states
    return (Nothing, Just b)
  else if splitPoint == end then -- all states match
    return (Just b, Nothing)
  else do
    -- update location for all states, because split moves them
    updateLocations partition start end

    let beforeSplitNum = splitPoint - start
        afterSplitNum = end - splitPoint

    if beforeSplitNum <= afterSplitNum then do
      setBlock partition b startOffset splitPoint
      setBlock partition b unmarkedOffset splitPoint
      new <- newBlock partition start splitPoint
      return (Just new, Just b)
    else do
      setBlock partition b endOffset splitPoint
      new <- newBlock partition splitPoint end
      return (Just b, Just new)

-- | Split a block into new blocks according to some atttribute of its states.
--
-- The result is maximally coarse list of blocks, such that all states in a new
-- block have the same value for the given attribute.
--
-- One of the blocks inherits the identity of the old block.
--
-- Runtime: O(n*log(n)) where n is the number of states in the old block.
groupBy :: RefinablePartition s -> Block -> (State -> State -> Ordering) -> ST s [Block]
groupBy partition b predicate = do
  start <- getBlock partition b startOffset
  end <- getBlock partition b endOffset

  VM.sortBy predicate $
    VU.slice start (end-start) (partition^.statesByBlock)

  updateLocations partition start end

  indices <- VU.groupBy (partition^.statesByBlock) predicate start end

  let splitAt (currentBlock,newBlocks) index = do
        setBlock partition currentBlock unmarkedOffset index
        (Just previousBlock, Just nextBlock) <- splitMarked partition currentBlock
        return (nextBlock, previousBlock:newBlocks)

  (last,blocks) <- foldM splitAt (b, []) indices

  return (last:blocks)

-- | Freeze the current refinable partition into an immutable one.
--
-- This copies the current state of the partition and doesn't modify it.
--
-- Runtime: O(number of states)
freeze :: RefinablePartition s -> ST s Partition
freeze partition = do
  let nrStates = VU.length (partition^.statesByBlock)
  nrBlocks <- numBlocks partition
  Partition.fromFunctionM nrBlocks nrStates (blockOfState partition)
{-# INLINE freeze #-}


getState :: VUU.Unbox a => RefinablePartition s -> State -> Lens' (States s) (VU.MVector s a) -> ST s a
getState !partition !s l = VU.unsafeRead (partition^.states.l) s

setState :: VUU.Unbox a => RefinablePartition s -> State -> Lens' (States s) (VU.MVector s a) -> a -> ST s ()
setState !partition !s l = VU.unsafeWrite (partition^.states.l) s

setStateAt :: VUU.Unbox a => RefinablePartition s -> Int -> Lens' (States s) (VU.MVector s a) -> a -> ST s ()
setStateAt !partition !loc l x = VU.unsafeRead (partition^.statesByBlock) loc >>= \state ->
  setState partition state l x


getBlock :: VUU.Unbox a => RefinablePartition s -> Block -> Lens' (Blocks s) (VU.MVector s a) -> ST s a
getBlock !partition (Block b) l = VU.unsafeRead (partition^.blocks.l) b


setBlock :: VUU.Unbox a => RefinablePartition s -> Block -> Lens' (Blocks s) (VU.MVector s a) -> a -> ST s ()
setBlock !partition (Block b) l = VU.unsafeWrite (partition^.blocks.l) b


newBlock :: RefinablePartition s -> Int -> Int -> ST s Block
newBlock partition beginning end = do
  blk <- allocateBlock partition
  setBlock partition blk startOffset beginning
  setBlock partition blk endOffset end
  setBlock partition blk unmarkedOffset beginning -- no marked blocks by default

  -- update block of all contained states
  forM_ [beginning..end-1] $ \pos ->
    setStateAt partition pos block blk

  return blk

allocateBlock :: RefinablePartition s -> ST s Block
allocateBlock partition = do
  current <- readMutVar (partition^.blockCount)
  writeMutVar (partition^.blockCount) (current+1)
  return $ Block current


-- Update location fields of all states in this block
updateLocations :: RefinablePartition s -> Int -> Int -> ST s ()
updateLocations partition start end =
  forM_ [start..end-1] $ \i ->
    setStateAt partition i location i
