{-# LANGUAGE BangPatterns #-}

-- | Additional functions for unboxed mutable vectors
module Data.Vector.Unboxed.Mutable.Utils
  ( -- * Partitioning
    partition
  , partitionM
  -- * Grouping
  , groupBy
  , groupByM
  ) where

import           Control.Monad (foldM)

import           Control.Monad.Primitive
import qualified Data.Vector.Unboxed.Mutable as VU

-- | Partition a region inside a vector according to some predicate.
--
-- Moves all elements for which the predicate is 'True' to the start and all
-- other elements to the end of the region. Return the index of the first
-- element for which the predicate is 'False'.
--
-- ==== __Example__
--
-- >>> import qualified Data.Vector.Unboxed as V
-- >>> vec' <- V.thaw (V.fromList [3, 2, 1] :: V.Vector Int)
-- >>> s <- partition vec' (<=2) 0 3
-- >>> vec <- V.freeze vec'
-- >>> V.splitAt s vec
-- ([1,2],[3])
partition :: (VU.Unbox a, PrimMonad m)
          => VU.MVector (PrimState m) a -- ^ Vector to modify
          -> (a -> Bool) -- ^ The predicate
          -> Int -- ^ The beginning of the region to partition (inclusive)
          -> Int -- ^ The end of the region to partition (exclusive)
          -> m Int
partition vec predicate  = go
  where
    go lower upper
      | lower >= upper = return lower
      | otherwise = do
          l <- VU.read vec lower
          r <- VU.read vec (upper-1)

          let predL = predicate l
          let predR = predicate r

          if predL then go (lower+1) upper
            else if not predR then go lower (upper-1)
            else VU.swap vec lower (upper-1) >> go (lower+1) upper
{-# INLINE partition #-}

-- | 'partition' with monadic predicate
partitionM :: (VU.Unbox a, PrimMonad m)
           => VU.MVector (PrimState m) a
           -> (a -> m Bool) -- ^ Vector to modify
           -> Int -- ^ The predicate
           -> Int -- ^ The beginning of the region to partition (inclusive)
           -> m Int -- ^ The end of the region to partition (exclusive)
partitionM vec predicate  = go
  where
    go lower upper
      | lower >= upper = return lower
      | otherwise = do
          l <- VU.read vec lower
          r <- VU.read vec (upper-1)

          predL <- predicate l
          predR <- predicate r

          if predL then partitionM vec predicate (lower+1) upper
            else if not predR then partitionM vec predicate lower (upper-1)
            else VU.swap vec lower (upper-1) >> partitionM vec predicate (lower+1) upper
{-# INLINE partitionM #-}


-- | Group a region inside a sorted vector according to a predicate.
--
-- Given a vector that is already sorted with respect to the predicate, this
-- returns a list of indices i where a block new of consecutive equal elements
-- (according to the predicate) begins. In other words, returns a list of
-- indices @i@, such that @p v[i-1]@ is different than @p v[i]@. The vector is
-- not modified (and must be already sorted).
--
-- ==== __Example__
--
-- >>> import Data.Vector.Unboxed (freeze,thaw,fromList,Vector)
-- >>> import Data.Ord
-- >>> v <- thaw (fromList [1,1,2,2,2,3,3] :: Vector Int)
-- >>> groupBy v (comparing id) 0 6
-- [2,5]
groupBy :: forall a m. (VU.Unbox a, PrimMonad m)
        => VU.MVector (PrimState m) a -- ^ The vector to group
        -> (a -> a -> Ordering) -- ^ The predicate
        -> Int -- ^ The beginning of the region (inclusive)
        -> Int -- ^ The end of the region (exclusive)
        -> m [Int]
groupBy vec predicate = groupByM vec (\a b -> return (predicate a b))
{-# INLINE groupBy #-}

-- | 'groupBy' with monadic predicate.
groupByM :: forall a m. (VU.Unbox a, PrimMonad m)
         => VU.MVector (PrimState m) a -- ^ The vector to group
         -> (a -> a -> m Ordering) -- ^ The predicate
         -> Int -- ^ The beginning of the region (inclusive)
         -> Int -- ^ The end of the region (exclusive)
         -> m [Int]
groupByM vec predicate lower upper
  | lower >= upper = return [] -- special case empty range
  | otherwise = do
      first <- VU.read vec lower
      reverse . snd <$> foldM groupByImpl (first, []) [lower+1..upper-1]

  where
    groupByImpl :: (a,[Int]) -> Int -> m (a,[Int])
    groupByImpl (current, !accu) i
      | i >= upper = return (current,accu)
      | otherwise = do
          x <- VU.read vec i
          predicate current x >>= \case
            EQ -> return (x, accu)
            _  -> return (x, i:accu)
{-# INLINE groupByM #-}
