{-# LANGUAGE PartialTypeSignatures #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ApplicativeDo #-}
{-# OPTIONS_GHC -Wno-partial-type-signatures #-}

module Main (main) where

import           Control.Arrow ((***))
import           Control.Exception (evaluate)
import           Control.Monad.ST
import           Data.IORef
import           Data.List (sortOn)
import           System.Exit
import           System.IO
import           System.Environment

import           Data.Map (Map)
import qualified Data.Map as M
import           Data.Text (Text)
import qualified Data.Text as T
import qualified Data.Text.IO as T
import           Data.Text.Prettyprint.Convert.AnsiWlPprint
import           Data.Text.Prettyprint
import           Data.Proxy
import qualified Data.Vector as V
import           Options.Applicative
import           Text.JSON
import           Text.Show.Pretty (pPrint)

import           Copar.Algorithm
import qualified Copar.Parser as P
import           Copar.PartitionPrinter
import           Copar.FunctorPrinter
import           Copar.Functors
import           Copar.FunctorDescription
import qualified Data.MorphismEncoding as Encoding
import           Copar.FunctorExpression.Sorts (Sort, sortedSort)
import qualified Data.Partition as Partition
import           Copar.Functors.SomeFunctor (SomeFunctor)
import           Copar.FunctorExpression.Type (FunctorExpression)
import           Copar.Dot
import           Copar.RewriteFunctors
import           Copar.Timing
import           Copar.Functors

import qualified Hardwired as Hardwired
import           Stats

----------------------------------------------------------------------
-- CLI Options
----------------------------------------------------------------------

newtype Options = Options
  { optCommand :: SubCommand }

data SubCommand
  = HelpCommand (Maybe HelpCommand)
  | RefineCommand RefineOptions
  | RefineHardwiredCommand RefineOptions
  | PrettyPrintCommand PrettyPrintOptions
  | GraphCommand GraphOptions
  | DebugCommand DebugCommand

subcommand :: Parser SubCommand
subcommand =
  hsubparser
    $  (command
         "help"
         (info (HelpCommand <$> (optional helpCommand)) (progDesc "Print documentation on CoPaR's input file syntax."))
       )
    <> (command
         "refine"
         (info
           (RefineCommand <$> refineOptions)
           (progDesc
             "Refine the initial partition of a given coalgebra and output the result as a list of blocks of states."
           <> footerDoc (Just (toAnsiWlPprint refineHelp))
           )
         )
       )
    <> (command
         "refine-hardwired"
         (info
           (RefineHardwiredCommand <$> refineOptions)
           (progDesc
             "Refine coalgebra and output the result as a list of blocks of states. (Experimental & Only for testing & only for hard-wired functors)"
           <> footerDoc (Just (toAnsiWlPprint refineHelp))
           )
         )
       )
    -- <> (command
    --      "pretty-print"
    --      (info
    --        (PrettyPrintCommand <$> prettyPrintOptions)
    --        (progDesc
    --          "Print the given input coalgebra after parsing."
    --        <> footerDoc (Just (toAnsiWlPprint prettyPrintHelp))
    --        )
    --      )
    --    )
    <> (command
         "graph"
         (info
           (GraphCommand <$> graphOptions)
           (  progDesc "Print coalgebra as graphviz graph."
           <> footerDoc (Just (toAnsiWlPprint graphHelp))
           )
         )
       )
    <> (command
         "debug"
         (info
           (DebugCommand <$> debugCommand)
           (  progDesc "Tools for debugging this program."
           <> footer
                "These commands will likely only be useful if you are working \
                \on the source code of this program."
           )
         )
       )

data RefineOptions = RefineOptions
  { refineStats :: Bool
  , refineStatsJson :: Bool
  , refineFunctor :: Maybe (FunctorExpression SomeFunctor Sort)
  , refineApplyTransformations :: Bool
  , refineEnableSanity :: Bool
  , refineEnableOpt :: Bool
  , refineInputFile :: Maybe FilePath
  , refineOutputFile :: Maybe FilePath
  }

refineHelp :: Doc AnsiStyle
refineHelp
  = reflow "This command reads a coalgebra as input, refines the implicit \
           \initial partition and outputs a list of blocks in the final \
           \partition."
  <> line <> line
  <> reflow "Please see the 'copar help' command for a description of the input \
            \file syntax."
  <> line <> line
  <> reflow "The output consists of a list of blocks. Each block is printed as \
            \a sepeate line of the format"
  <> softline <> "'Block i: x1, x2, ...'" <> softline
  <> reflow "where i is the number of the block (counted from 0) and x_1, .. \
            \are the states contained in this block."
  <> line <> line
  <> reflow "Note that the states printed are only those states of the input \
            \coalgebra that were explicitly named and not any intermediates \
            \arising from composite functors. Formally, these are the states in \
            \the first sort when the input coalgebra is transformed into a \
            \multisorted coalgebra."

refineOptions :: Parser RefineOptions
refineOptions = do
  refineInputFile <- optional
    (argument
      str
      (  metavar "INPUT_FILE"
      <> help
           "Name of file to read the input coalgebra from or \"-\" for standard input. \
            \See the \"help\" command for a description of available syntax."
      <> value "-"
      <> showDefault
      )
    )
  refineOutputFile <- optional
    (argument
      str
      (metavar "OUTPUT_FILE" <> help
        "Name of file to print the output to or \"-\" for standard output."
      )
    )
  refineStats <- switch
    (  long "stats"
    <> help
         "Print runtime statistics (such as timing information) to standard error."
    )
  refineStatsJson <- switch
    (  long "stats-json"
    <> help
         "Same as --stats, but in json format, suitable for parsing in other programs."
    )
  refineFunctor <- optional
    (option
      functorReader
      (  long "functor"
      <> short 'f'
      <> metavar "FUNCTOR_EXPR"
      <> help
           "Functor for the input coalgebra. This is normally the first line of \
               \the input, but can also alternatively be given here."
      )
    )
  refineApplyTransformations <- not <$> switch
    (  long "no-functor-transforms"
    <> help
         "Don't try to optimize functor expression. \
         \This flag can drastically *reduce* performace."
    )
  refineEnableSanity <- not <$> switch
    (  long "no-sanity-checks"
    <> help
         "Disable sanity checks in the parser. Do not use this lightly. It \
         \might speed up the parser but require the input to be absolutely \
         \correct. Otherwise, nasal demons might be created."
    )
  refineEnableOpt <- not <$> switch
    (  long "disable-optimizations"
    <> help "Disable some optimizations. Currently, this is just the \
            \one-element optimization. Obviously, this will reduce \
            \performance."
    )
  pure RefineOptions { .. }

data PrettyPrintOptions = PrettyPrintOptions
  { prettyPrintInputFile :: Maybe FilePath
  }

prettyPrintHelp :: Doc AnsiStyle
prettyPrintHelp = reflow "This directly prints an input coalgebra after parsing."

prettyPrintOptions :: Parser PrettyPrintOptions
prettyPrintOptions = do
  prettyPrintInputFile <- optional
    (argument
      str
      (  metavar "INPUT_FILE"
      <> help
           "Name of file to read the input coalgebra from or \"-\" for standard input. \
            \See the \"help\" command for a description of available syntax."
      <> value "-"
      <> showDefault
      )
    )
  pure PrettyPrintOptions { .. }

data GraphOptions = GraphOptions
  { graphInputFile :: Maybe FilePath
  , graphOutputFile :: Maybe FilePath
  , graphFunctor :: Maybe (FunctorExpression SomeFunctor Sort)
  , graphApplyTransformations :: Bool
  , graphDrawPartition :: Bool
  , graphDrawNodeLabels :: Bool
  , graphDrawEdgeLabels :: Bool
  }

graphHelp :: Doc AnsiStyle
graphHelp =
  reflow
    "This command outputs a graph representation of the input coalgebra in graphviz format." <>
  softline <>
  reflow "Use the dot(1) command to render the resulting file as an image." <>
  line <>
  line <>
  reflow
    "Nodes in the resulting graph are colored according to their sort. Labels depend on the functor."

graphOptions :: Parser GraphOptions
graphOptions = do
  graphInputFile <- optional
    (argument
      str
      (  metavar "INPUT_FILE"
      <> help
           "Name of file to read the input coalgebra from or \"-\" for standard input. \
            \See the \"help\" command for a description of available syntax."
      <> value "-"
      <> showDefault
      )
    )
  graphOutputFile <- optional
    (argument
      str
      (  metavar "OUTPUT_FILE"
      <> help
           "Name of the file to print the output to or \"-\" for standard output."
      <> value "-"
      <> showDefault
      )
    )
  graphFunctor <- optional
    (option
      functorReader
      (  long "functor"
      <> short 'f'
      <> metavar "FUNCTOR_EXPR"
      <> help
           "Functor for the input coalgebra. This is normally the first line of \
               \the input, but can also alternatively be given here."
      )
    )
  graphDrawPartition <- switch
    (long "draw-partition" <> short 'P' <> help
      "Calculate partition (see refine) and group nodes accordingly."
    )
  graphDrawNodeLabels <- not <$> switch
    (  long "no-node-labels"
    <> help
         "Don't include node labels (except for explicitly named nodes). \
         \This can make larger graphs more readable."
    )
  graphDrawEdgeLabels <- not <$> switch
    (long "no-edge-labels" <> help
      "Don't include edge labels. This can make larger graphs more readable."
    )
  graphApplyTransformations <- not <$> switch
    (  long "no-functor-transforms"
    <> help "Don't try to optimize functor expression."
    )
  pure GraphOptions { .. }

functorReader :: ReadM (FunctorExpression SomeFunctor Sort)
functorReader = eitherReader (\input -> P.parseFunctor input (T.pack input))

data HelpCommand = HelpListFunctors

helpCommand :: Parser HelpCommand
helpCommand = subparser $ command
  "functors"
  (info (pure HelpListFunctors)
   (progDesc "Show documentation for individual functors"))

options :: Parser Options
options = Options <$> subcommand

data DebugCommand = DebugFunctor Bool (FunctorExpression SomeFunctor Sort)

debugCommand :: Parser DebugCommand
debugCommand = hsubparser $ command
  "print-functor"
  (info
    (   DebugFunctor
    <$> (not <$> switch
          (  long "no-functor-transforms"
          <> help "Don't try to optimize functor expression."
          )
        )
    <*> argument functorReader (metavar "FUNCTOR_EXPRESSION")
    )
    (progDesc
      "Parse a functor expression and its internal representation."
    )
  )

----------------------------------------------------------------------
-- File handling
----------------------------------------------------------------------

readCoalgebra :: P.Config -> Maybe FilePath -> IO _
readCoalgebra config Nothing     = P.readStdin config
readCoalgebra config (Just "-" ) = P.readStdin config
readCoalgebra config (Just file) = P.readFile config file

outputPartition :: Maybe FilePath -> _
outputPartition Nothing = printPartition
outputPartition (Just "-") = printPartition
outputPartition (Just file) = writePartition file

mainHelp :: Text -> Doc AnsiStyle
mainHelp argv0' =
  reflow "The main command is" <> softline <> "'" <> argv0 <> " refine'."
  <> line <> line
  <> reflow "Please see the output of" <> softline <> "'" <> argv0
  <> " refine --help'" <> softline
  <> "and" <> softline <> "'" <> argv0 <> " help'."

  where argv0 = pretty argv0'

----------------------------------------------------------------------
-- Main
----------------------------------------------------------------------

main :: IO ()
main = do
  argv0 <- T.pack <$> getProgName
  termColumns <- flip fmap (lookupEnv "COLUMNS") $ \case
    Nothing  -> 120
    Just val -> min (read val) 120
  let optionParserPrefs = prefs (showHelpOnEmpty <> columns termColumns)
  params <- customExecParser optionParserPrefs
    $ info (options <**> helper)
           (fullDesc
           <> header "CoPaR - An Efficient Coalgebraic Partition Refiner"
           <> footerDoc (Just (toAnsiWlPprint (mainHelp argv0)))
           )

  case optCommand params of
    (HelpCommand   f) -> do
      printHelp argv0 f
    (RefineCommand r) -> do
      stats <- initStats (refineStats r) (refineStatsJson r)

      withTimeStat stats "overall-duration" $ do
        (f, (symbolTable, encoding)) <- withTimeStat stats "parse-duration" $ do

          let transPolicy = if refineApplyTransformations r
                then P.ApplyTransformations
                else P.DontApplyTransformations

          let sanity = if refineEnableSanity r
                then P.EnableSanityChecks
                else P.DisableSanityChecks

          let parserConfig = P.Config
                { functorTransforms = transPolicy
                , sanityChecks = sanity
                , functor = refineFunctor r
                }

          readCoalgebra parserConfig (refineInputFile r)
            >>= \case
                  Left  err -> hPutStrLn stderr err >> exitFailure
                  Right res -> evaluate res

        logStat stats "states" (tshow (Encoding.size encoding))
        logStat stats "edges"  (tshow (Encoding.numEdges encoding))

        let isFirstSort x = sortedSort x == 1
        logStat
          stats
          "explicit-states"
          (tshow (length (V.filter isFirstSort (Encoding.structure encoding))))

        partition <- case statsType stats of
          NoStats ->
            withTimeStat stats "algorithm-duration" (stToIO (refine f encoding (refineEnableOpt r)))
          _ -> do
            (part, algoStats) <-
              withTimeStat stats "algorithm-duration"
                $ refineWithStats f encoding (refineEnableOpt r)
            logStat stats
                    "initial-partition-size"
                    (tshow (initialBlocks algoStats))
            logStat stats
                    "initialize-duration"
                    (showTimeDiff (initTime algoStats))
            logStat stats "split-operation-count" (tshow (splitCount algoStats))
            logStat stats
                    "refine-duration"
                    (showTimeDiff (refineTime algoStats))
            logStat stats
                    "size1-skipped"
                    (tshow (size1Skipped algoStats))
            return part

        logStat stats
                "final-partition-size"
                (tshow (Partition.numBlocks partition))

        logStat stats
                "explicit-final-partition-size"
                (tshow (length (restrictPartitionToSort1 encoding partition)))

        withTimeStat
          stats
          "output-duration"
          (outputPartition (refineOutputFile r) encoding symbolTable partition)

      finalizeStats stats

    (RefineHardwiredCommand r) -> do
      stats <- initStats (refineStats r) (refineStatsJson r)
      Hardwired.main
        stats
        (refineInputFile r)
        (outputPartition (refineOutputFile r))
        (refineEnableOpt r)


    (PrettyPrintCommand r) -> do
      (f, (symbolTable, encoding)) <- do
          let parserConfig = P.Config
                { functorTransforms = P.DontApplyTransformations
                , sanityChecks = P.DisableSanityChecks
                , functor = Nothing
                }
          readCoalgebra parserConfig (prettyPrintInputFile r)
            >>= \case
                  Left  err -> hPutStrLn stderr err >> exitFailure
                  Right res -> evaluate res
      hPutStrLn stderr "tesetste"

    (GraphCommand r) -> do
      (f, (symbolTable, encoding)) <- do
        let transPolicy = if graphApplyTransformations r
              then P.ApplyTransformations
              else P.DontApplyTransformations

        let parserConfig = P.defaultConfig
              { P.functorTransforms = transPolicy
              , P.functor = graphFunctor r
              }

        readCoalgebra parserConfig (graphInputFile r) >>= \case
          Left  err -> hPutStrLn stderr err >> exitFailure
          Right res -> evaluate res

      part <- if graphDrawPartition r
        then Just <$> stToIO (refine f encoding True)
        else return Nothing

      let config = DotConfig { nodeLabels = graphDrawNodeLabels r
                             , edgeLabels = graphDrawEdgeLabels r
                             }

      case graphOutputFile r of
        Nothing   -> printDot config symbolTable encoding part
        Just "-"  -> printDot config symbolTable encoding part
        Just file -> writeDot file config symbolTable encoding part

    DebugCommand (DebugFunctor applyTrans f)
      | applyTrans -> T.putStrLn (printFunctor (applyFunctorRewrites f))
      | otherwise  -> T.putStrLn (printFunctor f)

helpOverview :: Text -> Doc AnsiStyle
helpOverview argv0 =
  reflow
      "This help describes the input syntax of CoPaR. For an overview of \
         \command line parameters, run"
    <> softline <>  pretty argv0 <+> "--help."
    <>  paragraph
    <> reflow "Input files for CoPaR describe a transition system as a finite \
              \coalgebra" <> noflow "X -> FX."
    <> reflow "They begin with a single line that specifies the functor F and \
              \each of the remaining lines define a state"
    <> noflow "x € X"
    <> reflow "of and its behavior. Everything in a line after a '#' \
              \character is considered a comment and ignored. As an example, a \
              \file containing"
    <> paragraph
    <>  indent 2 ("DX" <> line <> line
                 <> "q1: {q2: 0.5, q3: 0.5}" <> line
                 <> "q2: {q1: 0.5, q3: 0.5}" <> line
                 <> "q3: {q3: 1}" <> line)
    <> line
    <> reflow "describes a coalgebra for the distribution functor DX \
              \with three states q1, q2 and q3."
    <> paragraph
    <> annotate (italicized <> underlined) "Functor specification."
    <+> reflow "Functors are generally described as composites of basic \
               \building blocks; that is, the functor given in the first line \
               \of an input file is an expression determined by the grammar"
    <> paragraph
    <> indent 2 "T ::= X | F(T,...,T),"
    <> paragraph
    <> reflow "where X is a terminal symbol and F is one of the currently \
              \implemented basic functors described in" <> softline <> "\""
    <> pretty argv0 <+> "help functors\"."
    <> paragraph
    <> annotate (italicized <> underlined) "Coalgebra specification."
    <+> reflow "The remaining lines define a finite F-coalgebra"
    <> softline <> "xi: X -> FX." <> softline
    <> reflow "Each line of the form" <> noflow "'x: t'"
    <> reflow "defines a state" <> noflow "x ∈ X,"
    <> reflow "where x is a C-style identifier, and t represents the element"
    <> noflow "t = xi(x) ∈ FX."
    <> reflow "The syntax for t depends on the specified functor F and follows \
              \the structure of the term T defining F. For the base case"
    <> noflow "T = X,"
    <> reflow "t is an element of X, i.e. one of the named states specified in \
              \the file. For the other case" <> noflow "T = F(T),"
    <> reflow "where F is one of the basic functors, syntax and semantics is \
              \described in" <> softline <> "\"" <> pretty argv0
    <+> "help functors\"."
    <> paragraph <> line
    <> reflow "The 'examples/' directory in the source code of CoPaR also \
              \contains lots of ready to use example files."
    <> paragraph
    <> reflow "Please see" <> noflow "\"copar help functors\""
    <> reflow "for more help on the individual functors and their syntax."

  where
    paragraph = line <> line
    noflow x = softline <> x <> softline

printHelp :: Text -> Maybe HelpCommand -> IO ()
printHelp argv0 Nothing = putDoc (helpOverview argv0) >> putStrLn ""
printHelp _ (Just HelpListFunctors) =
  let functors = sortOn functorName registeredFunctors
  in do
    T.putStrLn "Available functors:"
    putDoc (formatFunctorDescriptions functors)
    -- tabularize (map (\f -> [("  - " <> name f), (syntaxExample f)]) functors)
  where functorName (DynFunctorDescription (Proxy :: Proxy f)) = name @f

tshow :: Show a => a -> Text
tshow = T.pack . show
