{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE TypeApplications #-}

module Main (main) where

import           System.IO
import           System.Exit
import           Control.Applicative
import           Data.Maybe (fromMaybe)

import           Data.Set (Set)
import           Data.Text (Text)
import qualified Data.Text as T
import qualified Data.Text.IO as T
import           Data.Text.Lazy.Builder (Builder)
import qualified Data.Text.Lazy.Builder as Build
import qualified Data.Text.Lazy.IO as TL
import           Data.Text.Prettyprint.Convert.AnsiWlPprint
import           Data.Text.Prettyprint.Doc
import           Data.Text.Prettyprint.Doc.Render.Terminal
import qualified Options.Applicative as OptParse
import           Text.Megaparsec

import           Parser
import           MarkovChain
import           Mdp
import           StatesFile

data ModelType = DTMC | CTMC | MDP

instance Show ModelType where
  show DTMC = "dtmc"
  show CTMC = "ctmc"
  show MDP = "mdp"

data OutputFormat = Valmari | Valmari2 | Copar | Mcrl2

instance Show OutputFormat where
  show Valmari = "valmari"
  show Valmari2 = "valmari2"
  show Copar = "copar"
  show Mcrl2 = "mcrl2"

parseOutputFormat :: String -> Either String OutputFormat
parseOutputFormat "valmari" = Right Valmari
parseOutputFormat "valmari2" = Right Valmari2
parseOutputFormat "copar" = Right Copar
parseOutputFormat "mcrl2" = Right Mcrl2
parseOutputFormat other = Left ("Unknown output format '" <> other <> "'")

data Options = Options
  { optModelType :: ModelType
  , optInputFile :: Maybe FilePath
  , optOutputFile :: Maybe FilePath
  , optStatesFile :: Maybe FilePath
  , optOutputFormat :: OutputFormat
  , optPartitionVariables :: Maybe (Set Text)
  }

readModelType :: OptParse.ReadM ModelType
readModelType = OptParse.eitherReader $ \case
  "dtmc" -> Right DTMC
  "ctmc" -> Right CTMC
  "mdp" -> Right MDP
  other -> Left ("Unknown model type \"" ++ other ++ "\"")

modelTypeParser :: OptParse.Parser ModelType
modelTypeParser =
  OptParse.option
    readModelType
    (OptParse.long "model-type" <> OptParse.short 'm' <>
     OptParse.help "Type of the PRISM model. Can be 'dtmc', 'ctmc' or 'mdp'." <>
     OptParse.metavar "TYPE" <>
     OptParse.showDefault <>
     OptParse.value MDP)

optionsParser :: OptParse.Parser Options
optionsParser =
  Options <$> modelTypeParser <*>
  optional
    (OptParse.strArgument
       (OptParse.metavar "INPUT_FILE" <> OptParse.value "-" <>
        OptParse.showDefault <>
        OptParse.help
          "Input file in PRISMs .tra-format, generated with `prism -exporttrans`. Can be a file name or \"-\" for stdin.")) <*>
  optional
    (OptParse.strArgument
       (OptParse.metavar "OUTPUT_FILE" <> OptParse.value "-" <>
        OptParse.showDefault <>
        OptParse.help
          "Output file for the corresponding coalgebra. Can be a file name or \"-\" for stdout.")) <*>
  optional
    (OptParse.strOption
       (OptParse.long "states-file" <>
        OptParse.help ".sta file generated by PRISM with -exportstates." <>
        OptParse.metavar "FILE")) <*>
  (OptParse.option
     (OptParse.eitherReader parseOutputFormat)
     (OptParse.long "output-format" <>
      OptParse.help
        "Syntax used for the output file. Can be either 'copar', 'valmari', 'valmari2' or 'mcrl2'" <>
      OptParse.metavar "FORMAT" <>
      OptParse.value Copar <>
      OptParse.showDefault)) <*>
  optional
    (OptParse.option
       (OptParse.eitherReader parseVarIndices)
       (OptParse.long "partition-on-variables" <>
        OptParse.help
          "comma separated list of variables that should be used to construct the initial partition." <>
        OptParse.metavar "VARS"))

parseInput :: FilePath -> Parser a -> IO a
parseInput file parser = do
  (name, input) <- case file of
    "-" -> ("(stdin)",) <$> T.getContents
    other -> (other,) <$> T.readFile other

  case parse (parser <* eof) name input of
    Left err -> do
      hPutStr stderr (errorBundlePretty err)
      exitFailure
    Right res ->
      return res

printOutput :: FilePath -> Builder -> IO ()
printOutput "-" = TL.putStr . Build.toLazyText
printOutput file = TL.writeFile file . Build.toLazyText

convert :: Options -> Parser a -> (a -> Builder) -> IO ()
convert opts parser builder =
  parseInput (fromMaybe "-" (optInputFile opts)) parser
  >>= printOutput (fromMaybe "-" (optOutputFile opts)) . builder

helpText :: Doc AnsiStyle
helpText =
  "convert PRISM transition matrices to coalgebras\n\
  \\n\
  \This reads transition matrices of PRISM[1] models and outputs coalgebras in our input format.\n\
  \\n\
  \[1]: https://www.prismmodelchecker.org"

main :: IO ()
main = do
  opts <-
    OptParse.customExecParser
      (OptParse.prefs (OptParse.columns 100))
      (OptParse.info
         (optionsParser <**> OptParse.helper)
         (OptParse.fullDesc <>
          OptParse.progDescDoc (Just (toAnsiWlPprint helpText))))

  states <- traverse (flip parseInput statesFileP) (optStatesFile opts)

  initPartition <- case (states, optPartitionVariables opts) of
    (Nothing, Just _) -> do
      hPutStrLn stderr "--partition-on-variables used, but no states files specified. Please use also --states-file."
      exitFailure
    (Just sf, Just vars) -> case variableIndices sf vars of
      Left err -> do
        hPutStrLn stderr $ "error while parsing --partition-on-variables: " <> T.unpack err
        exitFailure
      Right vs -> return (Just (computePartition sf vs))
    _ -> return Nothing

  case (optModelType opts, optOutputFormat opts) of
    (MDP, outType) ->
      let builder = case outType of
            Valmari -> valmariMdpB
            Valmari2 -> valmariMdp2B
            Copar -> mdpB
            Mcrl2 -> mcrl2B
      in convert opts mdpP (builder initPartition)
    (inType, outType) ->
      let mcType =
            case inType of
              DTMC -> DiscreteTime
              CTMC -> ContinuousTime
          builder =
            case outType of
              Valmari -> valmariMarkovChainB
              Valmari2 -> error "valmari2 only implemented for MDPs"
              Copar -> markovChainB
              Mcrl2 -> error "mcrl2-format only supports MDPs"
       in convert opts (markovChainP mcType) (builder initPartition)
