{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ScopedTypeVariables #-}

module MarkovChain
  ( MarkovType(..)
  , markovChainP
  , markovChainB
  , valmariMarkovChainB
  ) where

import           Data.Semigroup
import           Data.List (intersperse)
import           Data.Foldable

import           Data.IntMap (IntMap)
import qualified Data.IntMap as M
import qualified Data.Text.Lazy.Builder as Build
import qualified Data.Text.Lazy.Builder.Int as Build
import qualified Data.Text.Lazy.Builder.RealFloat as Build
import           Data.Vector (Vector)
import qualified Data.Vector as V

import           Parser
import           StatesFile (Partition(..))

----------------------------------------------------------------------
-- Types
----------------------------------------------------------------------

data MarkovChain = MarkovChain
  { markovType :: MarkovType
  , numStates :: Int
  , transitions :: Vector Transition
  }

data MarkovType = ContinuousTime | DiscreteTime

data Transition = Transition
  { from :: Int
  , to :: Int
  , label :: Double
  }

----------------------------------------------------------------------
-- Parser
----------------------------------------------------------------------

markovChainP :: MarkovType -> Parser MarkovChain
markovChainP mtype = MarkovChain mtype <$> decimalP <*> transitionsP

transitionsP :: Parser (Vector Transition)
transitionsP = do
  numTrans <- decimalP
  V.replicateM numTrans transitionP

transitionP :: Parser Transition
transitionP = Transition <$> decimalP <*> decimalP <*> doubleP

----------------------------------------------------------------------
-- Builder
----------------------------------------------------------------------

markovChainB :: Maybe Partition -> MarkovChain -> Build.Builder
markovChainB partition mc =
  functorB partition (markovType mc) <>
  "\n" <>
  transitionsB partition (transitions mc)

functorB :: Maybe Partition -> MarkovType -> Build.Builder
functorB partition DiscreteTime   = partitionFunctorB partition <> "DX\n"
functorB partition ContinuousTime = partitionFunctorB partition <> "R^X\n"

partitionFunctorB :: Maybe Partition -> Build.Builder
partitionFunctorB Nothing          = ""
partitionFunctorB (Just partition) = Build.decimal (numBlocks partition) <> "x "

transitionsB :: Maybe Partition -> Vector Transition -> Build.Builder
transitionsB partition ts =
  let stateMap :: IntMap [(Int, Double)] =
        V.foldl'
          (\m trans ->
             M.insertWith (++) (from trans) [(to trans, label trans)] m)
          M.empty
          ts
   in foldMap (uncurry (transitionB partition)) (M.toList stateMap)

transitionB :: Maybe Partition -> Int -> [(Int, Double)] -> Build.Builder
transitionB partition from successors =
  stateB from <> ": " <> partStart partition <> "{" <>
  fold (intersperse ", " (map successorB successors)) <>
  "}" <>
  partEnd partition <>
  "\n"
  where
    stateB :: Int -> Build.Builder
    stateB = ("s" <>) . Build.decimal
    successorB :: (Int, Double) -> Build.Builder
    successorB (to, weight) =
      stateB to <> ": " <> Build.formatRealFloat Build.Fixed Nothing weight
    partStart Nothing = ""
    partStart (Just part) =
      "(" <> Build.decimal (stateAssignment part V.! from) <> ", "
    partEnd Nothing = ""
    partEnd (Just _) = ")"

----------------------------------------------------------------------
-- Builder for Valmari format
--
-- This format is essentially:
--
-- MC            ::=  <sizes> <transitions> <blocks>
-- <sizes>       ::= num_states 0 0 num_trans num_blocks
-- <transitions> ::= <transition>*
-- <transition>  ::= from weight to
-- <blocks>      ::= <block>*
-- <block>       ::= state* 0
--
-- Where blocks are the blocks of the initial partition. Also, one block is
-- omitted.
----------------------------------------------------------------------

valmariMarkovChainB :: Maybe Partition -> MarkovChain -> Build.Builder
valmariMarkovChainB partition mc =
  valmariSizesB partition mc <> "\n" <> valmariTransitionsB (transitions mc) <>
  valmariBlocksB (numStates mc) partition

valmariSizesB :: Maybe Partition -> MarkovChain -> Build.Builder
valmariSizesB part mc =
  Build.decimal (numStates mc) <> " 0 0 " <>
  Build.decimal (length (transitions mc)) <>
  " " <>
  (Build.decimal partitionSize)
  where
    partitionSize =
      case part of
        Nothing -> 1 -- No partition: One block with all states
        Just p -> (numBlocks p)

valmariTransitionsB :: Vector Transition -> Build.Builder
valmariTransitionsB = foldMap (\t -> valmariTransitionB t <> "\n")

valmariTransitionB :: Transition -> Build.Builder
valmariTransitionB t =
  state (from t) <> " " <> Build.formatRealFloat Build.Fixed Nothing (label t) <>
  " " <>
  state (to t)
  where
    state = Build.decimal . (+ 1) -- states are numbered from 1

valmariBlocksB :: Int -> Maybe Partition -> Build.Builder
valmariBlocksB n Nothing = ""
valmariBlocksB _ (Just p) =
  let blockMap :: IntMap [Int]
      blockMap = V.ifoldl' (\m s b -> M.insertWith (++) b [s] m) M.empty (stateAssignment p)
  in
    foldMap (\b -> valmariBlockB b) (tail (M.elems blockMap))

valmariBlockB :: [Int] -> Build.Builder
valmariBlockB states = foldMap (\s -> Build.decimal (s+1) <> " ") states <> " 0\n"
