{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE TypeApplications #-}

module Mdp (mdpP, mdpB, valmariMdpB, valmariMdp2B, mcrl2B) where

import           Data.List                      ( intersperse )
import           Data.Foldable
import           Control.Applicative

import           Data.IntMap                    ( IntMap )
import qualified Data.IntMap                   as M
import qualified Data.Text.Lazy.Builder        as Build
import qualified Data.Text.Lazy.Builder.Int    as Build
import qualified Data.Text.Lazy.Builder.RealFloat
                                               as Build
import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import           Lens.Micro.Platform

import           Mdp.Types
import           Parser
import           StatesFile                     ( Partition(..) )
import qualified Mdp.Mcrl2                     as Mcrl2

----------------------------------------------------------------------
-- helpers
----------------------------------------------------------------------

convertToDouble :: Mdp Rational -> Mdp Double
convertToDouble = fmap (fromRational @Double)

----------------------------------------------------------------------
-- Parser
----------------------------------------------------------------------

mdpP :: Parser (Mdp Rational)
mdpP = Mdp <$> decimalP <*> decimalP <*> transitionsP

transitionsP :: Parser (Vector (Transition Rational))
transitionsP = do
  numTrans <- decimalP
  V.replicateM numTrans transitionP

transitionP :: Parser (Transition Rational)
transitionP =
  Transition
    <$> decimalP
    <*> decimalP
    <*> decimalP
    <*> rationalP
    <*> optional nameP

----------------------------------------------------------------------
-- Builder
----------------------------------------------------------------------

mdpB :: Maybe Partition -> (Mdp Rational) -> Build.Builder
mdpB partition mdp =
  let mdp' = convertToDouble mdp
  in functorB partition <> "\n" <> transitionsB partition (mdp' ^. transitions)

functorB :: Maybe Partition -> Build.Builder
functorB Nothing = "P(Nx(DX))\n"
functorB (Just partition) =
  (Build.decimal (numBlocks partition)) <> " x P(Nx(DX))\n"

transitionsB :: Maybe Partition -> Vector (Transition Double) -> Build.Builder
transitionsB partition ts =
  let stateMap :: IntMap [(Transition Double)] =
        V.foldl' (\m t -> M.insertWith (++) (t ^. source) [t] m) M.empty ts
  in  foldMap (uncurry (choicesB partition)) (M.toList stateMap)

choicesB :: Maybe Partition -> Int -> [(Transition Double)] -> Build.Builder
choicesB partition src ts =
  let choiceMap :: IntMap [(Transition Double)] =
        foldl' (\m t -> M.insertWith (++) (t ^. choice) [t] m) M.empty ts
  in  stateB src
      <> ": "
      <> partStart partition
      <> "{"
      <> fold
           (intersperse ", " (map (uncurry transitionB) (M.toList choiceMap)))
      <> "}"
      <> partEnd partition
      <> "\n"
  where
    partStart Nothing = ""
    partStart (Just part) =
      "(" <> Build.decimal (stateAssignment part V.! src) <> ", "
    partEnd Nothing  = ""
    partEnd (Just _) = ")"

transitionB :: Int -> [(Transition Double)] -> Build.Builder
transitionB choi successors =
  "("
    <> Build.decimal choi
    <> ", {"
    <> fold (intersperse ", " (map successorB successors))
    <> "})"
  where
    successorB :: (Transition Double) -> Build.Builder
    successorB t = stateB (t ^. target) <> ": " <> Build.formatRealFloat
      Build.Fixed
      Nothing
      (t ^. probability)

stateB :: Int -> Build.Builder
stateB = ("s" <>) . Build.decimal

----------------------------------------------------------------------
-- Builder for Valmari format
--
-- This format is essentially:
--
-- MC              ::=  <sizes> <l_transitions> <w_transitions> <blocks>
-- <sizes>         ::= num_states num_choices num_l_trans num_w_trans num_blocks
-- <l_transitions> ::= <l_transition>*
-- <l_transition>  ::= source choice target
-- <w_transitions> ::= <w_transition>*
-- <w_transition>  ::= source weight target
-- <blocks>        ::= <block>*
-- <block>         ::= state* 0
--
-- Where blocks are the blocks of the initial partition. Also, one block is
-- omitted.
----------------------------------------------------------------------

data ValmariMDP = ValmariMDP
  { valmariMDPNumStates :: Int
  , valmariMDPNumActionStates :: Int
  , valmariMDPNumChoices :: Int
  , valmariMDPChoices :: Vector ValmariChoice
  , valmariMDPProbabilities :: Vector ValmariTransition
  } deriving (Show)

data ValmariChoice = ValmariChoice
  { _valmariChoiceSource :: Int
  , _valmariChoiceChoice :: Int
  , _valmariChoiceTarget :: Int
  } deriving (Show)


-- From an action state to a normal state
data ValmariTransition = ValmariTransition
  { _valmariTransitionSource :: Int
  , _valmariTransitionProbability :: Double
  , _valmariTransitionTarget :: Int
  } deriving (Show)

makeFields ''ValmariMDP
makeFields ''ValmariChoice
makeFields ''ValmariTransition

type ValmariPartition = [[Int]]


convertToValmari :: Maybe Partition -> (Mdp Double) -> (ValmariMDP, ValmariPartition)
convertToValmari maybePart mdp =
  let nums    = mdp ^. numStates
      (numa, transMap) = mkValmariMap mdp
      numc    = mdp ^. numChoices
      actions = V.fromList (concatMap mkChoices (M.toList transMap))
      probs   = V.fromList (concatMap mkTransitions (transMap ^.. each . each))
      actionStates = map (+ nums) [0 .. numa - 1]
      normalStates = [0 .. nums - 1]
      part    = actionStates : maybe [normalStates] partToBlocks maybePart
  in  ( ValmariMDP { valmariMDPNumStates     = nums
                   , valmariMDPNumActionStates = numa
                   , valmariMDPChoices       = actions
                   , valmariMDPProbabilities = probs
                   , valmariMDPNumChoices    = numc
                   }
      , part
      )
  where
    mkChoices (s, cs) = map (mkChoice s) cs
    mkChoice s (t, c) = ValmariChoice s (fst c) t
    mkTransitions (s, (_, ts)) = map (uncurry (ValmariTransition s)) ts


-- | Alternative method of modeling the same coalgebra. Here, we don't use
-- labels but partition the Distribution states on the label that leads to them.
-- This models the functor P(AxDX) more faithfully.
convertToValmari2 :: Maybe Partition -> (Mdp Double) -> (ValmariMDP, ValmariPartition)
convertToValmari2 maybePart mdp =
  let
    nums    = mdp ^. numStates
    (numa, transMap) = mkValmariMap mdp
    actions = V.fromList (concatMap mkChoices (M.toList transMap))
    probs   = V.fromList (concatMap mkTransitions (transMap ^.. each . each))
    stateByChoice = M.fromListWith
      (++)
      (map (\x -> (x ^. choice, [(x ^. target) + nums])) (V.toList actions))
    part = (maybe [[0 .. nums - 1]] partToBlocks maybePart)
      ++ M.elems stateByChoice
  in
    ( ValmariMDP { valmariMDPNumStates     = nums
                 , valmariMDPNumActionStates = numa
                 , valmariMDPChoices       = actions & each . choice .~ 0
                 , valmariMDPProbabilities = probs
                 , valmariMDPNumChoices    = 1
                 }
    , part
    )
  where
    mkChoices (s, cs) = map (mkChoice s) cs
    mkChoice s (t, c) = ValmariChoice s (fst c) t
    mkTransitions (s, (_, ts)) = map (uncurry (ValmariTransition s)) ts


-- State -> Choices
-- each choice is (choice state, choice number, [(probability, successor)])
type ValmariMap = IntMap [(Int, (Int, [(Double, Int)]))]


-- | Returns the total number of action states and the valmari map
mkValmariMap :: (Mdp Double) -> (Int, ValmariMap)
mkValmariMap mdp = M.mapAccum uniquifyChoices
                              0
                              (V.foldl' ins M.empty (mdp ^. transitions))
  where
    ins
      :: IntMap (IntMap [(Double, Int)])
      -> (Transition Double)
      -> IntMap (IntMap [(Double, Int)])
    ins m t =
      m
        &  (at (t ^. source) . non M.empty)
        .  (at (t ^. choice) . non [])
        %~ ((t ^. probability, t ^. target) :)

    uniquifyChoices
      :: Int -> IntMap [(Double, Int)] -> (Int, [(Int, (Int, [(Double, Int)]))])
    uniquifyChoices current actions =
      let newActions = zip [current ..] (M.toList actions)
      in  (current + length newActions, newActions)


partToBlocks :: Partition -> [[Int]]
partToBlocks p =
  let blockMap = V.ifoldl' (\m s b -> m & at b . non [] %~ (s :))
                           M.empty
                           (stateAssignment p)
  in  M.elems blockMap


valmariMdpB :: Maybe Partition -> (Mdp Rational) -> Build.Builder
valmariMdpB partition mdp =
  let (vmdp, vpartition) = convertToValmari partition (convertToDouble mdp)
  in  valmariMdpBImpl vmdp vpartition


valmariMdp2B :: Maybe Partition -> (Mdp Rational) -> Build.Builder
valmariMdp2B partition mdp =
  let (vmdp, vpartition) = convertToValmari2 partition (convertToDouble mdp)
  in  valmariMdpBImpl vmdp vpartition


valmariMdpBImpl :: ValmariMDP -> ValmariPartition -> Build.Builder
valmariMdpBImpl vmdp vpartition =
  valmariSizesB vmdp vpartition
    <> "\n"
    <> valmariLTransitions vmdp
    <> "\n"
    <> valmariWTransitions vmdp
    <> "\n"
    <> valmariBlocks vpartition


valmariSizesB :: ValmariMDP -> ValmariPartition -> Build.Builder
valmariSizesB mdp part =
  Build.decimal (mdp ^. numStates + mdp ^. numActionStates)
    <> " "
    <> Build.decimal (mdp ^. numChoices)
    <> " "
    <> Build.decimal (length (mdp ^. choices))
    <> " "
    <> Build.decimal (length (mdp ^. probabilities))
    <> " "
    <> Build.decimal (length part)


valmariLTransitions :: ValmariMDP -> Build.Builder
valmariLTransitions mdp = foldMap vChoice (mdp ^.. choices . each)
  where
    vChoice :: ValmariChoice -> Build.Builder
    vChoice c =
      vState (c ^. source)
        <> " "
        <> Build.decimal (c ^. choice + 1)
        <> " "
        <> vActionState mdp (c ^. target)
        <> "\n"

valmariWTransitions :: ValmariMDP -> Build.Builder
valmariWTransitions mdp = foldMap vTrans (mdp ^.. probabilities . each)
  where
    vTrans :: ValmariTransition -> Build.Builder
    vTrans t =
      vActionState mdp (t ^. source)
        <> " "
        <> Build.formatRealFloat Build.Fixed Nothing (t ^. probability)
        <> " "
        <> vState (t ^. target)
        <> "\n"


vState :: Int -> Build.Builder
vState = Build.decimal . (+ 1)


vActionState :: ValmariMDP -> Int -> Build.Builder
vActionState mdp = Build.decimal . (+ (mdp ^. numStates)) . (+ 1)


valmariBlocks :: [[Int]] -> Build.Builder
valmariBlocks part =
  part
    & tail
    & (each . each %~ vState)
    & (each %~ intersperse " ")
    & (each %~ fold)
    & (each %~ (<> " 0\n"))
    & fold

----------------------------------------------------------------------
-- mcrl2
----------------------------------------------------------------------

-- FIXME Support partition

mcrl2B :: Maybe Partition -> Mdp Rational -> Build.Builder
mcrl2B = ((.) . (.)) Mcrl2.mcrl2B Mcrl2.convertToMcrl


