{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE OverloadedStrings #-}

module Mdp.Mcrl2
  ( Mcrl2Mdp(Mcrl2Mdp)
  , Mcrl2Transition(Mcrl2Transition)
  , Mcrl2Distribution
  , Mcrl2PropTrans(Mcrl2PropTrans)
  , numStates
  , transitions
  , outDistribution
  , source
  , label
  , probability
  , target
  , convertToMcrl
  , mcrl2B
  ) where

import           Data.Ord
import           Control.Arrow                  ((&&&))
import           Data.Ratio
import           Data.Maybe

import           Lens.Micro.Platform
import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import qualified Data.Vector.Algorithms.Intro  as V
import           Data.Text                      ( Text )
import qualified Data.Text                     as T
import qualified Data.Text.Lazy.Builder        as Build
import qualified Data.Text.Lazy.Builder.Int    as Build

import           Mdp.Types
import           StatesFile                     (Partition(..))

----------------------------------------------------------------------
-- Builder for .aut-Format used in mcrl2
--
-- This format is informally:
--
-- des (initial distribution, nr transitions, nr states)
-- (state, "label", out distribution)
-- ...
----------------------------------------------------------------------

data Mcrl2Mdp = Mcrl2Mdp
  { mcrl2MdpNumStates :: Int
  , mcrl2MdpTransitions :: Vector Mcrl2Transition
  , mcrl2MdpOutDistribution :: Mcrl2Distribution
  } deriving (Show)

data Mcrl2Transition = Mcrl2Transition
  { mcrl2TransitionSource :: Int
  , mcrl2TransitionLabel :: Text
  , mcrl2TransitionOutDistribution :: Mcrl2Distribution
  } deriving (Show, Eq)

type Mcrl2Distribution = Vector Mcrl2PropTrans

data Mcrl2PropTrans = Mcrl2PropTrans
  { mcrl2PropTransProbability :: Rational
  , mcrl2PropTransTarget :: Int
  } deriving (Show, Eq)

makeFields ''Mcrl2Mdp
makeFields ''Mcrl2Transition
makeFields ''Mcrl2PropTrans


convertToMcrl :: Maybe Partition -> Mdp Rational -> Mcrl2Mdp
convertToMcrl maybePart mdp = Mcrl2Mdp stateCount ts initDistri
  where
    partitionTransitions = maybe mempty convertPartition maybePart
    ts = convertTransitionsToMcrl (mdp^.transitions) <> partitionTransitions
    stateCount = (mdp ^.numStates) * (if isJust maybePart then 2 else 1)
    initDistri = V.fromList $
      map (Mcrl2PropTrans (1%fromIntegral stateCount)) [0 .. stateCount - 1]

-- | Introduce a new state for each existing state with one edge between them,
-- labeled with the block number.
convertPartition :: Partition -> Vector Mcrl2Transition
convertPartition part = V.imap mkPartTrans (stateAssignment part)
  where
    len = length (stateAssignment part)
    mkPartTrans src blk = Mcrl2Transition (len + src) (mkPartLabel blk) (mkPartDist src)
    mkPartLabel blk = "i" <> T.pack (show blk)
    mkPartDist src = V.singleton (Mcrl2PropTrans 1 src)

convertTransitionsToMcrl :: Vector (Transition Rational) -> Vector Mcrl2Transition
convertTransitionsToMcrl trans = mkMcrl2Tans <$> partitionVector (view source &&& view choice) sorted
  where
    sorted = V.create $ do
      v <- V.thaw trans -- TODO Maybe use unsafe thaw
      V.sortBy (comparing $ view source &&& view choice) v
      return v


mkMcrl2Tans :: Vector (Transition Rational) -> Mcrl2Transition
mkMcrl2Tans distri = Mcrl2Transition (h ^. source) (T.pack $ show $ h ^. choice) (mkPropTrans <$> distri)
  where h = V.head distri


mkPropTrans :: (Transition Rational) -> Mcrl2PropTrans
mkPropTrans t = Mcrl2PropTrans (t ^. probability) (t ^. target)


partitionVector :: Eq b => (a -> b) -> Vector a -> Vector (Vector a)
partitionVector p = V.unfoldr helper
  where
    helper vec
      | null vec  = Nothing
      | otherwise = Just $ V.span (((p (V.head vec)) ==) . p) vec

----------------------------------------------------------------------
-- Pretty Printer
--
-- The format is something like the following:
--
-- des (1 1/4 2 1/4 3 1/4 0,240,240)
-- (0,"step",5 1/4 6 1/4 7 1/4 4)
-- (1,"step",9 1/4 10 1/4 11 1/4 8)
-- (2,"step",13 1/4 14 1/4 15 1/4 12)
-- (3,"step",17 1/4 18 1/4 19 1/4 16)
-- (4,"step",21 1/4 22 1/4 23 1/4 20)
-- (5,"step",1 1/4 2 1/4 3 1/4 0)
-- ...
----------------------------------------------------------------------

-- TODO Rationals!!!
mcrl2B :: Mcrl2Mdp -> Build.Builder
mcrl2B mdp =
  "des ("
    <> distributionB (mdp ^. outDistribution)
    <> ","
    <> Build.decimal (mdp ^. transitions . to length)
    <> ","
    <> Build.decimal (mdp ^. numStates)
    <> ")\n"
    <> foldMap transitionB (mdp^.transitions)


distributionB :: Mcrl2Distribution -> Build.Builder
distributionB distri = foldMap (\x -> propTransB x <> " ") (V.init distri)
  <> Build.decimal (V.last distri ^. target)


propTransB :: Mcrl2PropTrans -> Build.Builder
propTransB t = Build.decimal (t ^. target) <> " "
  <> ratioB (t ^. probability)


transitionB :: Mcrl2Transition -> Build.Builder
transitionB t = "(" <> Build.decimal (t ^. source) <> ","
  <> "\"" <> Build.fromText (t ^. label)
  <> "\"," <> distributionB (t ^. outDistribution)
  <> ")\n"


ratioB :: Rational -> Build.Builder
ratioB r = Build.decimal (numerator r) <> "/" <> Build.decimal (denominator r)
