{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE OverloadedStrings #-}

module StatesFile
  ( statesFileP
  , parseVarIndices
  , variableIndices
  , computePartition
  , Partition(..)
  ) where

import           Control.Monad
import           Data.Bifunctor (first)

import           Data.Text (Text)
import qualified Data.Text as T
import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Vector.Mutable as VM
import           Text.Megaparsec
import           Data.Set (Set)
import qualified Data.Set as S
import           Data.Map (Map)
import qualified Data.Map as M

import           Parser

----------------------------------------------------------------------
-- Types
----------------------------------------------------------------------

data Value
  = Number Int
  | String Text
  deriving (Eq, Ord)

data StatesFile = StatesFile
  { variables :: Vector Text
  , states :: Vector (Vector Value)
  }

----------------------------------------------------------------------
-- Parser
----------------------------------------------------------------------

statesFileP :: Parser StatesFile
statesFileP = StatesFile <$> variablesP <*> statesP

variablesP :: Parser (Vector Text)
variablesP = V.fromList <$> parens (nameP `sepBy` symbolP ",")

statesP :: Parser (Vector (Vector Value))
statesP = V.fromList <$> (many stateP)

stateP :: Parser (Vector Value)
stateP =
  V.fromList <$>
  (decimalP *> symbolP ":" *> parens (valueP `sepBy` symbolP ","))

valueP :: Parser Value
valueP = (Number <$> decimalP) <|> (String <$> nameP)

parens :: Parser a -> Parser a
parens = between (symbolP "(") (symbolP ")")

varIndicesP :: Parser (Set Text)
varIndicesP = S.fromList <$> nameP `sepBy` symbolP ","

parseVarIndices :: String -> Either String (Set Text)
parseVarIndices input =
  first errorBundlePretty (parse (varIndicesP <* eof) input (T.pack input))

----------------------------------------------------------------------
-- Partitioning
----------------------------------------------------------------------

-- | Given a set of variable names, return the indices of those variables in the
-- states file or an error in case one or more variables aren't defined.
variableIndices :: StatesFile -> Set Text -> Either Text (Set Int)
variableIndices sf vars = S.fromList <$> (mapM lookupVar (S.toList vars))
  where
    lookupVar :: Text -> Either Text Int
    lookupVar var =
      maybe
        (Left ("Variable " <> var <> " not found in the states file."))
        Right
        (V.elemIndex var (variables sf))

-- assigns each states the index of it's block
data Partition = Partition
  { numBlocks :: Int
  , stateAssignment :: (Vector Int)
  }

-- | Given a states file and the indices of the variables to compare, return a partition.
computePartition :: StatesFile -> Set Int -> Partition
computePartition sf vars
 =
      -- vector version of @vars@
  let varVector :: Vector Int = V.fromList (S.toAscList vars)
      -- map from a vector of variable values to a list of each state having those exact values.
      stateMap :: Map (Vector Value) [Int] =
        V.ifoldl'
          (\m i vals -> M.insertWith (++) (vecIndices vals varVector) [i] m)
          M.empty
          (states sf)
      -- Vector of disjunct lists of states
      blocks :: Vector [Int] = V.fromList (M.elems stateMap)
   in Partition (length blocks) $
      V.create $ do
        v <- VM.new (length (states sf))
        iforM_ blocks $ \block states -> forM_ states $ \state -> VM.write v state block
        return v


-- | Given an original vector v of length @n@ and a vector of indices
-- @[x_1,...,x_m]@ of length @m@, return the vector @[v[x_1],...,v[x_m]]@ of
-- length @m@.
vecIndices :: Vector a -- ^ Original vector
           -> Vector Int -- ^ Index vector
           -> Vector a
vecIndices orig indices = V.map (orig V.!) indices

iforM_ :: Monad m => Vector a -> (Int -> a -> m b) -> m ()
iforM_ = flip V.imapM_
