{-# LANGUAGE TupleSections #-}
{-# LANGUAGE TemplateHaskell #-}
module Main (main) where

import           Control.Applicative
import           Control.Monad                  ( forM_ )
import           Data.Maybe                     ( listToMaybe )
import qualified Data.List.NonEmpty            as NE
import           Data.Word

import qualified Data.Text.Lazy.IO             as TL
import qualified Data.Text.Lazy.Builder        as Build
import qualified Data.Vector                   as V
import qualified Data.Vector.Unboxed           as VU
import           System.Random
import qualified Options.Applicative           as OptParse
import           Lens.Micro.TH
import           Lens.Micro.Platform

import           Type
import           DotOutput
import           CoalgOutput
import           ValmariOutput


-- | Generate a random DFA
randomDFA
  :: Word32 -- ^ Number of states
  -> Word32 -- ^ Size of the alphabet
  -> IO Bool -- ^ Random finality
  -> IO DFA
randomDFA s a randomIsFinal = do
  finals <- VU.replicateM (fromIntegral s) randomIsFinal
  matrix <- V.replicateM
    (fromIntegral s)
    (VU.replicateM (fromIntegral a) (randomRIO (0, s - 1)))
  return DFA { dfaStates      = s
             , dfaLetters     = a
             , dfaIsFinal     = finals
             , dfaTransitions = matrix
             }


percentTrue :: Int -> IO Bool
percentTrue percent = (< percent) <$> randomRIO (0, 99)

data OutputFormat = Dot | Copar | Valmari

instance Show OutputFormat where
  show Dot     = "dot"
  show Copar   = "copar"
  show Valmari = "valmari"


data Options = Options
  { _optStates :: Word32
  , _optLetters :: Word32
  , _optFinal :: Int
  , _optOutputs :: [(OutputFormat, Maybe (NE.NonEmpty Char))]
  } deriving (Show)
makeLensesWith abbreviatedFields ''Options


options :: OptParse.Parser Options
options =
  Options
    <$> OptParse.argument
          OptParse.auto
          (OptParse.metavar "N" <> OptParse.help "Number of states")
    <*> OptParse.argument
          OptParse.auto
          (OptParse.metavar "M" <> OptParse.help "Size of the alphabet")
    <*> OptParse.option
          OptParse.auto
          (  OptParse.long "final"
          <> OptParse.metavar "PERCENT"
          <> OptParse.value 50
          <> OptParse.help "Percent of final states"
          )
    <*> some
          (OptParse.argument
            readOutput
            (  OptParse.metavar "FORMAT:[OUTPUT_FILE]..."
            <> OptParse.help "FORMAT can be 'dot', 'copar' or 'valmari'"
            )
          )


readFormat :: OptParse.ReadM OutputFormat
readFormat = OptParse.maybeReader $ \case
  "copar" -> Just Copar
  "dot"   -> Just Dot
  "valmari" -> Just Valmari
  _       -> Nothing


readOutput :: OptParse.ReadM (OutputFormat, Maybe (NE.NonEmpty Char))
readOutput = OptParse.eitherReader $ \s -> case span (/= sep) s of
  ([]    , _ ) -> Left "a"
  (_     , []) -> Left "b"
  (format, _ : file) -> (, NE.nonEmpty file) <$> case format of
    "copar" -> Right Copar
    "dot"   -> Right Dot
    "valmari" -> Right Valmari
    _       -> Left "c"
  where sep = ':'


main :: IO ()
main = do
  opts <- OptParse.execParser
    (OptParse.info
      (options <**> OptParse.helper)
      (  OptParse.fullDesc
      <> OptParse.progDesc "Generate a random determinstic finite automaton"
      )
    )

  dfa <- randomDFA (opts ^. states)
                   (opts ^. letters)
                   (percentTrue (opts ^. final))

  forM_ (opts ^. outputs) $ \(outputFormat, outputFile) -> do
    let builder = case outputFormat of
          Copar   -> coalgB
          Dot     -> dotB
          Valmari -> valmariB

    let text = Build.toLazyText (builder dfa)
    case outputFile of
      Nothing       -> TL.putStr text
      Just filename -> TL.writeFile (NE.toList filename) text

    return ()
