{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE ExistentialQuantification #-}

module Main (main) where

import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import qualified Data.Text.Lazy.IO             as T
import qualified Data.Text.Lazy.Builder        as Build
import qualified Options.Applicative           as Options
import qualified Text.Megaparsec               as Mega
import qualified Text.Megaparsec.Char          as Mega
import qualified Text.Megaparsec.Char.Lexer    as Mega
import           Data.Void
import           System.Random
import           Numeric
import           System.IO
import           System.Exit
import           Control.Applicative
import           Data.Text.Prettyprint.Doc
import           Data.Text.Prettyprint.Doc.Util
import           Data.Text.Prettyprint.Doc.Render.Terminal
import           Data.Text.Prettyprint.Convert.AnsiWlPprint

import           Types
import           Generator
import           Output
import           Hoegberg
import           Probability

data SomeMonoid = forall m. SomeMonoid (MonoidType m)


data Opts = Opts
  { optMonoid :: SomeMonoid
  , optStates :: Int
  , optSymbols :: SymbolSpec
  , optEdgeConfig :: EdgeConfig
  , optRandomState :: Maybe StdGen
  , optDifferentValues :: Maybe Int
  , optHoegberg :: Bool
  }

readMonoid :: Options.ReadM SomeMonoid
readMonoid = Options.eitherReader $ \case
  "Z,max"    -> Right (SomeMonoid MaxInt)
  "Word,or"  -> Right (SomeMonoid OrWord)
  "powerset" -> Right (SomeMonoid Powerset)
  _ ->
    Left
      $  "Unknown monoid type. Valid values are '"
      <> "Z,max"
      <> "', '"
      <> "Word,or"
      <> "' and '"
      <> "powerset"
      <> "'"

readSymbols :: Options.ReadM SymbolSpec
readSymbols = Options.maybeReader (Mega.parseMaybe parser)
  where
    parser :: Mega.Parsec Void String (Vector Int)
    parser = V.fromList <$> Mega.decimal `Mega.sepBy` Mega.string ","

readCount :: String -> Either String Int
readCount input = case readDec input of
  [(0, "")] -> Left "Count must be >0"
  [(x, "")] -> Right x
  _ -> Left "Count not parse Number"

parseOpts :: Options.Parser Opts
parseOpts =
  Opts
    <$> Options.option
          readMonoid
          (Options.long "monoid" <> Options.metavar "MONOID" <> Options.help
            (  "Monoid to use. Valid valies are '"
            <> "Z,max"
            <> "', '"
            <> "Word,or"
            <> "' and '"
            <> "powerset"
            <> "'"
            )
          )
    <*> Options.option
          Options.auto
          (Options.long "states" <> Options.metavar "NUM" <> Options.help
            "Number of states to generate"
          )
    <*> Options.option
          readSymbols
          (  Options.long "symbols"
          <> Options.metavar "NUM,NUM,..."
          <> Options.help
               (  "Comma separated list of symbols per arity."
               <> " E.g. 2,0,1 means two symbols with arity 0,"
               <> "non with arity 1 and one with arity two"
               )
          )
    <*> parseEdgeConfig
    <*> Options.optional
          (Options.option Options.auto (Options.long "random-state"))
    <*> Options.optional
          (Options.option
            (Options.eitherReader readCount)
            (  Options.long "different-values"
            <> Options.metavar "NUM"
            <> Options.help
                 ("Maximal number of differnt monoid values to generate."
                 <> "This is useful to limit the number of blocks in the initial partition."
                 )
            )
          )
    <*> Options.switch
          (Options.long "hoegberg" <> Options.help
            ("Generate output in format suiteable for Johanna Högbergs Java implementation."
            <> " See https://people.cs.umu.se/~johanna/bisimulation/"
            )
          )

parseEdgeConfig :: Options.Parser EdgeConfig
parseEdgeConfig =
  (ZeroFrequency <$> Options.option
      (Options.eitherReader readProbability)
      (  Options.long "zero-frequency"
      <> Options.showDefault
      <> Options.value 0.7
      <> Options.metavar "FREQ"
      <> Options.help
           "Frequency of edges with zero weight as number between 0 and 1."
      )
    )
    <|> (NumTransitions <$> Options.option
          Options.auto
          (Options.long "transitions" <> Options.metavar "NUM" <> Options.help
            (  "Number of transitions to generate."
            <> "They will be distributed randomly over states."
            )
          )
        )

withSpec :: Opts -> (forall m . WTASpec m -> x) -> x
withSpec opts f = case optMonoid opts of
  SomeMonoid m -> f WTASpec { monoid     = m
                            , numStates  = optStates opts
                            , numSymbols = optSymbols opts
                            }


example :: Doc AnsiStyle
example =
  "Example:"
    <> line
    <> line
    <> "    random-wta --monoid Z,max --states 3 --symbols 1,2,3"
    <> line
    <> line
    <> reflow "This generates a WTA with three states for the functor"
    <> softline
    <> "Z×X^(1 + 2×X + 3×X^2)"
    <> softline
    <> reflow "with roughly 30% of all possible edges, chosen at random."

    <> softline
    <> reflow "To control the number of edges, use --zero-frequency"
    <> softline
    <> reflow "or --transitions."

main :: IO ()
main = do
  let optSpec = Options.info
        (parseOpts Options.<**> Options.helper)
        (  Options.fullDesc
        <> Options.header "Generator for random weighted tree automata"
        <> Options.footerDoc (Just (toAnsiWlPprint example))
        )
  opts <- Options.execParser optSpec

  case optMonoid opts of
    SomeMonoid Powerset | maybe False (> 2) (optDifferentValues opts) -> do
      hPutStrLn stderr "error: Powerset can't have more than 2 different values"
      exitFailure
    _ -> return ()

  mapM_ setStdGen (optRandomState opts)

  withSpec opts $ \spec -> do
    randGen <- getStdGen
    wta     <- runGenerator
      (GeneratorConfig spec (optEdgeConfig opts) (optDifferentValues opts))
      genWTA
    putStrLn $ "# Random state for this automaton: '" <> show randGen <> "'"
    case monoid spec of
      Powerset | optHoegberg opts ->
        T.putStr (Build.toLazyText (hoegbergWTA wta))
      _ | optHoegberg opts ->
        hPutStrLn stderr "error: Hoegberg output only supports powerset"
      _ -> T.putStr (Build.toLazyText (buildWTA wta))
