{-# LANGUAGE GADTs #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE LambdaCase #-}

-- | Genator for random weighted tree automata.
module Generator
  ( genWTA
  , runGenerator
  , GeneratorConfig(..)
  , EdgeConfig(..)
  , uniqueTransitions
  )
where

import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import           System.Random
import           Control.Monad.Reader
import           Data.Coerce
import           Data.Maybe
import           Data.Foldable
import           Control.Arrow                  ( (&&&) )
import qualified Data.Map.Strict               as M
import qualified Data.Set                      as S
import           Data.Ratio
import           System.IO

import           Types                   hiding ( spec )
import           Probability
import           IndexedTransition              ( IndexedTransition )
import qualified IndexedTransition


-- | Decides how many non-zero transitions are generated.
data EdgeConfig
     -- | Generate zero transitions with a given probability.
     = ZeroFrequency Probability
     -- | Generate a fixed number of non-zero transitions.
     | NumTransitions Int


-- Configuration for the automaton generator.
data GeneratorConfig m = GeneratorConfig
   { spec :: WTASpec m -- ^ The automaton to generate.
   , zeroPolicy :: EdgeConfig -- ^ How many edges to generate.
   , differentValues :: Maybe Int -- ^ How many different monoid values to generate.
   }


type Generator m = ReaderT (GeneratorConfig m) IO


-- | Acutally run the generator with the given config.
runGenerator :: GeneratorConfig m -> Generator m a -> IO a
runGenerator config action = runReaderT action config


zeroFreq :: GeneratorConfig m -> Probability
zeroFreq GeneratorConfig { zeroPolicy = ZeroFrequency p } = p
zeroFreq _ = error "zeroFreq: unexpected out degree" -- Ugly as hell

genMonoidValue :: Generator m m
genMonoidValue = asks ((monoid . spec) &&& differentValues) >>= \case
  (Powerset, Nothing) -> liftIO randomIO
  (Powerset, Just 1 ) -> return True
  (Powerset, Just 2 ) -> liftIO randomIO
  (Powerset, _      ) -> error "differentValues >2 not supported for powerset"
  (OrWord  , Nothing) -> liftIO randomIO
  (OrWord  , Just x ) -> liftIO $ randomRIO (1, fromIntegral x)
  (MaxInt  , Nothing) -> liftIO randomIO
  (MaxInt  , Just x ) -> liftIO $ randomRIO (1, x)

genStates :: Generator m (Vector m)
genStates = do
  n <- asks (numStates . spec)
  V.replicateM n genMonoidValue

aritySummand :: Int -> Generator m (Maybe Int)
aritySummand arity = do
  arities <- asks (numSymbols . spec)
  if V.length (V.filter (/=0) arities) == 1
    then return Nothing -- only one summand. => no injections necessary
    else return $ Just (V.length (V.filter (/= 0) (V.take arity arities)))

decideZero :: Generator m Bool
decideZero = asks zeroFreq >>= liftIO . decide

-- Generates Nothing, when it decides that a zero value would be in order
genTransition :: Int -> Int -> [State] -> Generator m (Maybe (Transition m))
genTransition arity symbol succs = decideZero >>= \case
  True -> return Nothing
  False ->
    fmap Just
      $   Transition
      <$> genMonoidValue
      <*> aritySummand arity
      <*> return symbol
      <*> return (V.fromList succs)

genForSymbol :: Int -> Int -> Generator m (Vector (Transition m))
genForSymbol arity symbol = do
  states <- asks (numStates . spec)
  V.fromList . catMaybes <$> traverse
    (genTransition arity symbol)
    (replicateM arity (coerce [0 .. states - 1]))

genForArity :: Int -> Generator m (Vector (Transition m))
genForArity arity = do
  n <- asks ((V.! arity) . numSymbols . spec)
  fold <$> V.generateM n (genForSymbol arity)

genStateTransitions :: Generator m (Vector (Transition m))
genStateTransitions = do
  arities <- asks (numSymbols . spec)
  fold <$> traverse genForArity (V.findIndices (/= 0) arities)

genTransitionsZeroFreq :: Generator m (Vector (Vector (Transition m)))
genTransitionsZeroFreq = do
  n <- asks (numStates . spec)
  V.replicateM n genStateTransitions

uniqueTransitions :: Int -> IndexedTransition -> IO [IndexedTransition]
uniqueTransitions num idxMax@(IndexedTransition.Index max)
  | fromIntegral num < fromIntegral max * ((7 :: Integer) % 10) = uniqueTransitionsByGeneration
    num
    idxMax
  | otherwise = uniqueTransitionsByElimination num idxMax

uniqueTransitionsByGeneration
  :: Int -> IndexedTransition -> IO [IndexedTransition]
uniqueTransitionsByGeneration num max = helper
  S.empty
  num
  where
    helper m 0 = return $ S.toList m
    helper m c = do
      x <- randomRIO (0, max - 1)
      if x `S.member` m then helper m c else helper (S.insert x m) (c - 1)

uniqueTransitionsByElimination
  :: Int -> IndexedTransition -> IO [IndexedTransition]
uniqueTransitionsByElimination num max = helper
  whole
  num
  where
    helper free 0 = return $ S.toList (S.difference whole free)
    helper free c = do
      idx <- randomRIO (0, S.size free - 1)
      let x = S.elemAt idx free
      helper (S.delete x free) (c - 1)

    whole = S.fromList [0 .. max - 1]

genTransitionsNumTrans :: Int -> Generator m (Vector (Vector (Transition m)))
genTransitionsNumTrans numTransitions = do
  wtaSpec <- asks spec
  let n    = numStates wtaSpec
      maxT = IndexedTransition.maxIndex wtaSpec

  numTransitions' <- if fromIntegral numTransitions > maxT
    then do
      let cap = IndexedTransition.fromIndexdTransition maxT
      lift $ hPutStrLn
        stderr
        (  "warning: More transitions than possible requested. Capping at "
        <> show cap
        )
      return (fromIntegral cap)
    else return numTransitions

  transitions <-
    lift
    $   map (IndexedTransition.fromIndex wtaSpec)
    <$> uniqueTransitions numTransitions' maxT
  weightedTransitions <- (traverse . traverse . traverse)
    (const genMonoidValue)
    transitions

  let byState = foldl' (\m (State s, t) -> M.insertWith (++) s [t] m)
                       M.empty
                       weightedTransitions

  let stateVec = V.generate n $ \i -> case M.lookup i byState of
        Nothing  -> V.empty
        Just lst -> V.fromList lst

  return stateVec

genWTA :: Generator m (WTA m)
genWTA = asks zeroPolicy >>= \case
  NumTransitions d ->
    WTA <$> asks spec <*> genStates <*> genTransitionsNumTrans d
  ZeroFrequency _ -> WTA <$> asks spec <*> genStates <*> genTransitionsZeroFreq
