{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- | A bijection from transitions to integers.
--
-- This module provides functions to convert transitions to and from continuous
-- integer indices without loss of information (for a given WTA).
--
-- In particular, this allows for easy enumeration of all transitions and
-- enables choosing a small set of transitions from a potentially very large set
-- of /all/ possible transitions for a given WTA.
--
-- Transitions are ordered by
--
-- 1. Source state
-- 2. Arity
-- 3. Symbol with that arity
-- 4. Successor states in lexicographic order
--
-- Transition indices are then simply the indices into the list of all
-- transitions for the given WTA, ordered by the above criteria.
module IndexedTransition
  ( IndexedTransition(..)
  , maxIndex
  , fromIndex
  , index
  )
where

import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import           Data.Maybe
import           Data.Tuple
import           System.Random

import           Types

-- | Index into the set of all transitions for a given WTA
newtype IndexedTransition = Index { fromIndexdTransition :: Integer }
  deriving newtype (Num, Eq, Ord, Random, Enum)
  deriving (Show)

-- | Return __one more than__ the maximum transition index for a given WTA.
maxIndex :: WTASpec m -> IndexedTransition
maxIndex spec =
  let n      = numStates spec
      (t, _) = transitionsPerState spec
  in  Index (fromIntegral n * t)

-- | Convert the given index to its corresponding transition.
fromIndex :: WTASpec m -> IndexedTransition -> (State, Transition ())
fromIndex spec (Index i) =
  let n     = numStates spec
      (t    , symbolSums     ) = transitionsPerState spec

      (state, stateTransition) = i `divMod` t

      -- fromJust is justified (ho ho) here, since `stateTransition` should
      -- never be greater than the total number of possible transitions for this
      -- state (which is the last value in symbolSums).
      arity = fromJust (V.findIndex (> stateTransition) symbolSums) - 1

      arityTransition = stateTransition - (symbolSums V.! arity)

      symbolBounds :: Vector Int =
          V.cons (numSymbols spec V.! arity) (V.replicate arity n)
      symbolDigits = decodeFromInt symbolBounds arityTransition
      symbol     = V.head symbolDigits
      successors = V.tail symbolDigits


      trans      = Transition { weight     = ()
                              , summand    = aritySummand spec arity
                              , symbol     = symbol
                              , successors = V.map State successors
                              }
  in  (State (fromIntegral state), trans)

-- | Convert the given transition to its corresponding index.
index :: WTASpec m1 -> State -> Transition m2 -> IndexedTransition
index spec (State state) trans =
  let (t, symbolSums) = transitionsPerState spec

      arity :: Int    = summandArity spec (summand trans)
      symbolBounds :: Vector Int = V.cons
        (numSymbols spec V.! arity)
        (V.replicate arity (numStates spec))
      arityIdx :: Integer = encodeAsInt
        symbolBounds
        (V.cons (symbol trans) (V.map fromState $ successors trans))
      stateLocal :: Integer = symbolSums V.! arity + arityIdx
  in  Index $ (fromIntegral state) * t + stateLocal

-- Helpers

-- | Convert a given summand number to its arity
summandArity :: WTASpec m -> Maybe Int -> Int
summandArity spec summand =
  V.findIndices (/= 0) (numSymbols spec) V.! (fromMaybe 0 summand)

-- | Convert a given arity to its summand index
aritySummand :: WTASpec m -> Int -> Maybe Int
aritySummand spec arity =
  let arities = numSymbols spec
  in if V.length (V.filter (/= 0) arities) == 1
     then Nothing
     else Just $ V.length (V.filter (/= 0) (V.take arity arities))

-- | Return the maximum number of transitions per state for this WTA as well as
-- a vector that contains at index @i@ the maximum number of transitions per
-- state with arity @i@ or less.
--
-- Therefore,
--
-- @
--  fst (transitionsPerState w) == last (snd (transitionsPerState w))
-- @
transitionsPerState :: WTASpec m -> (Integer, Vector Integer)
transitionsPerState spec =
  let n :: Integer = fromIntegral $ numStates spec
      tPerSymbol =
          (V.imap (\i syms -> fromIntegral syms * n ^ i) (numSymbols spec))
      runningTotal = V.scanl' (+) 0 tPerSymbol
  in  (V.last runningTotal, runningTotal)

encodeAsInt :: Vector Int -> Vector Int -> Integer
encodeAsInt maxBounds digits =
  let factors = V.prescanr' (*) 1 (fmap fromIntegral maxBounds)
  in  sum (V.zipWith (*) factors (fmap fromIntegral digits))

decodeFromInt :: Vector Int -> Integer -> Vector Int
decodeFromInt maxBounds encoded = V.map (fromIntegral . fst)
  $ V.postscanr' doDigit (0, encoded) (fmap fromIntegral maxBounds)
  where doDigit bound (_, current) = (swap $ current `divMod` bound)
