{-# LANGUAGE GADTs #-}

-- | Printer for CoPaR's input format, i.e. coalgebra specifications.
module Output (buildWTA) where

import qualified Data.Vector                   as V
import           Data.Vector                    ( Vector )
import           Data.Text.Lazy.Builder         ( Builder )
import qualified Data.Text.Lazy.Builder.Int    as Build
import           Data.Foldable
import           Data.List
import           Data.String
import           Data.Maybe

import           Types

wtaFunctor :: WTASpec m -> Builder
wtaFunctor wta = monoidForStates (monoid wta) <> " × " <> withMonoidForWeights
  (monoid wta)
  (polynomial (numSymbols wta))

monoidForStates :: MonoidType m -> Builder
monoidForStates MaxInt   = "Z"
monoidForStates OrWord   = "N"
monoidForStates Powerset = "2"

withMonoidForWeights :: MonoidType m -> Builder -> Builder
withMonoidForWeights MaxInt   inner = "(Z, max)^(" <> inner <> ")"
withMonoidForWeights OrWord   inner = "(Word, or)^(" <> inner <> ")"
withMonoidForWeights Powerset inner = "P(" <> inner <> ")"

polynomial :: SymbolSpec -> Builder
polynomial = sepList " + " id . catMaybes . toList . V.imap summand
  where
    summand :: Int -> Int -> Maybe Builder
    summand _     0    = Nothing
    summand 0     syms = Just (Build.decimal syms)
    summand arity syms =
      Just $ fold (intersperse "×" (Build.decimal syms : replicate arity "X"))

data ValueUse = Weight | StateVal

buildValue :: ValueUse -> MonoidType m -> m -> Builder
buildValue _      MaxInt   i = Build.decimal i
buildValue Weight OrWord   w = "0x" <> Build.hexadecimal w
buildValue StateVal OrWord w = Build.decimal w
buildValue _      Powerset b = if b then "1" else "0"

buildTransition :: MonoidType m -> Transition m -> Builder
buildTransition mon trans =
  inject (summand trans)
    <> buildSuccs (successors trans)
    <> buildWeight mon (weight trans)

  where
    inject Nothing = ""
    inject (Just i) = "inj" <> Build.decimal i <> " "
    buildSuccs succs = if V.null succs
      then Build.decimal (symbol trans)
      else
        "("
        <> Build.decimal (symbol trans)
        <> ", "
        <> sepList ", " buildStateName (successors trans)
        <> ")"

buildWeight :: MonoidType m -> m -> Builder
buildWeight Powerset _   = ""
buildWeight mon      val = ": " <> buildValue Weight mon val

buildTransitions :: MonoidType m -> Vector (Transition m) -> Builder
buildTransitions mon trans =
  "{ " <> sepList ", " (buildTransition mon) trans <> " }"

buildState :: WTASpec m -> (Int, m, Vector (Transition m)) -> Builder
buildState wtaSpec (state, value, trans) =
  buildStateName (State state)
    <> ": ("
    <> buildValue StateVal (monoid wtaSpec) value
    <> ", "
    <> buildTransitions (monoid wtaSpec) trans
    <> ")\n"

buildStateName :: State -> Builder
buildStateName = ("s" <>) . Build.decimal

buildStates :: WTA m -> Builder
buildStates wta = foldMap
  (buildState (spec wta))
  (V.zip3 (V.fromList indices) (stateValue wta) (stateTransitions wta))
  where indices = [0 .. V.length (stateValue wta) - 1]


-- | Print a WTA as coalgebra specification suitable for CoPaR.
buildWTA :: WTA m -> Builder
buildWTA a = wtaFunctor (spec a) <> "\n" <> buildStates a


-- helpers

sepList :: (Foldable f, Monoid m, IsString m) => m -> (a -> m) -> f a -> m
sepList m action = fold . intersperse m . map action . toList
