{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- | Arbitrary precision floating point numbers between [0,1] that can be used
-- as a probability value.
module Probability (Probability(..), readProbability, decide) where

import           System.Random
import           Data.Scientific


-- | Arbitrary precision floating point numbers between [0,1] that can be used
-- as a probability value.
newtype Probability = Probability Scientific
  deriving newtype (Show,Fractional,Num)


-- | Read an arbitrary precision floating point number from a string. Valid
-- formats are:
--
-- - 0
-- - 1
-- - 1.0
-- - 0.[0..9]*
--
-- which means that leading zeros are not allowed. in particular, @00@, @1.@ and
-- similar verboten for simplicity.
readProbability :: String -> Either String Probability
readProbability input = Probability <$> case input of
  "0" -> Right (scientific 0 0)
  ('0' : '.' : rest) -> case reads rest of
    [(digits, "")] -> Right (scientific digits (negate (length rest)))
    _ -> failure
  "1" -> Right (scientific 1 0)
  ('1' : '.' : rest) | all (== '0') rest -> Right (scientific 1 0)
                     | otherwise -> failure
  _ -> failure
  where failure = Left "Could not parse probability"


-- | Return 'True' randomly with the given probability and 'False' otherwise.
--
-- Uses the global random number generator.
decide :: Probability -> IO Bool
decide (Probability science) = do
  let digits = coefficient science
  let exp = negate (base10Exponent science)
  randomNumber <- randomRIO (0, (10 ^ exp) - 1)
  return $ randomNumber < digits
