{-# LANGUAGE GADTs #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE DeriveFunctor #-}
{-# LANGUAGE DeriveFoldable #-}
{-# LANGUAGE DeriveTraversable #-}
{-# LANGUAGE StandaloneDeriving #-}

module Types
  ( MonoidType(..)
  , WTASpec(..)
  , Transition(..)
  , WTA(..)
  , SymbolSpec
  , State(..)
  )
where

import           Data.Vector                    ( Vector )

-- | Supported monoids
--
-- Note that powerset is not really the monoid, it's ({True,False}, or)^X
data MonoidType m where
  MaxInt :: MonoidType Int
  OrWord :: MonoidType Word
  Powerset :: MonoidType Bool

deriving instance Show (MonoidType m)

-- | For each arity, the number of symbols
type SymbolSpec = Vector Int

-- | Type and size of an automaton.
data WTASpec m = WTASpec
  { monoid :: MonoidType m
  -- | Number of states
  , numStates :: Int
  -- | Number of symbols per arity
  , numSymbols :: SymbolSpec
  }

deriving instance Show m => Show (WTASpec m)

-- | A state in the automaton
newtype State = State { fromState :: Int}
  deriving (Num,Show,Eq,Integral,Enum,Real,Ord)

-- | A transition in the automaton. Note that the source state is not part of
-- this record, since it is implicit in the arry 'stateTransitions' of the 'WTA'
-- type.
data Transition m = Transition
  { weight :: m
  -- | 'summand' is not the arity, because arities with zero symbols don't have
  -- summands. This means that @X^2@ will only have summand 0, but arity 0,1 and
  -- 2. Can be 'Nothing' as an optimization if there is only one summand, as in
  -- this case we don't have to generate coproduct injections.
  , summand :: Maybe Int
  -- | Symbol number
  , symbol :: Int
  , successors :: Vector State
  } deriving (Show, Functor, Foldable, Traversable)

-- | A weighted tree automaton.
data WTA m = WTA
  { spec :: WTASpec m
  -- | Each state has a value from the respective monoid associated with it. The
  -- 'output' value.
  , stateValue :: Vector m
  -- | Outgoing transitions per state.
  , stateTransitions :: Vector (Vector (Transition m))
  }

deriving instance (Show m) => Show (WTA m)
