module Copar.Algorithm.InitializeSpec (spec) where

import           Test.Hspec

import           Prelude hiding (pred)
import           Control.Monad.ST
import           Data.STRef

import qualified Data.Vector as V

import           Data.MorphismEncoding (Encoding, State, EdgeRef(..))
import qualified Data.MorphismEncoding as Encoding
import           Copar.Algorithm.Initialize
import           Copar.Functors.Powerset
import           Copar.Algorithm.Types
import           Data.Partition.Common

spec :: Spec
spec = do
  blocksSpec
  toSubSpec
  predSpec
  lastWSpec


blocksSpec :: Spec
blocksSpec = describe "returned block list" $ do
  it "works correctly when no state has successors" $
    runST (fst <$> initialize' [False, False] [(0, 1)]) `shouldBe` [0]

  it "works correctly when all states have successors" $
    runST (fst <$> initialize' [True, True] [(0, 1), (1, 0)]) `shouldBe` [0]

  it "works correctly with mixed F1s" $
    runST (fst <$> initialize' [True, False] [(0, 1)]) `shouldMatchList` [0, 1]


toSubSpec :: Spec
toSubSpec = describe "returned toSub vector" $ do
  let getSub = getState (fmap V.toList . V.freeze . toSub)

  it "is empty for states that have no successors" $
    getSub [False, False] [] `shouldMatchList` [[], []]

  it "is empty for states that have successors" $
    getSub [True, True] [(0, 1), (1, 0)] `shouldMatchList` [[], []]


predSpec :: Spec
predSpec = describe "returned pred vector" $ do
  let getPred = getState (return . V.toList . pred)

  it "works for states that have no predecessors" $
    getPred [False, False] [] `shouldMatchList` [[], []]

  it "works for states that no predecessors" $
    getPred [True, True] [(0, 1), (1, 1)] `shouldMatchList`
    [[], [EdgeRef 1, EdgeRef 0]]


lastWSpec :: Spec
lastWSpec = describe "returned lastW vector" $ do
  let getLastW = getState $ \s -> do
        v <- V.freeze (lastW s)
        mapM readSTRef (V.toList v)

  it "works without edges" $
    getLastW [False, False] [] `shouldBe` []

  it "works with some edges" $
    getLastW [True, True] [(0, 1), (0, 0), (1, 0)]
      `shouldBe` [packWeight False 2, packWeight False 2, packWeight False 1]


enc :: [f1] -> [(State, label, State)] -> Encoding label f1
enc f1s transitions = Encoding.new (V.fromList f1s) edges
  where
    edges = V.fromList (map mkEdge transitions)
    mkEdge (from, lab, to) = Encoding.Edge from lab to

initialize' :: [Bool] -> [(State, State)] -> ST s ([Block], AlgoState s Powerset)
initialize' f1s transitions =
  initialize @Powerset (enc f1s (map mkEdge transitions)) True
  where
    mkEdge (from, to) = (from, (), to)

getState :: (forall s. AlgoState s Powerset -> ST s a) -> [Bool] -> [(State, State)] -> a
getState projection f1s transitions = runST $ do
  res <- snd <$> initialize' f1s transitions
  projection res
