{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
module Copar.Coalgebra.ParserSpec (spec) where

import           Test.Hspec
import           Test.Hspec.Megaparsec
import           TestHelpers

import           Data.Void
import           Control.Exception (evaluate)
import           Type.Reflection
import           Data.List (nub)

import qualified Data.Set as S
import qualified Data.Vector as V
import           Data.Text (Text)
import           Text.Megaparsec
import           Lens.Micro.Platform
import qualified Data.HashMap.Strict as HM

import           Data.MorphismEncoding (Encoding)
import qualified Data.MorphismEncoding as Encoding
import           Copar.Coalgebra.Parser
import           Copar.RefinementInterface
import           Copar.FunctorExpression.Sorts (Sort, Sorted(..))
import           Copar.FunctorExpression.Type
import           Copar.Parser.Lexer

spec :: Spec
spec = parseMorphismsSpec

parseMorphismsSpec :: Spec
parseMorphismsSpec = describe "parseMorphisms" $ do
  it "errors when the functor expression doesn't have a function symbol" $
    evaluate (parsing @Id Variable "") `shouldThrow` anyErrorCall

  it "fails to parse empty input" $
    parsing (Functor 1 (Id Variable)) `shouldFailOn` ""

  it "fails to parse garbage" $ do
    parsing (Functor 1 (Id Variable)) `shouldFailOn` "a"
    parsing (Functor 1 (Id Variable)) `shouldFailOn` "a:"
    parsing (Functor 1 (Id Variable)) `shouldFailOn` ":"

  it "fails on duplicate variable definitions" $
    parsing (Functor 1 (Id Variable)) `shouldFailOn` "a: a\na: a"

  it "fails on undefined variable references" $
    parsing (Functor 1 (Id Variable)) `shouldFailOn` "a: b"

  it "parses a minimal example" $
    parsing (Functor 1 (Id Variable)) "a: a" `shouldParse`
    encoding [(Sorted 1 ())] [(0, (Sorted 1 ()), 0)]

  it "parses a sligtly less minimal example" $
    parsing (Functor 1 (Id Variable)) "a: b\nb: a" `shouldParse`
    encoding
      [(Sorted 1 ()), (Sorted 1 ())]
      [(1, (Sorted 1 ()), 0), (0, (Sorted 1 ()), 1)]

  it "parses a powerset example" $
    parsing (Functor 1 (P Variable)) "a: [a, b]\nb: []" `shouldParse`
    encoding
      [(Sorted 1 False), (Sorted 1 True)]
      [(0, (Sorted 1 ()), 0), (0, (Sorted 1 ()), 1)]

  it "parses a fixed product example" $
    parsing (Functor 1 (FP ["l1", "l2"] Variable)) "a: (l1, b)\nb: (l2, a)" `shouldParse`
    encoding
      [(Sorted 1 "l1"), (Sorted 1 "l2")]
      [(1, (Sorted 1 ()), 0), (0, (Sorted 1 ()), 1)]

  it "handles mixed functors" $
    parsing (Functor 1 (SomeFunctor (FP ["l1", "l2"] (Functor 2 (SomeFunctor (P Variable))))))
    "a: (l1, [a, b])\nb: (l2, [])"
    `shouldParse`
    encoding [ (Sorted 1 (SomeF1 ("l1" :: Text)))
             , (Sorted 2 (SomeF1 False))
             , (Sorted 1 (SomeF1 ("l2" :: Text)))
             , (Sorted 2 (SomeF1 True )) ]
             [ (2, (Sorted 1 (SomeLabel ())), 3)
             , (0, (Sorted 1 (SomeLabel ())), 1)
             , (1, (Sorted 2 (SomeLabel ())), 0)
             , (1, (Sorted 2 (SomeLabel ())), 2)
             ]

  it "doesn't confuse sorts" $
    fmap (nub . V.toList . Encoding.structure)
    (parsing (Functor 1 (SomeFunctor (FP ["a"] (Functor 2 (SomeFunctor (FP ["a"] Variable))))))
    "x: (a, (a, x))")
    `shouldBe`
    Right [ (Sorted 1 (SomeF1 ("a" :: Text))), (Sorted 2 (SomeF1 ("a" :: Text))) ]

  context "the symbol table" $ do
    let p x = fromSymbolTable . fst <$> parseMorphisms (Functor 1 (P Variable)) EnableSanityChecks "" x

    it "contains the defined symbols" $ do
      (HM.elems <$> p "a: []\nb: []\nc: []") `shouldParse` ["a", "b", "c"]

    it "maps all symbols to different nodes" $ do
      (HM.keys <$> p "a: [b]\nb: [c]\nc: [d]\nd: [a]") `parseSatisfies` allDifferent


allDifferent :: Ord a => [a] -> Bool
allDifferent lst = length (S.fromList lst) == length lst

data Id a = Id a
  deriving (Functor)

type instance F1 Id = ()
type instance Label Id = ()

instance ParseMorphism Id where
  parseMorphismPoint (Id inner) = do
    suc <- inner
    return ((), (V.singleton (suc, ())))

data Powerset a = P a
  deriving (Functor)

type instance F1 Powerset = Bool
type instance Label Powerset = ()

instance ParseMorphism Powerset where
  parseMorphismPoint (P inner) = do
    succs <- V.fromList <$> brackets (inner `sepBy` comma)
    return (null succs, V.map (,()) succs)

data FixedProduct a = FP [Text] a
  deriving (Functor)

type instance F1 FixedProduct = Text
type instance Label FixedProduct = ()

instance ParseMorphism FixedProduct where
  parseMorphismPoint (FP labels inner) = do
    (f1, suc) <-
      parens ((,) <$> choice (map (try . symbol) labels) <*> (comma *> inner))
    return (f1, V.singleton (suc, ()))

data SomeFunctor a where
  SomeFunctor
    :: ( Functor f
       , ParseMorphism f
       , Show (Label f)
       , Show (F1 f)
       , Eq (Label f)
       , Eq (F1 f)
       , Typeable (Label f)
       , Typeable (F1 f)
       )
    => f a
    -> SomeFunctor a

deriving instance Functor SomeFunctor

data SomeLabel where
  SomeLabel
    :: forall l. (Show l, Eq l, Typeable l)
    => l
    -> SomeLabel

deriving instance Show SomeLabel

instance Eq SomeLabel where
  (SomeLabel (a :: ta)) == (SomeLabel (b :: tb)) =
    case eqTypeRep (typeRep @ta) (typeRep @tb) of
      Nothing -> False
      Just HRefl -> a == b

data SomeF1 where
  SomeF1 :: forall f. (Show f, Eq f, Typeable f) => f -> SomeF1

deriving instance Show SomeF1

instance Eq SomeF1 where
  (SomeF1 (a :: ta)) == (SomeF1 (b :: tb)) =
    case eqTypeRep (typeRep @ta) (typeRep @tb) of
      Nothing -> False
      Just HRefl -> a == b

type instance F1 SomeFunctor = SomeF1
type instance Label SomeFunctor = SomeLabel

instance ParseMorphism SomeFunctor where
  parseMorphismPoint (SomeFunctor f) = do
    convertOuter (parseMorphismPoint f)

    where
      convertOuter = fmap (\(f1, succs) -> (SomeF1 f1, fmap (_2 %~ SomeLabel) succs))

parsing :: (Functor f, ParseMorphism f) => FunctorExpression f Sort
  -> Text
  -> Either (ParseErrorBundle Text Void) (Encoding (Sorted (Label f)) (Sorted (F1 f)))
parsing expr = fmap snd . parseMorphisms expr EnableSanityChecks ""
