module Copar.FunctorExpression.ParserSpec (spec) where

import Test.Hspec
import Test.Hspec.Megaparsec

import Data.Functor.Classes
import Type.Reflection

import Data.Text (Text)

import Copar.FunctorExpression.Type
import Copar.FunctorExpression.Parser
import Copar.Parser.Lexer
import Copar.Parser.Types

spec :: Spec
spec = do
  parseFunctorExpressionSpec


parseFunctorExpressionSpec :: Spec
parseFunctorExpressionSpec = describe "parseFunctorExpression" $ do
  it "refuses an empty file" $
    parsing `shouldFailOn` ""

  it "parses a single prefix operator" $
    parsing "Pre X" `shouldParse` (Functor 2 (SomeFunctor (PrefixFunctor (Variable))))

  it "parses a single postfix operator" $
    parsing "X Post" `shouldParse`
    (Functor 1 (SomeFunctor (PostfixFunctor Variable)))

  it "parses nested operators" $ do
    parsing "Pre (Pre X)" `shouldParse`
      (Functor
         2
         (SomeFunctor
            (PrefixFunctor (Functor 2 (SomeFunctor (PrefixFunctor Variable))))))
    parsing "(X Post) Post" `shouldParse`
      (Functor
         1
         (SomeFunctor
            (PostfixFunctor (Functor 1 (SomeFunctor (PostfixFunctor Variable))))))
    parsing "Pre (X Post)" `shouldParse`
      (Functor
         2
         (SomeFunctor
            (PrefixFunctor (Functor 1 (SomeFunctor (PostfixFunctor Variable))))))

  it "fails to parse an expression without X" $
    parsing `shouldFailOn` "Pre"

  it "gets precedence right" $
    parsing "Pre X Post" `shouldParse`
    (Functor
       1
       (SomeFunctor
          (PostfixFunctor (Functor 2 (SomeFunctor (PrefixFunctor Variable))))))

  it "handles whitespace at the begining" $
    parsing `shouldSucceedOn` "  Pre X"

  it "parses Pre(X) with X in parenthesis" $
    parsing "Pre(X)" `shouldParse`
    (Functor 2 (SomeFunctor (PrefixFunctor (Variable))))


-- A simple prefix parser

data PrefixFunctor a = PrefixFunctor a
  deriving (Functor, Foldable, Traversable, Show)

instance Eq1 PrefixFunctor where
  liftEq f (PrefixFunctor x) (PrefixFunctor y) = f x y

instance Show1 PrefixFunctor where
  liftShowsPrec show' _ prec (PrefixFunctor x) rest
    | prec > 10 = "(PrefixFunctor " ++ show' 11 x (rest ++ ")")
    | otherwise = "PrefixFunctor " ++ show' 11 x rest

prefixParser :: FunctorParser PrefixFunctor
prefixParser = prefix (symbol "Pre" >> return PrefixFunctor)

-- A simple postfix parser

data PostfixFunctor a = PostfixFunctor a
  deriving (Functor, Foldable, Traversable, Show)

instance Eq1 PostfixFunctor where
  liftEq f (PostfixFunctor x) (PostfixFunctor y) = f x y

instance Show1 PostfixFunctor where
  liftShowsPrec show' _ prec (PostfixFunctor x) rest
    | prec > 10 = "(PostfixFunctor " ++ show' 11 x (rest ++ ")")
    | otherwise = "PostfixFunctor " ++ show' 11 x rest

postfixParser :: FunctorParser PostfixFunctor
postfixParser = postfix (symbol "Post" >> return PostfixFunctor)

-- List of functors

data SomeFunctor a where
  SomeFunctor :: (Eq1 f, Typeable f, Show1 f) => f a -> SomeFunctor a

instance Eq1 SomeFunctor where
  liftEq eq (SomeFunctor (f1 :: tf1 a)) (SomeFunctor (f2 :: tf2 b)) =
    case eqTypeRep (typeRep @tf1) (typeRep @tf2) of
      Nothing -> False -- different types
      Just HRefl -> liftEq eq f1 f2

instance Show1 SomeFunctor where
  liftShowsPrec show' showList' prec (SomeFunctor f) = liftShowsPrec show' showList' prec f

functors :: [FunctorParser SomeFunctor]
functors = [transParser SomeFunctor prefixParser, transParser SomeFunctor postfixParser]

-- Helpers

parsing :: Text -> Either ParseErr (FunctorExpression SomeFunctor Precedence)
parsing = parseFunctorExpression functors ""
