{-# OPTIONS_GHC -Wno-orphans #-}
module Copar.FunctorExpression.SortsSpec (spec) where

import           Test.Hspec

import           Data.Functor.Classes
import           Data.Functor.Identity
import           Data.Monoid
import           Type.Reflection

import qualified Data.Vector as V

import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Sorts
import           Copar.FunctorExpression.Type
import           Copar.FunctorExpression.Printable
import           Copar.RefinementInterface

import           Copar.Functors.SomeFunctor

spec :: Spec
spec = do
  annotateSortsSpec
  sortTableSpec

annotateSortsSpec :: Spec
annotateSortsSpec = describe "annotateSorts" $ do
  it "handles a single variable" $
    annotateSorts @Identity Variable `shouldBe` Variable

  it "correctly annotates an expression with one functor" $
    annotateSorts (Functor () (Identity Variable)) `shouldBe`
    (Functor 1 (Identity Variable))

  it "correctly annotates a nested expression" $
    annotateSorts (Functor () (Identity (Functor () (Identity Variable)))) `shouldBe`
    (Functor 1 (Identity (Functor 2 (Identity Variable))))

  it "correctly annotates an expression with different functors" $
    annotateSorts
      (Functor () (SomeFunctor (Identity (Functor () (SomeFunctor (Just Variable)))))) `shouldBe`
    (Functor 1 (SomeFunctor (Identity (Functor 2 (SomeFunctor (Just Variable))))))

  it "correctly annotates an expression with two argument functors" $
    annotateSorts
      (Functor
         ()
         (SomeFunctor
            (Twice
               (Functor () (SomeFunctor (Identity Variable)))
               (Functor () (SomeFunctor (Identity Variable)))))) `shouldBe`
    (Functor
       1
       (SomeFunctor
          (Twice
             (Functor 2 (SomeFunctor (Identity Variable)))
             (Functor 3 (SomeFunctor (Identity Variable))))))

sortTableSpec :: Spec
sortTableSpec = describe "sortTable" $ do
  it "works for a single variable" $
    sortTable @Identity Variable `shouldBe` V.empty

  it "works for a single functor" $ do
    sortTable (Functor 1 (Identity Variable)) `shouldBe` V.fromList [Identity ()]

  it "works for two functors" $ do
    sortTable (Functor 1 (Identity (Functor 2 (Identity Variable)))) `shouldBe`
      V.fromList [Identity (), Identity ()]

  it "works for functors with multiple arguments" $ do
    sortTable
      (Functor
         1
         (SomeFunctor
            (Twice
               (Functor 3 (SomeFunctor (Identity Variable)))
               (Functor 2 (SomeFunctor (Just Variable)))))) `shouldBe`
      V.fromList [ SomeFunctor (Twice () ())
                 , SomeFunctor (Just ())
                 , SomeFunctor (Identity ())
                 ]

instance Printable Identity where
  precedence _ = 0
  printMe cfg (Identity x) = withName cfg "Identity" <> x

type instance Label Identity = ()
type instance Weight Identity = ()
type instance F1 Identity = () 
type instance F3 Identity = ()
instance RefinementInterface Identity where
  init   = undefined
  update = undefined
instance ParseMorphism Identity where
  parseMorphismPoint = undefined

data Twice a = Twice a a
  deriving (Functor, Foldable, Traversable)

instance Eq1 Twice where
  liftEq f (Twice a1 b1) (Twice a2 b2) = f a1 a2 && f b1 b2

type instance Label Twice = ()
type instance Weight Twice = ()
type instance F1 Twice = () 
type instance F3 Twice = () 
instance RefinementInterface Twice where
  init   = undefined
  update = undefined
instance ParseMorphism Twice where
  parseMorphismPoint = undefined
instance Show1 Twice where
  liftShowsPrec = undefined

instance Printable Twice where
  precedence _ = 0
  printMe cfg (Twice a b) = withName cfg "Twice" <> a <> " " <> b

type instance Label Maybe = ()
type instance Weight Maybe = ()
type instance F1 Maybe = () 
type instance F3 Maybe = () 
instance RefinementInterface Maybe where
  init   = undefined
  update = undefined
instance ParseMorphism Maybe where
  parseMorphismPoint = undefined

instance Printable Maybe where
  precedence _ = 0
  printMe _ maybe = case maybe of
    Just x -> "Just " <> x
    Nothing -> "Nothing" 
