{-# LANGUAGE FlexibleContexts #-}
module Copar.Functors.AbsorbingPolynomialSpec (spec) where

import           Test.Hspec
import           Test.Hspec.Megaparsec
import           TestHelpers

import qualified Data.List.NonEmpty            as NonEmpty
import           Data.Bifunctor
import           Control.Monad.ST
import           Data.Proxy
import           Data.Either
import           Data.Foldable

import           Data.Vector                    ( Vector )
import qualified Data.Vector                   as V
import           Data.Text                      ( Text )

import           Data.MorphismEncoding          ( Encoding )
import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Desorting
import           Copar.FunctorExpression.Type
import           Copar.Functors.AbsorbingPolynomial
import           Copar.Functors.Polynomial  hiding ( PolyF1(..) )
import           Copar.Functors.SomeFunctor
import           Copar.RefinementInterface
import qualified Data.Partition                as Part
import           Copar.Algorithm

spec :: Spec
spec = do
  parseMorphismPointSpec
  refineSpec


c :: [Text] -> Factor a
c = Const . ExplicitSet . V.fromList


e :: a -> [Text] -> Factor a
e inner = Exponential inner . ExplicitExp . V.fromList


removeSomeFunctor
  :: Encoding
       (Sorted (Label (AbsorbingPolynomial SomeFunctor)))
       (Sorted (F1 (AbsorbingPolynomial SomeFunctor)))
  -> Encoding
       (Sorted (Int, Maybe ()))
       (Sorted (Int, Vector (Maybe ()), Vector Int))
removeSomeFunctor = bimap (fmap processLabels) (fmap processF1)
  where
    processF1 f1 =
      ( polyF1Summand f1
      , V.fromList (map (fmap (const ())) (toList (polyF1Variables f1)))
      , V.convert (polyF1Constants f1)
      )
    processLabels (PolyLabel i x) = (i, () <$ x)


parseMorphismPointSpec :: Spec
parseMorphismPointSpec = describe "parseMorphismPoint" $ do
  let morphp fexpr input =
        (removeSomeFunctor . snd) <$> parseMorphisms (Functor 1 fexpr) EnableSanityChecks "" input

  it "parses a constant" $ do
    morphp (mkPoly [[c ["a", "b", "c"]]]) "x: inj 0 (a)"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []
    morphp (mkPoly [[c ["a", "b", "c"]]]) "x: inj 0 (b)"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [1]))] []

  it "parses the identity"
    $ morphp (mkPoly [[Identity Variable]]) "x: inj 0 (y)\ny: inj 0 (x)"
    `shouldParse` encoding
                    [ (Sorted 1 (f1 0 [Nothing] []))
                    , (Sorted 1 (f1 0 [Nothing] []))
                    ]
                    [ (1, (Sorted 1 (0, Nothing)), 0)
                    , (0, (Sorted 1 (0, Nothing)), 1)
                    ]

  it "gives a useful error if the injection index is out of bounds" $ do
    morphp (mkPoly [[c ["a"]]]) `shouldFailOn` "x: inj 5 (a)"
    morphp (mkPoly [[c ["a"]]]) `shouldFailOn` "x: inj -1 (a)"

  it "parses a product of a constant and an X"
    $ morphp (mkPoly [[c ["a"], Identity Variable]]) "x: inj 0 (a, x)"
    `shouldParse` encoding [(Sorted 1 (f1 0 [Nothing] [0]))]
                           [(0, (Sorted 1 (0, Nothing)), 0)]

  it "parses a product of two elements"
    $ morphp (mkPoly [[Identity Variable, Identity Variable]]) "x: inj 0 (x, x)"
    `shouldParse` encoding
                    [(Sorted 1 (f1 0 [Nothing, Nothing] []))]
                    [ (0, (Sorted 1 (0, Nothing)), 0)
                    , (0, (Sorted 1 (1, Nothing)), 0)
                    ]

  it "parses a sum of two constants"
    $ morphp (mkPoly [[c ["a"]], [c ["b"]]]) "x: inj 0 (a)\ny: inj 1 (b)"
    `shouldParse` encoding
                    [(Sorted 1 (f1 0 [] [0])), (Sorted 1 (f1 1 [] [0]))]
                    []

  it "parses X+(AxX)"
    $ morphp (mkPoly [[Identity Variable], [c ["a"], Identity Variable]])
             "x: inj 0 (y)\ny: inj 1 (a, x)"
    `shouldParse` encoding
                    [ (Sorted 1 (f1 0 [Nothing] []))
                    , (Sorted 1 (f1 1 [Nothing] [0]))
                    ]
                    [ (1, (Sorted 1 (0, Nothing)), 0)
                    , (0, (Sorted 1 (0, Nothing)), 1)
                    ]

  it "allows to omit 'inj' for co-products with only one factor" $ do
    morphp (mkPoly [[c ["a"]]]) "x: (a)"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []

    morphp (mkPoly [[Identity Variable, Identity Variable]]) "x: (x, x)"
      `shouldParse` encoding
                      [(Sorted 1 (f1 0 [Nothing, Nothing] []))]
                      [ (0, (Sorted 1 (0, Nothing)), 0)
                      , (0, (Sorted 1 (1, Nothing)), 0)
                      ]

  it "doesn't confuse a constant called inj and an injection" $ do
    morphp (mkPoly [[c ["injection"]]]) "x: injection"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []
    morphp (mkPoly [[c ["inj"]]]) "x: inj"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []
    morphp (mkPoly [[c ["inj"]]]) "x: inj 0 inj"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []

  it "allows to omit parens for products with only one factor" $ do
    morphp (mkPoly [[c ["a"]], [c ["b"]]]) "x: inj 0 a"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []

  it "allows to omit both 'inj' and parens" $ do
    morphp (mkPoly [[c ["a"]]]) "x: a"
      `shouldParse` encoding [(Sorted 1 (f1 0 [] [0]))] []

  it "parses an exponential" $ do
    morphp (mkPoly [[e Variable ["a"]]]) "x: {a: x}" `shouldParse` encoding
      [(Sorted 1 (f1 0 [Nothing] []))]
      [(0, (Sorted 1 (0, Nothing)), 0)]

  it "fails to parse an exponential that isn't totally defined" $ do
    morphp (mkPoly [[e Variable ["a", "b"]]]) `shouldFailOn` "x: {a: x}"

  it "fails to parse an exponential that isn't uniquely defined" $ do
    morphp (mkPoly [[e Variable ["a", "b"]]])
      `shouldFailOn` "x: {a: x, b: x, b: x}"

  it "parses an exponential under a coproduct" $ do
    morphp (mkPoly [[e Variable ["a", "b"]], [Identity Variable]])
           "x: inj0 {a: x, b: y}\ny: inj1 y"
      `shouldParse` (encoding
                      [ (Sorted 1 (f1 0 [Nothing, Nothing] []))
                      , (Sorted 1 (f1 1 [Nothing] []))
                      ]
                      [ (1, (Sorted 1 (0, Nothing)), 1)
                      , (0, (Sorted 1 (0, Nothing)), 0)
                      , (0, (Sorted 1 (1, Nothing)), 1)
                      ]
                    )

  it "allows positive numbers as constants for integer set" $ do
    morphp (mkPoly [[Const IntSet]]) "x: 5\ny:30"
      `shouldParse` (encoding
                      [(Sorted 1 (f1 0 [] [5])), (Sorted 1 (f1 0 [] [30]))]
                      []
                    )

  it "allows positive numbers as constants for integer set" $ do
    morphp (mkPoly [[Const IntSet]]) "x: -3\ny:-77"
      `shouldParse` (encoding
                      [ (Sorted 1 (f1 0 [] [(-3)]))
                      , (Sorted 1 (f1 0 [] [(-77)]))
                      ]
                      []
                    )

  it "allows positive numbers as constants for naturals set" $ do
    morphp (mkPoly [[Const NatSet]]) "x: 5\ny:30"
      `shouldParse` (encoding
                      [(Sorted 1 (f1 0 [] [5])), (Sorted 1 (f1 0 [] [30]))]
                      []
                    )

  it "doesn't allow negative numbers as constants for naturals set" $ do
    morphp (mkPoly [[Const NatSet]]) `shouldFailOn` "x: -35\ny:-77"

  it "allows numbers in the correct range as constants for FiniteNatSet" $ do
    let f = mkPoly [[Const (FiniteNatSet 4)]]
    morphp f "x: 0" `shouldParse` (encoding [(Sorted 1 (f1 0 [] [0]))] [])
    morphp f `shouldSucceedOn` "x: 1"
    morphp f `shouldSucceedOn` "x: 2"
    morphp f `shouldSucceedOn` "x: 3"

  it "disallows numbers outside bounds for FiniteNatSet" $ do
    let f = mkPoly [[Const (FiniteNatSet 4)]]
    morphp f `shouldFailOn` "x: -1"
    morphp f `shouldFailOn` "x: 4"
    morphp f `shouldFailOn` "x: 10"

  it "parses an exponential with FiniteNatExp" $ do
    morphp (mkPoly [[Exponential Variable (FiniteNatExp 2)]]) "x: {0: x, 1: x}"
      `shouldParse` (encoding
                      [(Sorted 1 (f1 0 [Nothing, Nothing] []))]
                      [ (0, (Sorted 1 (0, Nothing)), 0)
                      , (0, (Sorted 1 (1, Nothing)), 0)
                      ]
                    )

  it "requires all values of the exponent set for FiniteNatExp" $ do
    morphp (mkPoly [[Exponential Variable (FiniteNatExp 2)]])
      `shouldFailOn` "x: {0: x}"

  it "can correctly parse constants from a set with shared prefixes" $ do
    morphp (mkPoly [[Const (ExplicitSet (v ["a1", "a10"]))]]) "x: a10"
      `shouldParse` (encoding [(Sorted 1 (f1 0 [] [1]))] [])


refineSpec :: Spec
refineSpec = describe "refining" $ do
  let morphp fexpr input = snd <$> parseMorphisms (Functor 1 fexpr) EnableSanityChecks "" input

  it "distinguishes constants" $ do
    let Right enc =
          morphp (mkPoly [[c ["a", "b"]]]) "x: inj 0 (a)\ny: inj 0 (b)"
    part <- stToIO (refine (Proxy @(Desorted (AbsorbingPolynomial SomeFunctor))) enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "distinguishes co-factors" $ do
    let f = mkPoly [[c ["a"]], [c ["a"]]]
    let Right enc = morphp f "x: inj 0 (a)\ny: inj 1 (a)"
    part <- stToIO (refine (Proxy @(Desorted (AbsorbingPolynomial SomeFunctor))) enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "correctly identifies factors" $ do
    -- {a,b} + X^2
    let f = mkPoly [[c ["A", "B"]], [Identity Variable, Identity Variable]]
    let
      res = morphp
        f
        "a: inj 0 (A)\n\
          \b: inj 0 (B)\n\
          \x: inj 1 (a, b)\n\
          \y: inj 1 (b, a)"
    res `shouldSatisfy` isRight
    let Right enc = res
    part <- stToIO (refine (Proxy @(Desorted (AbsorbingPolynomial SomeFunctor))) enc True)
    (Part.numBlocks part) `shouldBe` 4

  it "correctly distinguishes different exponential values" $ do
    -- {a,b} + X^2
    let f = mkPoly [[c ["A", "B"]], [e Variable ["i", "j"]]]
    let
      res = morphp
        f
        "a: inj 0 (A)\n\
          \b: inj 0 (B)\n\
          \x: inj 1 {i: a, j: b}\n\
          \y: inj 1 {i: b, j: a}"
    res `shouldSatisfy` isRight
    let Right enc = res
    part <- stToIO (refine (Proxy @(Desorted (AbsorbingPolynomial SomeFunctor))) enc True)
    (Part.numBlocks part) `shouldBe` 4


-- Helpers

mkPoly :: [[Factor a]] -> AbsorbingPolynomial SomeFunctor a
mkPoly =
  AbsorbingPolynomial . fmap Direct . Polynomial . Sum . NonEmpty.fromList . map
    (Product . NonEmpty.fromList)


f1 :: Int -> [Maybe ()] -> [Int] -> (Int, Vector (Maybe ()), Vector Int)
f1 s v c = (s, V.fromList v, V.fromList c)


v :: [a] -> Vector a
v = V.fromList
