module Copar.Functors.DistributionSpec (spec) where

import           Test.Hspec
import           Test.Hspec.Megaparsec
import           TestHelpers

import           Data.Proxy
import           Control.Monad.ST
import           Data.Either

import qualified Data.Partition as Part
import           Copar.Algorithm
import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Desorting
import           Copar.FunctorExpression.Parser
import           Copar.FunctorExpression.Type
import           Copar.Functors.Distribution
import           Copar.FunctorDescription


spec :: Spec
spec = do
  functorExpressionSpec
  parseMorphismPointSpec
  refineSpec

functorExpressionSpec :: Spec
functorExpressionSpec = describe "functorExpression" $ do
  it "parses a normal D correctly" $
    parseFunctorExpression [functorExprParser @Distribution] "" "DX" `shouldParse` (Functor 1 (Distribution Variable))

  it "parses a unicode D (Ɗ) correctly" $
    parseFunctorExpression [functorExprParser @Distribution] "" "ƊX" `shouldParse` (Functor 1 (Distribution Variable))

parseMorphismPointSpec :: Spec
parseMorphismPointSpec = describe "parseMorphismPoint" $ do
  it "works for a simple example" $
    (snd <$>
     parseMorphisms
       (Functor 1 (Distribution Variable))
       EnableSanityChecks
       ""
       "x: {x: 0.5, y: 0.5}\ny: {x: 1.0}") `shouldParse`
    encoding
      [(Sorted 1 1), (Sorted 1 1)]
      [(1, (Sorted 1 1), 0), (0, (Sorted 1 0.5), 0), (0, (Sorted 1 0.5), 1)]

  it "errors if edge weight sum isn't 1" $
    parseMorphisms (Functor 1 (Distribution Variable)) EnableSanityChecks ""
    `shouldFailOn`
      "x: {x: 0.5}"

  it "errors if an edge weight is below zero" $
    parseMorphisms (Functor 1 (Distribution Variable)) EnableSanityChecks ""
    `shouldFailOn`
      "x: {x: -1}"

  it "errors if an edge weight is above one" $
    parseMorphisms (Functor 1 (Distribution Variable)) EnableSanityChecks ""
    `shouldFailOn`
      "x: {x: 1.3}"

  it "uses approximate comparison for doubles" $
    parseMorphisms (Functor 1 (Distribution Variable)) EnableSanityChecks "" `shouldSucceedOn`
      "s0: {s0: 0.1, s1: 0.1, s2: 0.1, s3: 0.1, s4: 0.1, s5: 0.1, s6: 0.1, s7: 0.1, s8: 0.1, s9: 0.1}\n\
      \s1: {s1: 1.0}\n\
      \s2: {s2: 1.0}\n\
      \s3: {s3: 1.0}\n\
      \s4: {s4: 1.0}\n\
      \s5: {s5: 1.0}\n\
      \s6: {s6: 1.0}\n\
      \s7: {s7: 1.0}\n\
      \s8: {s8: 1.0}\n\
      \s9: {s9: 1.0}"

refineSpec :: Spec
refineSpec = describe "refining" $ do
  let f = Functor 1 (Distribution Variable)

  it "handles states with incoming edges greater than 1" $ do
    let x = parseMorphisms f EnableSanityChecks "" "x: {z: 1.0}\ny: {z: 1.0}\nz: {y: 1.0}"
    x `shouldSatisfy` isRight
    let Right (_, enc) = x
    stToIO (refine (Proxy @(Desorted Distribution)) enc True) `shouldReturn` Part.fromBlocks [[0, 1, 2]]
