module Copar.Functors.GroupValuedSpec (spec) where

import           Test.Hspec
import           Test.Hspec.Megaparsec
import           TestHelpers

import           Data.Complex
import           Data.Ratio

import           Copar.FunctorExpression.Parser
import           Copar.FunctorDescription
import           Copar.Functors.GroupValued
import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Type
import           Copar.FunctorExpression.Sorts     ( Sorted(..) )
import           Data.Float.Utils               ( EqDouble )


spec :: Spec
spec = do
  parseIntSpec
  parseMorphismPointIntSpec
  parseMorphismPointDoubleSpec
  parseMorphismPointComplexSpec
  parseMorphismPointRationalSpec

parseIntSpec :: Spec
parseIntSpec = describe "Int-valued parser" $ do
  it "parses Z^X" $
    parseFunctorExpression [functorExprParser @IntValued] "" "Z^X"
    `shouldParse` (Functor 1 (GroupValued Variable))

parseMorphismPointIntSpec :: Spec
parseMorphismPointIntSpec = describe "parseMorphismPoint (Int)" $ do
  let p = fmap snd . parseMorphisms (Functor 1 (GroupValued @Int Variable)) EnableSanityChecks ""

  it "parses an empty successor list"
    $ p "x: {}"
    `shouldParse` encoding [(Sorted 1 0)] []

  it "parses a simple example"
    $ p "x: {x: 2, y: 3}\ny: {}"
    `shouldParse` encoding [(Sorted 1 5), (Sorted 1 0)]
                           [(0, (Sorted 1 2), 0), (0, (Sorted 1 3), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 2, x: 3}"

  it "works with negative numbers" $ p `shouldSucceedOn` "x: {x: -2}"


parseMorphismPointDoubleSpec :: Spec
parseMorphismPointDoubleSpec = describe "parseMorphismPoint (Double)" $ do
  let
    p =
      fmap snd . parseMorphisms (Functor 1 (GroupValued @EqDouble Variable)) EnableSanityChecks ""

  it "parses an empty successor list"
    $ p "x: {}"
    `shouldParse` encoding [(Sorted 1 0)] []

  it "parses a simple example"
    $ p "x: {x: 0.5, y: 3.7}\ny: {}"
    `shouldParse` encoding [(Sorted 1 4.2), (Sorted 1 0)]
                           [(0, (Sorted 1 0.5), 0), (0, (Sorted 1 3.7), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 0.5, x: 3.7}"

  it "works with negative numbers" $ p `shouldSucceedOn` "x: {x: -2.3}"


parseMorphismPointComplexSpec :: Spec
parseMorphismPointComplexSpec = describe "parseMorphismPoint (Complex)" $ do
  let p = fmap snd . parseMorphisms
        (Functor 1 (GroupValued @OrderedComplex Variable))
        EnableSanityChecks
        ""

  it "parses an empty successor list"
    $ p "x: {}"
    `shouldParse` encoding [(Sorted 1 0)] []

  it "parses a simple example with only real parts"
    $ p "x: {x: 0.5, y: 3.7}\ny: {}"
    `shouldParse` encoding
                    [(Sorted 1 (c 4.2 0)), (Sorted 1 0)]
                    [(0, (Sorted 1 (c 0.5 0)), 0), (0, (Sorted 1 (c 3.7 0)), 1)]


  it "parses a simple example with only imaginary parts"
    $ p "x: {x: 0.5i, y: 3.7i}\ny: {}"
    `shouldParse` encoding
                    [(Sorted 1 (c 0 4.2)), (Sorted 1 0)]
                    [(0, (Sorted 1 (c 0 0.5)), 0), (0, (Sorted 1 (c 0 3.7)), 1)]


  it "parses a simple exmaple with mixed real and fractional parts"
    $ p "x: {x: 1 + 0.5i, y: 3.7i + 2}\ny: {}"
    `shouldParse` encoding
                    [(Sorted 1 (c 3 4.2)), (Sorted 1 0)]
                    [(0, (Sorted 1 (c 1 0.5)), 0), (0, (Sorted 1 (c 2 3.7)), 1)]


  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 2, x: 3}"

  it "works with negative numbers" $ p "x: {x: -2.3}" `shouldParse` encoding
    [Sorted 1 (c (-2.3) 0)]
    [(0, Sorted 1 (c (-2.3) 0), 0)]

  where c a b = OrderedComplex (a :+ b)

parseMorphismPointRationalSpec :: Spec
parseMorphismPointRationalSpec = describe "parseMorphismPoint (Rational)" $ do
  let p = fmap snd . parseMorphisms
        (Functor 1 (GroupValued @(Ratio Int) Variable))
        EnableSanityChecks
        ""

  it "parses a simple example"
    $ p "x: {x: 1/5, y: 3/7}\ny: {}"
    `shouldParse` encoding [(Sorted 1 (22%35)), (Sorted 1 0)]
                           [(0, (Sorted 1 (1%5)), 0), (0, (Sorted 1 (3%7)), 1)]

  it "parses negative values"
    $ p "x: {x: -1/5, y: -3/7}\ny: {}"
    `shouldParse` encoding [(Sorted 1 (-22%35)), (Sorted 1 0)]
                           [(0, (Sorted 1 (-1%5)), 0), (0, (Sorted 1 (-3%7)), 1)]
