{-# LANGUAGE ConstraintKinds #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE AllowAmbiguousTypes #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Copar.Functors.MonoidValuedSpec (spec) where

import           Test.Hspec
import           TestHelpers (encoding)

import           Data.Semigroup                 ( Max(..)
                                                , Min(..)
                                                )
import           Control.Monad.ST
import           Data.Proxy
import           Data.Void
import           Data.Word

import           Test.Hspec.Megaparsec
import           Data.Text                      ( Text )
import qualified Data.Text                     as T
import           Text.Megaparsec                ( ParseErrorBundle )

import           Copar.FunctorExpression.Parser
import           Copar.Functors.MonoidValued
import           Copar.FunctorExpression.Type
import           Copar.FunctorDescription
import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Sorts     ( Sorted(..) )
import           Copar.FunctorExpression.Desorting
import qualified Data.Partition                as Part
import           Copar.Algorithm
import           Data.Float.Utils
import           Data.Bits.Monoid
import           Data.MorphismEncoding          ( Encoding )
import           Copar.RefinementInterface      ( Label, F1 )

spec :: Spec
spec = do
  maxIntParseSpec
  minIntParseSpec
  maxIntRefineSpec
  minIntRefineSpec
  maxRealParseSpec
  minRealParseSpec
  maxRealRefineSpec
  minRealRefineSpec
  andWordParseSpec
  andWordRefineSpec
  orWordParseSpec
  orWordRefineSpec


maxIntParseSpec :: Spec
maxIntParseSpec = describe "maxInt parsing" $ do
  makeFunctorParseSpec @(Max Int) ("Z", Just "ℤ") "max"

  let p = makeMorphParser @(Max Int)

  itParsesEmpty @(Max Int)

  it "parses a simple example"
    $ p "x: {x: 2, y: 3}\ny: {}"
    `shouldParse` encoding [Sorted 1 3, Sorted 1 minBound]
                           [(0, (Sorted 1 2), 0), (0, (Sorted 1 3), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 2, x: 3}"

  it "works with negative numbers" $ p `shouldSucceedOn` "x: {x: -2}"

minIntParseSpec :: Spec
minIntParseSpec = describe "minIntParse" $ do
  makeFunctorParseSpec @(Min Int) ("Z", Just "ℤ") "min"

  let p = makeMorphParser @(Min Int)

  itParsesEmpty @(Min Int)

  it "parses a simple example"
    $ p "x: {x: 2, y: 3}\ny: {}"
    `shouldParse` encoding [Sorted 1 2, Sorted 1 maxBound]
                           [(0, (Sorted 1 2), 0), (0, (Sorted 1 3), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 2, x: 3}"

  it "works with negative numbers" $ p `shouldSucceedOn` "x: {x: -2}"


maxIntRefineSpec :: Spec
maxIntRefineSpec = describe "maxInt refine" $ do
  let p = makeMorphParser @(Max Int)
      proxy = Proxy @(Desorted MaxIntValued)

  it "it distinguishes different maximas with equal sums" $ do
    let Right enc = p "x: {x: 1, y: 3}\ny: {x: 2, y: 2}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "identifies equal maximas with different sums" $ do
    let Right enc = p "x: {x: 1, y: 3}\ny: {x: 3, y: 2}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0, 1]]


minIntRefineSpec :: Spec
minIntRefineSpec = describe "minInt refine" $ do
  let p = makeMorphParser @(Min Int)
      proxy = Proxy @(Desorted MinIntValued)

  it "it distinguishes different minimas with equal sums" $ do
    let Right enc = p "x: {x: 1, y: 3}\ny: {x: 2, y: 2}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "identifies equal minimas with different sums" $ do
    let Right enc = p "x: {x: 1, y: 3}\ny: {x: 4, y: 1}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0, 1]]


maxRealParseSpec :: Spec
maxRealParseSpec = describe "maxReal parsing" $ do
  makeFunctorParseSpec @MaxDouble ("R", Just "ℝ") "max"

  let p = makeMorphParser @MaxDouble

  itParsesEmpty @MaxDouble

  it "parses a simple example"
    $ p "x: {x: 2.5, y: 3.7}\ny: {}"
    `shouldParse` encoding [Sorted 1 3.7, Sorted 1 mempty]
                           [(0, (Sorted 1 2.5), 0), (0, (Sorted 1 3.7), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 2.3, x: 3.6}"

  it "works with negative numbers" $ p `shouldSucceedOn` "x: {x: -2.8}"


minRealParseSpec :: Spec
minRealParseSpec = describe "minReal parsing" $ do
  makeFunctorParseSpec @MinDouble ("R", Just "ℝ") "min"

  let p = makeMorphParser @MinDouble

  itParsesEmpty @MinDouble

  it "parses a simple example"
    $ p "x: {x: 2.5, y: 3.7}\ny: {}"
    `shouldParse` encoding [Sorted 1 2.5, Sorted 1 mempty]
                           [(0, (Sorted 1 2.5), 0), (0, (Sorted 1 3.7), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 2.3, x: 3.6}"

  it "works with negative numbers" $ p `shouldSucceedOn` "x: {x: -2.8}"


maxRealRefineSpec :: Spec
maxRealRefineSpec = describe "maxReal refine" $ do
  let p = makeMorphParser @MaxDouble
      proxy = Proxy @(Desorted MaxDoubleValued)

  it "it distinguishes different maximas with equal sums" $ do
    let Right enc = p "x: {x: 1.1, y: 3.1}\ny: {x: 2.1, y: 2.1}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "identifies equal maximas with different sums" $ do
    let Right enc = p "x: {x: 1.1, y: 3.1}\ny: {x: 3.1, y: 2.1}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0, 1]]


minRealRefineSpec :: Spec
minRealRefineSpec = describe "minReal refine" $ do
  let p = makeMorphParser @MinDouble
      proxy = Proxy @(Desorted (SlowMonoidValued MinDouble))

  it "it distinguishes different minimas with equal sums" $ do
    let Right enc = p "x: {x: 1.1, y: 3.1}\ny: {x: 2.1, y: 2.1}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "identifies equal minimas with different sums" $ do
    let Right enc = p "x: {x: 1.1, y: 3.1}\ny: {x: 1.1, y: 2.1}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0, 1]]


andWordParseSpec :: Spec
andWordParseSpec = describe "bit-and parsing" $ do
  makeFunctorParseSpec @(BitAnd Word64) ("Word", Nothing) "and"

  let p = makeMorphParser @(BitAnd Word64)

  itParsesEmpty @(BitAnd Word64)

  it "parses a simple example"
    $ p "x: {x: 0xA0, y: 0x0A}\ny: {}"
    `shouldParse` encoding [Sorted 1 0, Sorted 1 mempty]
                           [(0, (Sorted 1 0xA0), 0), (0, (Sorted 1 0x0A), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 0xA, x: 0xA}"


andWordRefineSpec :: Spec
andWordRefineSpec = describe "andWord refine" $ do
  let p = makeMorphParser @(BitAnd Word64)
      proxy = Proxy @(Desorted (SlowMonoidValued (BitAnd Word64)))

  it "it distinguishes different meets with equal sums" $ do
    let Right enc = p "x: {x: 0xAA, y: 0xCC}\ny: {x: 0xBB, y: 0xBB}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "identifies equal meets with different sums" $ do
    let Right enc = p "x: {x: 0xA0, y: 0x0A}\ny: {x: 0x0B, y: 0xB0}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0, 1]]


orWordParseSpec :: Spec
orWordParseSpec = describe "bit-or parsing" $ do
  makeFunctorParseSpec @(BitOr Word64) ("Word", Nothing) "or"

  let p = makeMorphParser @(BitOr Word64)

  itParsesEmpty @(BitOr Word64)

  it "parses a simple example"
    $ p "x: {x: 0xA0, y: 0x0A}\ny: {}"
    `shouldParse` encoding [Sorted 1 0xAA, Sorted 1 mempty]
                           [(0, (Sorted 1 0xA0), 0), (0, (Sorted 1 0x0A), 1)]

  it "fails on duplicate edges" $ p `shouldFailOn` "x: {x: 0xA, x: 0xA}"


orWordRefineSpec :: Spec
orWordRefineSpec = describe "orWord refine" $ do
  let p = makeMorphParser @(BitOr Word64)
      proxy = Proxy @(Desorted (SlowMonoidValued (BitOr Word64)))

  it "it distinguishes different joins with equal sums" $ do
    let Right enc = p "x: {x: 0xAA, y: 0xCC}\ny: {x: 0xBB, y: 0xBB}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "identifies equal joins with different sums" $ do
    let Right enc = p "x: {x: 0xA0, y: 0x0A}\ny: {x: 0xAA, y: 0x00}"
    part <- stToIO (refine proxy enc True)
    (Part.toBlocks part) `shouldMatchList` [[0, 1]]


makeFunctorParseSpec :: forall m. FunctorDescription (SlowMonoidValued m)
                        => (Text, Maybe Text)
                        -> Text
                        -> Spec
makeFunctorParseSpec (setAscii, setUnicode) operation = do
  let expr = syntax setAscii operation "X"
  it ("can parse " <> T.unpack expr <> " as functor expression")
    $ parseFunctorExpression [functorExprParser @(SlowMonoidValued m)] "" expr
    `shouldParse` (Functor 1 (SlowMonoidValued Variable))

  case setUnicode of
    Nothing -> return ()
    Just set -> do
      let expr = syntax set operation "X"
      it ("can parse " <> T.unpack expr <> " as functor expression")
        $ parseFunctorExpression [functorExprParser @(SlowMonoidValued m)] "" expr
        `shouldParse` (Functor 1 (SlowMonoidValued Variable))

  it "nests correctly in functor expressions"
    $ parseFunctorExpression [functorExprParser @(SlowMonoidValued m)]
                             ""
                             (syntax setAscii operation ("(" <> syntax setAscii operation "X" <> ")"))
    `shouldParse` (Functor
                    1
                    (SlowMonoidValued (Functor 1 (SlowMonoidValued Variable)))
                  )

  it "still parses parenthesis in functor expressions correctly" $ do
    parseFunctorExpression [functorExprParser @(SlowMonoidValued m)] "" ("(" <> syntax setAscii operation "X" <> ")")
      `shouldParse` (Functor 1 (SlowMonoidValued Variable))
    parseFunctorExpression [functorExprParser @(SlowMonoidValued m)] "" (syntax setAscii operation "(X)")
      `shouldParse` (Functor 1 (SlowMonoidValued Variable))

  where syntax set op power = "(" <> set <> ", " <> op <> ")^" <> power

makeMorphParser
  :: ParseMorphism (SlowMonoidValued m)
  => Text
  -> ( Either
         (ParseErrorBundle Text Void)
         ( Encoding
             (Label (Desorted (SlowMonoidValued m)))
             (F1 (Desorted (SlowMonoidValued m)))
         )
     )
makeMorphParser = fmap snd . parseMorphisms
  (Functor 1 (SlowMonoidValued Variable))
  EnableSanityChecks
  ""

itParsesEmpty :: forall m. (Show m, Eq m, Monoid m, ParseMorphism (SlowMonoidValued m)) => Spec
itParsesEmpty =
  it "parses an empty successor list"
    $ makeMorphParser @m "x: {}"
    `shouldParse` encoding [Sorted 1 mempty] []
