module Copar.Functors.PolynomialSpec (spec) where

import           Test.Hspec
import           Test.Hspec.Megaparsec
import           TestHelpers

import           Data.List (isInfixOf)
import           Control.Monad.ST
import           Data.Either (isRight)
import           Data.Functor.Classes
import qualified Data.List.NonEmpty as NonEmpty
import           Data.Proxy
import           Type.Reflection

import           Text.Megaparsec (errorBundlePretty)
import           Data.Vector (Vector)
import qualified Data.Vector as V
import qualified Data.Vector.Unboxed as VU
import           Data.Text (Text)

import qualified Copar.Parser.Lexer as L
import           Copar.Algorithm
import qualified Data.Partition as Part
import           Copar.Coalgebra.Parser
import           Copar.FunctorExpression.Desorting
import           Copar.FunctorExpression.Type
import           Copar.Functors.Polynomial
import           Copar.FunctorExpression.Parser
import           Copar.FunctorDescription

spec :: Spec
spec = do
  functorExpressionSpec
  parseMorphismPointSpec
  refineSpec

functorExpressionSpec :: Spec
functorExpressionSpec = describe "functorExpression" $ do
  let p = parseFunctorExpression [functorExprParser @Polynomial] ""
      ident = Identity Variable

  it "parses a constant" $
    p "{a, b}" `shouldParse` (Functor 1 (mkPoly [[c ["a", "b"]]]))

  it "parses a product" $ do
    p "X x X" `shouldParse` (Functor 1 (mkPoly [[ident, ident]]))
    p "{a} x {b}" `shouldParse` (Functor 1 (mkPoly [[c ["a"], c ["b"]]]))
    p "X x {b}" `shouldParse` (Functor 1 (mkPoly [[ident, c ["b"]]]))
    p "{a} x X" `shouldParse` (Functor 1 (mkPoly [[c ["a"], ident]]))

  it "allows to use the unicode multiplication sign for products" $
    p "X×X" `shouldParse` (Functor 1 (mkPoly [[ident, ident]]))

  it "parses a co-product" $ do
    p "X + X" `shouldParse` (Functor 1 (mkPoly [[ident], [ident]]))
    p "{a} + {b}" `shouldParse` (Functor 1 (mkPoly [[c ["a"]], [c ["b"]]]))
    p "{a} + X" `shouldParse` (Functor 1 (mkPoly [[c ["a"]], [ident]]))
    p "X + {b}" `shouldParse` (Functor 1 (mkPoly [[ident], [c ["b"]]]))

  it "parses a product with more than two factors" $ do
    p "X x X x X" `shouldParse` (Functor 1 (mkPoly [[ident, ident, ident]]))

  it "parses a co-product with more than two factors" $ do
    p "X + X + X" `shouldParse` (Functor 1 (mkPoly [[ident], [ident], [ident]]))

  it "parses a complex nested expression" $ do
    p "{a} + Xx{b,c} + Xx{d}xX" `shouldParse`
      (Functor 1 (mkPoly [ [c ["a"]]
                         , [ident, c ["b", "c"]]
                         , [ident, c ["d"], ident]]))

  it "correctly parses a single subfunctor" $ do
    parseFunctorExpression [ transParser SomeFunctor prefixParser
                           , transParser SomeFunctor (functorExprParser @Polynomial)
                           ]
      ""
      "P X" `shouldParse` (Functor 2 (SomeFunctor (PrefixFunctor Variable)))

  it "parses a product of two subfunctors" $ do
    parseFunctorExpression [ transParser SomeFunctor prefixParser
                           , transParser SomeFunctor (functorExprParser @Polynomial)
                           ]
      ""
      "PX x PX"
      `shouldParse`
      (Functor 1
       (SomeFunctor
        (mkPoly [[ Identity (Functor 2 (SomeFunctor (PrefixFunctor Variable)))
                 , Identity (Functor 2 (SomeFunctor (PrefixFunctor Variable)))
                 ]])))

  it "parses a co-product of two subfunctors" $ do
    parseFunctorExpression [ transParser SomeFunctor prefixParser
                           , transParser SomeFunctor (functorExprParser @Polynomial)
                           ]
      ""
      "PX + PX"
      `shouldParse`
      (Functor 1
       (SomeFunctor
        (mkPoly [[ Identity (Functor 2 (SomeFunctor (PrefixFunctor Variable))) ]
                ,[ Identity (Functor 2 (SomeFunctor (PrefixFunctor Variable))) ]
                ])))

  it "parses an exponential" $ do
    p "X^{a,b}"
      `shouldParse` (Functor 1 (mkPoly [[e Variable ["a", "b"]]]))

  it "parses an exponential under a product" $ do
    p "X^{a} x X" `shouldParse`
      (Functor 1 (mkPoly [[e Variable ["a"], ident]]))

    p "XxX^{a}" `shouldParse`
      (Functor 1 (mkPoly [[ident, e Variable ["a"]]]))

  it "parses an exponential under a coproduct" $ do
    p "X^{a} + X" `shouldParse`
      (Functor 1 (mkPoly [[e Variable ["a"]], [ident]]))
    p "X + X^{a}" `shouldParse`
      (Functor 1 (mkPoly [[ident], [e Variable ["a"]]]))

  it "parses a deply nested exponential" $ do
    p "X^{a}xX + X" `shouldParse`
      (Functor 1 (mkPoly [[e Variable ["a"], ident], [ident]]))

  it "errors if exponential has non-uniquely defined domain" $
    p `shouldFailOn` "X^{a, a}"

  -- This is a bit of a hack: We need to exclude the following behaviour:
  -- 1. The parser tries to parse ^{a, a} and fails because of non-unique keys
  -- 2. The parser parses the "X" instead as an identity
  -- 3. The whole parser fails with an error saying that you can't just have an
  --    "X" as functor.
  it "gives a useful error message for non-unique keys" $ do
    let Left err = p "X^{a, a}"
    errorBundlePretty err `shouldSatisfy` ("unique" `isInfixOf`)

  it "allows to use intergers as set for constants" $ do
    p "Z" `shouldParse` (Functor 1 (mkPoly [[Const IntSet]]))
    p "ℤ" `shouldParse` (Functor 1 (mkPoly [[Const IntSet]]))

  it "allows to use naturals as set for constants" $ do
    p "N" `shouldParse` (Functor 1 (mkPoly [[Const NatSet]]))
    p "ℕ" `shouldParse` (Functor 1 (mkPoly [[Const NatSet]]))

  it "allows to use explicit sets of a certain size as constants" $ do
    p "2" `shouldParse` (Functor 1 (mkPoly [[Const (FiniteNatSet 2)]]))
    p "55" `shouldParse` (Functor 1 (mkPoly [[Const (FiniteNatSet 55)]]))
    p `shouldFailOn` "-2"

  it "allows to use explicit sets of a certain size as exponents" $ do
    p "X^2" `shouldParse`
      (Functor 1 (mkPoly [[Exponential Variable (FiniteNatExp 2)]]))
    p "X^100" `shouldParse`
      (Functor 1 (mkPoly [[Exponential Variable (FiniteNatExp 100)]]))

  it "disallows negative exponents" $ do
    p `shouldFailOn` "X^-1"

data SomeFunctor a where
  SomeFunctor :: (Eq1 f, Typeable f, Show1 f) => f a -> SomeFunctor a

instance Eq1 SomeFunctor where
  liftEq eq (SomeFunctor (f1 :: tf1 a)) (SomeFunctor (f2 :: tf2 b)) =
    case eqTypeRep (typeRep @tf1) (typeRep @tf2) of
      Nothing -> False -- different types
      Just HRefl -> liftEq eq f1 f2

instance Show1 SomeFunctor where
  liftShowsPrec show' showList' prec (SomeFunctor f) = liftShowsPrec show' showList' prec f

data PrefixFunctor a = PrefixFunctor a
  deriving (Functor, Foldable, Traversable, Show)

instance Eq1 PrefixFunctor where
  liftEq f (PrefixFunctor x) (PrefixFunctor y) = f x y

instance Show1 PrefixFunctor where
  liftShowsPrec show' _ prec (PrefixFunctor x) rest
    | prec > 10 = "(PrefixFunctor " ++ show' 11 x (rest ++ ")")
    | otherwise = "PrefixFunctor " ++ show' 11 x rest

prefixParser :: FunctorParser PrefixFunctor
prefixParser = prefix (L.symbol "P" >> return PrefixFunctor)

c :: [Text] -> Factor a
c = Const . ExplicitSet . V.fromList

e :: a -> [Text] -> Factor a
e inner = Exponential inner . ExplicitExp . V.fromList

parseMorphismPointSpec :: Spec
parseMorphismPointSpec = describe "parseMorphismPoint" $ do
  let morphp fexpr input = snd <$> parseMorphisms (Functor 1 fexpr) EnableSanityChecks "" input

  it "parses a constant" $ do
    morphp (mkPoly [[c ["a", "b", "c"]]]) "x: inj 0 (a)" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []
    morphp (mkPoly [[c ["a", "b", "c"]]]) "x: inj 0 (b)" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [1]))] []

  it "parses the identity" $
    morphp (mkPoly [[Identity Variable]]) "x: inj 0 (y)\ny: inj 0 (x)" `shouldParse`
    encoding
      [(Sorted 1 (f1 0 1 [])), (Sorted 1 (f1 0 1 []))]
      [(1, (Sorted 1 0), 0), (0, (Sorted 1 0), 1)]

  it "gives a useful error if the injection index is out of bounds" $ do
    morphp (mkPoly [[c ["a"]]]) `shouldFailOn `"x: inj 5 (a)"
    morphp (mkPoly [[c ["a"]]]) `shouldFailOn `"x: inj -1 (a)"

  it "parses a product of a constant and an X" $
    morphp (mkPoly [[c ["a"], Identity Variable]]) "x: inj 0 (a, x)" `shouldParse`
    encoding [(Sorted 1 (f1 0 1 [0]))] [(0, (Sorted 1 0), 0)]

  it "parses a product of two elements" $
    morphp (mkPoly [[Identity Variable, Identity Variable]]) "x: inj 0 (x, x)" `shouldParse`
    encoding [(Sorted 1 (f1 0 2 []))] [(0, (Sorted 1 0), 0), (0, (Sorted 1 1), 0)]

  it "parses a sum of two constants" $
    morphp (mkPoly [[c ["a"]], [c ["b"]]]) "x: inj 0 (a)\ny: inj 1 (b)" `shouldParse`
    encoding [(Sorted 1 (f1 0 0 [0])), (Sorted 1 (f1 1 0 [0]))] []

  it "parses X+(AxX)" $
    morphp
      (mkPoly [[Identity Variable], [c ["a"], Identity Variable]])
      "x: inj 0 (y)\ny: inj 1 (a, x)" `shouldParse`
    encoding
      [ (Sorted 1 (f1 0 1 []))
      , (Sorted 1 (f1 1 1 [0]))
      ]
      [(1, (Sorted 1 0), 0), (0, (Sorted 1 0), 1)]

  it "allows to omit 'inj' for co-products with only one factor" $ do
    morphp (mkPoly [[c ["a"]]]) "x: (a)" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []

    morphp (mkPoly [[Identity Variable, Identity Variable]]) "x: (x, x)" `shouldParse`
      encoding [(Sorted 1 (f1 0 2 []))] [(0, (Sorted 1 0), 0), (0, (Sorted 1 1), 0)]

  it "doesn't confuse a constant called inj and an injection" $ do
    morphp (mkPoly [[c ["injection"]]]) "x: injection" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []
    morphp (mkPoly [[c ["inj"]]]) "x: inj" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []
    morphp (mkPoly [[c ["inj"]]]) "x: inj 0 inj" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []

  it "allows to omit parens for products with only one factor" $ do
    morphp (mkPoly [[c ["a"]], [c ["b"]]]) "x: inj 0 a" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []

  it "allows to omit both 'inj' and parens" $ do
    morphp (mkPoly [[c ["a"]]]) "x: a" `shouldParse`
      encoding [(Sorted 1 (f1 0 0 [0]))] []

  it "parses an exponential" $ do
    morphp (mkPoly [[e Variable ["a"]]]) "x: {a: x}" `shouldParse`
      encoding [(Sorted 1 (f1 0 1 []))] [(0, (Sorted 1 0), 0)]

  it "fails to parse an exponential that isn't totally defined" $ do
    morphp (mkPoly [[e Variable ["a", "b"]]])
      `shouldFailOn` "x: {a: x}"

  it "fails to parse an exponential that isn't uniquely defined" $ do
    morphp (mkPoly [[e Variable ["a", "b"]]])
      `shouldFailOn` "x: {a: x, b: x, b: x}"

  it "parses an exponential under a coproduct" $ do
    morphp (mkPoly [[e Variable ["a", "b"]], [Identity Variable]])
        "x: inj0 {a: x, b: y}\ny: inj1 y"
      `shouldParse`
      (encoding
        [ (Sorted 1 (f1 0 2 []))
        , (Sorted 1 (f1 1 1 []))
        ]
        [ (1, (Sorted 1 0), 1)
        , (0, (Sorted 1 0), 0)
        , (0, (Sorted 1 1), 1)
        ])

  it "allows positive numbers as constants for integer set" $ do
    morphp (mkPoly [[Const IntSet]]) "x: 5\ny:30" `shouldParse`
      (encoding
         [ (Sorted 1 (f1 0 0 [5]))
         , (Sorted 1 (f1 0 0 [30]))
         ]
         [])

  it "allows positive numbers as constants for integer set" $ do
    morphp (mkPoly [[Const IntSet]]) "x: -3\ny:-77" `shouldParse`
      (encoding
         [ (Sorted 1 (f1 0 0 [(-3)]))
         , (Sorted 1 (f1 0 0 [(-77)]))
         ]
         [])

  it "allows positive numbers as constants for naturals set" $ do
    morphp (mkPoly [[Const NatSet]]) "x: 5\ny:30" `shouldParse`
      (encoding
         [ (Sorted 1 (f1 0 0 [5]))
         , (Sorted 1 (f1 0 0 [30]))
         ]
         [])

  it "doesn't allow negative numbers as constants for naturals set" $ do
    morphp (mkPoly [[Const NatSet]]) `shouldFailOn` "x: -35\ny:-77"

  it "allows numbers in the correct range as constants for FiniteNatSet" $ do
    let f = mkPoly [[Const (FiniteNatSet 4)]]
    morphp f "x: 0" `shouldParse` (encoding [(Sorted 1 (f1 0 0 [0]))] [])
    morphp f `shouldSucceedOn` "x: 1"
    morphp f `shouldSucceedOn` "x: 2"
    morphp f `shouldSucceedOn` "x: 3"

  it "disallows numbers outside bounds for FiniteNatSet" $ do
    let f = mkPoly [[Const (FiniteNatSet 4)]]
    morphp f `shouldFailOn` "x: -1"
    morphp f `shouldFailOn` "x: 4"
    morphp f `shouldFailOn` "x: 10"

  it "parses an exponential with FiniteNatExp" $ do
    morphp (mkPoly [[Exponential Variable (FiniteNatExp 2)]]) "x: {0: x, 1: x}"
      `shouldParse`
      (encoding
       [(Sorted 1 (f1 0 2 []))]
       [(0, (Sorted 1 0), 0), (0, (Sorted 1 1), 0)])

  it "requires all values of the exponent set for FiniteNatExp" $ do
    morphp (mkPoly [[Exponential Variable (FiniteNatExp 2)]])
      `shouldFailOn` "x: {0: x}"

  it "can correctly parse constants from a set with shared prefixes" $ do
    morphp (mkPoly [[Const (ExplicitSet (v ["a1", "a10"]))]]) "x: a10"
      `shouldParse`
      (encoding [(Sorted 1 (f1 0 0 [1]))] [])

refineSpec :: Spec
refineSpec = describe "refining" $ do
  let morphp fexpr input = snd <$> parseMorphisms (Functor 1 fexpr) EnableSanityChecks "" input

  it "distinguishes constants" $ do
    let Right enc = morphp (mkPoly [[c ["a", "b"]]]) "x: inj 0 (a)\ny: inj 0 (b)"
    part <- stToIO (refine (Proxy @(Desorted Polynomial)) enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "distinguishes co-factors" $ do
    let f = mkPoly [[c ["a"]], [c ["a"]]]
    let Right enc = morphp f "x: inj 0 (a)\ny: inj 1 (a)"
    part <- stToIO (refine (Proxy @(Desorted Polynomial)) enc True)
    (Part.toBlocks part) `shouldMatchList` [[0], [1]]

  it "correctly identifies factors" $ do
    -- {a,b} + X^2
    let f = mkPoly [[c ["A", "B"]], [Identity Variable, Identity Variable]]
    let res = morphp f
          "a: inj 0 (A)\n\
          \b: inj 0 (B)\n\
          \x: inj 1 (a, b)\n\
          \y: inj 1 (b, a)"
    res `shouldSatisfy` isRight
    let Right enc = res
    part <- stToIO (refine (Proxy @(Desorted Polynomial)) enc True)
    (Part.numBlocks part) `shouldBe` 4

  it "correctly distinguishes different exponential values" $ do
    -- {a,b} + X^2
    let f = mkPoly [[c ["A", "B"]], [e Variable ["i", "j"]]]
    let res = morphp f
          "a: inj 0 (A)\n\
          \b: inj 0 (B)\n\
          \x: inj 1 {i: a, j: b}\n\
          \y: inj 1 {i: b, j: a}"
    res `shouldSatisfy` isRight
    let Right enc = res
    part <- stToIO (refine (Proxy @(Desorted Polynomial)) enc True)
    (Part.numBlocks part) `shouldBe` 4

-- Helpers

mkPoly :: [[Factor a]] -> Polynomial a
mkPoly =
  Polynomial . Sum . NonEmpty.fromList . map (Product . NonEmpty.fromList)


f1 :: Int -> Int -> [Int] -> PolyF1
f1 s v = PolyF1 s v . VU.fromList


v :: [a] -> Vector a
v = V.fromList
