module Copar.Parser.LexerSpec (spec) where

import           Test.Hspec
import           Test.Hspec.Megaparsec

import           Data.Complex
import           Data.Ratio

import           Text.Megaparsec

import qualified Copar.Parser.Lexer as L

spec :: Spec
spec = do
  spaceSpec
  complexSpec
  hexSpec
  fractionSpec
  nameSpec

spaceSpec :: Spec
spaceSpec = describe "whitespace" $ do
  describe "space" $ do
    let p = parse (L.space <* eof) ""
    it "skips spaces" $
      p `shouldSucceedOn` "    "

    it "skips tabs" $
      p `shouldSucceedOn` "\t\t"

    it "doesn't parse letters" $
      p `shouldFailOn` "   a"

    it "doesn't parse newlines" $
      p `shouldFailOn` "   \n"

  describe "newlines1" $ do
    let p = parse (L.newlines1 <* eof) ""

    it "skips many newlines" $
      p `shouldSucceedOn` "\n\n\n"

    it "skips newlines interleaved with spaces" $
      p `shouldSucceedOn` "\n   \n  \n"

    it "skips comments" $
      p `shouldSucceedOn` "#foo\n#bar\n\n  # lala\n"

    it "fails on the empty input" $
      p `shouldFailOn` ""

    it "fails with leading whitespace" $
      p `shouldFailOn` "   \n"

  describe "newlines" $ do
    let p = parse (L.newlines <* eof) ""

    it "skips many newlines" $
      p `shouldSucceedOn` "\n\n\n"

    it "skips newlines interleaved with spaces" $
      p `shouldSucceedOn` "\n   \n  \n"

    it "skips comments" $
      p `shouldSucceedOn` "#foo\n#bar\n\n  # lala\n"

    it "parses the empty input" $
      p `shouldSucceedOn` ""

    it "fails with leading whitespace" $
      p `shouldFailOn` "   \n"


complexSpec :: Spec
complexSpec = describe "complex" $ do
  context "over ints" $ do
    let p = parse (L.complex L.decimal <* eof) ""

    it "parses an unsigned real part" $
      p "42" `shouldParse` (42 :+ 0)

    it "parses a negative real part" $
      p "-42" `shouldParse` ((-42) :+ 0)

    it "parses an unsigned imaginary part" $
      p "42i" `shouldParse` (0 :+ 42)

    it "parses a negative imaginary part" $
      p "-42i" `shouldParse` (0 :+ (-42))

    it "parses a sum with positive summands" $
      p "42 + 42i" `shouldParse` (42 :+ 42)

    it "parses a sum with negative real part" $
      p "-42 + 42i" `shouldParse` ((-42) :+ 42)

    it "parses a sum with negative imaginary part" $
      p "42 - 42i" `shouldParse` (42 :+ (-42))

    it "parses a sum with negative summands" $
      p "-42 - 42i" `shouldParse` ((-42) :+ (-42))

    it "parses a sum without whitespace" $
      p "-42-42i" `shouldParse` ((-42) :+ (-42))

  context "over floats" $ do
    let p = parse (L.complex L.float <* eof) ""

    it "parses an unsigned real part" $
      p "8.5" `shouldParse` (8.5 :+ 0)

    it "parses a negative real part" $
      p "-8.5" `shouldParse` ((-8.5) :+ 0)

    it "parses an unsigned imaginary part" $
      p "8.5i" `shouldParse` (0 :+ 8.5)

    it "parses a negative imaginary part" $
      p "-8.5i" `shouldParse` (0 :+ (-8.5))

    it "parses a sum with positive summands" $
      p "8.5 + 8.5i" `shouldParse` (8.5 :+ 8.5)

    it "parses a sum with negative real part" $
      p "-8.5 + 8.5i" `shouldParse` ((-8.5) :+ 8.5)

    it "parses a sum with negative imaginary part" $
      p "8.5 - 8.5i" `shouldParse` (8.5 :+ (-8.5))

    it "parses a sum with negative summands" $
      p "-8.5 - 8.5i" `shouldParse` ((-8.5) :+ (-8.5))

    it "parses a sum without whitespace" $
      p "-8.5-8.5i" `shouldParse` ((-8.5) :+ (-8.5))

hexSpec :: Spec
hexSpec = describe "hex" $ do
  let p = parse (L.hex <* eof) ""

  it "parses uppercase digits" $
    p "0xDEADBEEF" `shouldParse` (0xdeadbeef :: Word)

  it "parses lowercase digits" $
    p "0xdeadbeef" `shouldParse` (0xdeadbeef :: Word)

  it "parses mixed digits" $
    p "0xdeadBEEF" `shouldParse` (0xdeadbeef :: Word)

  it "required the 0x prefix" $
    p `shouldFailOn` "deadbeef"


fractionSpec :: Spec
fractionSpec = describe "fraction" $ do
  let p = parse (L.fraction L.decimal <* eof) ""

  it "parses simple decimals" $ do
    p "0" `shouldParse` (0 :: Ratio Int)
    p "1" `shouldParse` (1 :: Ratio Int)

  it "parses actual fractions" $ do
    p "1 / 2" `shouldParse` (1%2 :: Ratio Int)
    p "3/5" `shouldParse` (3%5 :: Ratio Int)

  it "refuses to parse division by zero" $
    p `shouldFailOn` "2 / 0"

nameSpec :: Spec
nameSpec = describe "name" $ do
  let p = parse (L.name <* eof) ""

  it "parses alphabetic" $ do
    p "foo" `shouldParse` "foo"
    p "q" `shouldParse` "q"
  
  it "parses underscore" $ do
    p "_" `shouldParse` "_"
    p "_a_" `shouldParse` "_a_"
  
  it "parses integers" $ do
    p "12" `shouldParse` "12"
    p "1_a43" `shouldParse` "1_a43"
  
  it "refuses empty" $ do
    p `shouldFailOn` ""
  
  it "refuses braces" $ do
    p `shouldFailOn` "("
    p `shouldFailOn` "{}"
    p `shouldFailOn` "[]"
    p `shouldFailOn` "]"
  
  it "refuses ambiguous symbols" $ do
    p `shouldFailOn` "^"
    p `shouldFailOn` ","
