{-# OPTIONS_GHC -fno-warn-orphans #-}
module Data.PartitionSpec (spec) where

import Test.Hspec

import Data.List (sort)

import Data.Partition
import Data.Partition.Common

spec :: Spec
spec = do
  fromFunctionSpec
  fromBlocksSpec
  statesOfBlockSpec
  toBlocksSpec
  isRefinementOfSpec

fromFunctionSpec :: Spec
fromFunctionSpec = describe "fromFunction" $ do
  it "allows to create an empty partition" $
    fromFunction 0 0 (const undefined) `shouldBe` fromBlocks []

  it "assigns the correct blocks" $
    let
      stateAssignment = [1,3,5,2,1,5,4]
      part = fromFunction 5 (length stateAssignment) (stateAssignment!!)
    in
      map (blockOfState part) [0..length stateAssignment-1] `shouldBe` stateAssignment

  it "sets the correct blocks count" $
    numBlocks (fromFunction 2 10 (const $ Block 1)) `shouldBe` 2

  it "sets the correct state count" $
    numStates (fromFunction 2 10 (const $ Block 1)) `shouldBe` 10


fromBlocksSpec :: Spec
fromBlocksSpec = describe "fromBlocks" $ do
  it "assigns the correct blocks" $
    let
      part = fromBlocks [[0], [1,3], [2,4,5]]
    in
      map (blockOfState part) [0..5] `shouldBe` [0,1,2,1,2,2]

  it "sets the correct blocks count" $
    numBlocks (fromBlocks [[0], [1,3], [2,4,5]]) `shouldBe` 3

  it "sets the correct state count" $
    numStates (fromBlocks [[0], [1,3], [2,4,5]]) `shouldBe` 6

statesOfBlockSpec :: Spec
statesOfBlockSpec = describe "statesOfBlock" $ do
  it "returns the correct blocks" $
    let
      blcks = [[0], [1,3], [2,4,5]]
      part = fromBlocks blcks
    in
      map (statesOfBlock part) [0..2] `shouldBe` blcks

toBlocksSpec :: Spec
toBlocksSpec = describe "toBlocks" $ do
  it "works for an example" $
    let part = [[0], [1,3], [2,4,5]]
    in map sort (toBlocks (fromBlocks part)) `shouldBe` part

isRefinementOfSpec :: Spec
isRefinementOfSpec = describe "isRefinementOf" $ do
  it "correctly decides a positive example" $
    let part1 = fromBlocks [[0], [1,3], [2,4,5]]
        part2 = fromBlocks [[0], [1], [3], [4], [2,5]]
    in
      (part2 `isRefinementOf` part1) `shouldBe` True

  it "correctly decides a negative example" $
    let part1 = fromBlocks [[0], [1,3], [2,4,5]]
        part2 = fromBlocks [[0], [2], [3], [4], [1,5]]
    in
      (part2 `isRefinementOf` part1) `shouldBe` False

  it "is reflexive" $
    let part = fromBlocks [[0], [1,3], [2,4,5]]
    in
      (part `isRefinementOf` part) `shouldBe` True

  it "is not commutative" $
    let part1 = fromBlocks [[0], [1,3], [2,4,5]]
        part2 = fromBlocks [[0], [1], [3], [4], [2,5]]
    in
      (part1 `isRefinementOf` part2) `shouldBe` False
