module Data.RefinablePartitionSpec (spec) where

import Test.Hspec
import Test.QuickCheck

import Control.Exception (ErrorCall(..))
import Control.Monad.ST
import Data.Foldable
import Data.Function (on)
import Data.List (intersect, isInfixOf,sort,nubBy)
import Data.Maybe (isJust,isNothing,fromJust)
import Data.Ord (comparing)

import qualified Data.Vector as V
import qualified Data.Vector.Algorithms.Intro as V
import qualified Data.Vector.Unboxed as VU

import Data.RefinablePartition

spec :: Spec
spec = do
  makeSpec
  statesOfBlockSpec
  markSpec
  unmarkSpec
  hasMarkedSpec
  markedStatesSpec
  splitMarkedSpec
  splitBySpec
  groupBySpec

makeSpec :: Spec
makeSpec = describe "make" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "creates blocks of the correct size" $
    let test p = (,,) <$> blockSize p 0 <*> blockSize p 1 <*> blockSize p 2
    in runST (make 10 3 initPart >>= test) `shouldBe` (3,5,2)

  it "assignes the states to blocks correctly" $
    let test p = mapM (blockOfState p) [0..9]
    in runST (make 10 3 initPart >>= test) `shouldBe` [0,0,0,1,1,1,1,2,2,1]

  it "errors out with more blocks than states" $
    stToIO (make 2 3 initPart) `shouldThrow` errorContaining "states"

  it "errors out with zero blocks, but some states" $
    stToIO (make 2 0 initPart) `shouldThrow` errorContaining "blocks"

  it "doesn't mark any states initially" $
    let test p = mapM (isMarked p) [0..9]
    in runST (make 10 3 initPart >>= test) `shouldBe` replicate 10 False

statesOfBlockSpec :: Spec
statesOfBlockSpec = describe "statesOfBlock" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "returns the correct states" $
    VU.toList (runST (make 10 3 initPart >>= flip statesOfBlock 1))
      `shouldMatchList` [3,4,5,6,9]

  it "works with empty blocks" $
    VU.toList (runST (make 5 3 initPart >>= flip statesOfBlock 2))
      `shouldMatchList` []

  it "works with block containing all states" $
    VU.toList (runST (make 3 3 initPart >>= flip statesOfBlock 0))
      `shouldMatchList` [0,1,2]

markSpec :: Spec
markSpec = describe "mark" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "it marks the correct state" $
    let doMark i = runST $ do
          p <- make 10 3 initPart
          mark p i
          isMarked p i
    in
      property $ forAll (elements [0..9]) $ \(i::Int) ->
        doMark (fromIntegral i) === True

  it "doesn't mark another state" $
    let doMark i = runST $ do
          p <- make 10 3 initPart
          mark p i
          mapM (isMarked p) [x | x <- [0..9], x /= i]
    in
      property $ forAll (elements [0..9]) $ \i ->
        doMark i === replicate 9 False

  it "works when marking many states in one block" $
    let res = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [4,5,9]
          mapM (isMarked p) [4,5,9,6]
    in
      res `shouldBe` [True,True,True,False]

  it "works when marking all states in one block" $
    let res = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [4,5,6,9]
          mapM (isMarked p) [4,5,9,6]
    in
      res `shouldBe` [True,True,True,True]

unmarkSpec :: Spec
unmarkSpec = describe "unmark" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "it unmarks the correct state" $
    let doMark i = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [0..9]
          unmark p i
          isMarked p i
    in
      property $ forAll (elements [0..9]) $ \(i::Int) ->
        doMark (fromIntegral i) === False

  it "doesn't unmark other states" $
    let doMark i = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [0..9]
          unmark p i
          mapM (isMarked p) [x | x <- [0..9], x /= i]
    in
      property $ forAll (elements [0..9]) $ \i ->
        doMark i === replicate 9 True

  it "works when unmarking many states in one block" $
    let res = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [0..9]
          mapM_ (unmark p) [4,5,9]
          mapM (isMarked p) [4,5,9,6]
    in
      res `shouldBe` [False,False,False,True]

  it "works when marking all states in one block" $
    let res = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [0..9]
          mapM_ (unmark p) [4,5,6,9]
          mapM (isMarked p) [4,5,9,6]
    in
      res `shouldBe` [False,False,False,False]

  it "it is the inverse of mark" $
    let doMark i = runST $ do
          p <- make 10 3 initPart
          mark p i
          unmark p i
          hasMarked p =<< blockOfState p i
    in
      property $ forAll (elements [0..9]) $ \(i::Int) ->
        doMark (fromIntegral i) === False

hasMarkedSpec :: Spec
hasMarkedSpec = describe "hasMarked" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "initially reports false for every block" $
    let res = runST $ do
          p <- make 10 3 initPart
          mapM (hasMarked p) [0,1,2]
    in res `shouldBe` [False,False,False]

  it "reports false if we mark and then umark all states" $
    let res = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [0..9]
          mapM_ (unmark p) [0..9]
          mapM (hasMarked p) [0,1,2]
    in
      res `shouldBe` [False,False,False]

  it "reports a block as marked when one of its states get marked" $
    let doMark st = runST $ do
          p <- make 10 3 initPart
          mark p st
          blk <- blockOfState p st
          hasMarked p blk
    in
      property $ forAll (elements [0..9]) $ \i ->
        doMark i === True

  it "can handle the case where every state in the block is marked" $
    let doMark is blk = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) is
          hasMarked p blk
    in
      doMark [0,1,2] 0 `shouldBe` True

markedStatesSpec :: Spec
markedStatesSpec = describe "markedStates" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "returns the marked states" $
    let doMark states blk = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) states

          marked <- markedStates p blk
          blockStates <- VU.toList <$> statesOfBlock p blk
          return $ sort (intersect states blockStates) == sort (VU.toList marked)
    in
      property $
        forAll (arbitraryBlock 3) $ \block ->
        forAll (arbitraryStates 10) $ \states ->
          doMark states block === True

splitMarkedSpec :: Spec
splitMarkedSpec = describe "splitMarkedSpec" $ do
  let initPart s
        | s < 3 = 0
        | s < 7 || s > 8 = 1
        | otherwise = 2

  it "assignes all previously marked states to block 1" $
    let doIt markStates block = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) markStates
          markedPreviously <- statesOfBlock p block >>= VU.filterM (isMarked p)
          (a, _) <- splitMarked p block
          nowInBlock <- maybe (return VU.empty) (statesOfBlock p) a
          return (nowInBlock, markedPreviously)
    in
      property $
        forAll (arbitraryBlock 3) $ \block ->
        forAll (arbitraryStates 10) $ \states ->
          let (nowInBlock, markedPreviously) = doIt states block
          in nowInBlock === markedPreviously

  it "assignes all previously unmarked states to block 2" $
    let doIt markStates block = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) markStates
          unMarkedPreviously <- statesOfBlock p block
            >>= VU.filterM (fmap not . isMarked p)
          (_, b) <- splitMarked p block
          nowInBlock <- maybe (return VU.empty) (statesOfBlock p) b
          return (nowInBlock, unMarkedPreviously)
    in
      property $
        forAll (arbitraryBlock 3) $ \block ->
        forAll (arbitraryStates 10) $ \states ->
          let (nowInBlock, markedPreviously) = doIt states block
          in nowInBlock === markedPreviously

  it "unmarks all states in the new blocks" $
    let doIt markStates block = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) markStates
          blockStates <- statesOfBlock p block
          _ <- splitMarked p block
          VU.mapM (isMarked p) blockStates
    in
      property $
        forAll (arbitraryBlock 3) $ \block ->
        forAll (arbitraryStates 10) $ \states ->
          doIt states block `shouldSatisfy` VU.all (==False)

  it "works without marked states" $
    let doIt = runST $ do
          p <- make 10 3 initPart
          (Nothing, Just b) <- splitMarked p 1
          statesOfBlock p b
    in
      VU.toList doIt `shouldMatchList` [3,4,5,6,9]

  it "works without unmarked states" $
    let doIt = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) [3,4,5,6,9]
          (Just b, Nothing) <- splitMarked p 1
          statesOfBlock p b
    in
      VU.toList doIt `shouldMatchList` [3,4,5,6,9]

  it "gives the identity of the original block to the largest new block" $
    let doIt markStates block = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) markStates
          (a, b) <- splitMarked p block
          a' <- traverse (\x -> (x,) <$> blockSize p x) a
          b' <- traverse (\x -> (x,) <$> blockSize p x) b
          return (a', b')
    in
      property $
        forAll (arbitraryBlock 3) $ \block ->
        forAll (arbitraryStates 10) $ \states ->
          case doIt states block of
            (Just (a,asize), Just (b,bsize)) ->
              counterexample (show ((a,asize), (b,bsize))) $
                if asize > bsize then a === block else b === block
            _ -> property Discard

  it "shares the identity of the original block if only one new block is created" $
    let doIt markStates block = runST $ do
          p <- make 10 3 initPart
          mapM_ (mark p) markStates
          (a, b) <- splitMarked p block
          return (block, a, b)
    in
      property $
        forAll (arbitraryBlock 3) $ \block ->
        forAll (arbitraryStates 10) $ \states ->
          let (old,marked,unmarked) = doIt states block
          in isNothing unmarked ==> (Just old === marked)

splitBySpec :: Spec
splitBySpec = describe "splitBy" $ do
  let predicate = odd

  it "assignes all states matching the predicate to block 1" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let
          (oldStates, newA, _) = split block predicate
          matchingStates = V.convert $ V.filter predicate oldStates
        in
          not (null matchingStates) ==>
            (isJust newA .&. vSort (snd (fromJust newA)) === vSort matchingStates)

  it "assignes all states not matching the predicate to block 2" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let
          (oldStates, _, newB) = split block predicate
          nonMatchingStates = V.filter (not . predicate) oldStates
        in
          not (null nonMatchingStates) ==>
            (isJust newB .&. vSort (snd (fromJust newB)) === vSort nonMatchingStates)

  it "works when no states match the predicate" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let
          res@(oldStates, newA, newB) = split block (const False)
        in
          counterexample (show res) $
              isNothing newA
          .&. isJust newB
          .&. vSort (snd (fromJust newB)) === vSort oldStates

  it "works when all states match the predicate" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let
          res@(oldStates, newA, newB) = split block (const True)
        in
          counterexample (show res) $
              isNothing newB
          .&. isJust newA
          .&. vSort (snd (fromJust newA)) === vSort oldStates

  it "gives the old block identity to the largest block by default" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        case split block predicate of
          (_, Just (a, aStates), Just (b, bStates))
            | length aStates > length bStates -> a === block
            | otherwise                       -> b === block
          _ -> property Discard

  it "gives the old block identity to the second block if the first doesn't exist" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let
          (_, _, newB) = split block (const False)
        in
          isJust newB ==> fst (fromJust newB) === block

  where
    initPart s
      | s < 3 = 0
      | s < 7 || s > 8 = 1
      | otherwise = 2

    split ::
         Block
      -> (State -> Bool)
      -> ( V.Vector State
         , Maybe (Block, V.Vector State)
         , Maybe (Block, V.Vector State))
    split block predicate =
      runST $ do
        p <- make 10 3 initPart
        oldStates <- V.convert <$> statesOfBlock p block
        (a, b) <- splitBy p block predicate
        aStates <- traverse (fmap V.convert . statesOfBlock p) a
        bStates <- traverse (fmap V.convert . statesOfBlock p) b
        return (oldStates, (,) <$> a <*> aStates, (,) <$> b <*> bStates)


groupBySpec :: Spec
groupBySpec = describe "groupBy" $ do
  it "distributes the old states over the new blocks" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let (oldStates,newBlocks) = split block (`div` 3)
        in
          toList oldStates `shouldMatchList` toList (V.concat newBlocks)

  it "returns blocks that have each equal elements (according to attribute)" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let (_,newBlocks) = split block (`div` 3)
        in
          all (haveEqualElems (`div` 3)) newBlocks

  it "returns nonempty blocks" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let (_, newBlocks) = split block (`div` 3)
        in
          all (not . null) newBlocks

  it "returns blocks that have distinct elements from each other" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let (_, newBlocks) = split block (`div` 3)
            oneElemFromAll = map V.head newBlocks
        in
          nubBy ((==) `on` (`div` 3)) oneElemFromAll === oneElemFromAll

  it "gives one of the new blocks the old identity" $
    property $
      forAll (arbitraryBlock 3) $ \block ->
        let newBlocks = runST $ do
              p <- make 10 3 initPart
              groupBy p block (comparing (`div` 3))
        in
          filter (==block) newBlocks === [block]

  where
    initPart s
      | s < 3 = 0
      | s < 7 || s > 8 = 1
      | otherwise = 2

    split :: Block -> (State -> State) -> (V.Vector State, [V.Vector State])
    split block attribute = runST $ do
      p <- make 10 3 initPart
      oldStates <- V.convert <$> statesOfBlock p block
      newBlocks <- groupBy p block (comparing attribute)

      (oldStates,) <$> mapM (fmap (V.convert @VU.Vector @State @V.Vector) . statesOfBlock p) newBlocks

    haveEqualElems attribute lst =
      length (nubBy ((==) `on` attribute) (sort (toList lst))) <= 1

arbitraryBlock :: Int -> Gen Block
arbitraryBlock n = elements [Block x | x <- [0..n-1]]

arbitraryStates :: Int -> Gen [State]
arbitraryStates n = sublistOf [x | x <- [0..n-1]]

errorContaining :: String -> Selector ErrorCall
errorContaining s (ErrorCallWithLocation msg _) = s `isInfixOf` msg

vSort :: Ord a => V.Vector a -> V.Vector a
vSort = V.modify V.sort
